/*
	$Id: MeasuredBioMaterialTransfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BioMaterialEvent;
import net.sf.basedb.core.CommonItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.Protocol;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
This method provides methods to transfer MeasuredBioMaterial from base 1 to
BASE 2.

@author Gregory
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

abstract class MeasuredBioMaterialTransfer extends CommonItemTransfer
{
	/**
	Created date column index. In base 1 this should match one of the
	following sampleDate, labelingDate, extractionDate.
	 */
	private final int CREATED_DATE = 9;
	/**
	Quantity column index.
	 */
	private final int QUANTITY = 10;
	/**
	Protocol column index.
	 */
	private final int PROTOCOL = 11;
	
	private ProtocolTransfer protocolT;
	
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.MeasuredBioMaterialTransfer");

	MeasuredBioMaterialTransfer(UserTransfer userT, GroupTransfer groupT,
			ProtocolTransfer protocolT)
	{
		super(userT, groupT);
		log.debug("Checking that necessary transfers have been done before proceeding with MeasuredBioMaterials");
		this.protocolT = protocolT;
		if (!protocolT.isDone())
		{
			log.debug("Dependency: Protocols haven't been transfered");
			dependencyFailure();
		}
	}

	/**
	@param item <code>MeasuredBioMaterial</code>
	@param rs <code>ResultSet</code> IMPORTANT that the select statement
	       for this resultset begins with
	       <p>
	       "SELECT id, owner, gid, groupAccess, worldAccess, name, descr,
	       removed, date, quantity, protocol, ... "
	       </p>
	       The last date column should be one of sampleDate, labelingDate or
	       extractionDate.
	 */
	protected void setMeasuredBioMaterialData(DbControl dbc,
			MeasuredBioMaterial item, ResultSet rs)
	{
		setCommonItemData(dbc, (CommonItem) item, rs);
		try
		{
			item.setOriginalQuantity(rs.getFloat(QUANTITY));
			/*
			 * date maps to an event indicating when the bioMaterial was created
			 * This should be a of Type BioMaterialEvent.Type.CREATION
			 */
			BioMaterialEvent event = item.getCreationEvent();
			try
			{
				event.setEventDate(rs.getDate(CREATED_DATE));
			}
			catch (SQLException e)
			{
				event.setEventDate(null);
				if (log.isDebugEnabled())
				{
					log.debug("No EventDate was set on item:" + item.getType() + ":" + item.getName() + " Event:" + event.getEventType());
				}
			}
			Integer protocolId = protocolT.getBase2Id(rs.getInt(PROTOCOL));
			if (protocolId != null)
			{
				event.setProtocol(Protocol.getById(dbc, protocolId));
			}
		}
		catch (Exception e)
		{
			log.warn("Couldn't set MeasuredBioMaterialData");
			if (log.isDebugEnabled())
			{
				log.debug("Item:" + item.getType() + ":" + item.getName(), e);
			}
		}
	}
}
