/*
	$Id: ProgressThread.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Gregory Vincic
	Copyright (C) 2006 Jari Häkkinen, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import java.text.SimpleDateFormat;
import java.util.Date;

/**
This class provides methods for displaying the progress of a transfer. At a
set interval the thread checks whether it's counter has increased, if so
progress is printed to standard output.

@author Gregory
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

public class ProgressThread implements Runnable
{
	/**
	Progress counter limit. E.g. number of items to transfer.
	 */
	private int endCount = 0;
	private int counter = 0;
	/**
	Last counter printed to standard output.
	 */
	private int lastOutputCounter = 0;
	/**
	True if this thread is active.
	 */
	private boolean active = true;
	/**
	True if thread hasn't been started. One progress object can only be used
	started and stopped once.
	 */
	private boolean unused = true;
	/**
	Number of milliseconds to wait before trying to display the progress.
	 */
	private int updateInterval = 1500;
	/**
	printf format for an output
	 */
	private final String OUT_FORMAT = "%10d : %10d  %3d%%";
	/**
	Total duration between start and stop in milliseconds
	 */
	private int duration = 0;
	/**
	This progress thread
	 */
	Thread progress = null;
	/**
	 The itemType that is transfered
	 */
	private String itemType;

	/**
	@param endCount Total number of items that will be transfered
	 */
	ProgressThread(int endCount, String itemType)
	{
		this.endCount = endCount;
		this.itemType = itemType;
		progress = new Thread(this);
	}

	/**
	Starts progress
	 */
	public void start()
	{
		if (unused)
		{
			if (itemType != null)
			{
				printDate();
				System.out.println(itemType);
			}
			printDate();
			System.out.printf(
					"%10s   %10s     %10s %10s\n", "Left", "Done", "Items/s",
					"ETE");
			printDate();
			System.out.printf(OUT_FORMAT, endCount, 0, 0);
			System.out.println("");
			progress.start();
			unused = false;
		}
	}

	public void run()
	{
		while (active)
		{
			try
			{
				duration = duration + updateInterval;
				Thread.sleep(updateInterval);
				display();
			}
			catch (InterruptedException e)
			{}
		}
	}

	/**
	Decrease the value of the counter by one.
	 */
	public void decrease()
	{
		if (active)
		{
			counter--;
		}
	}

	/**
	Increase the counter by one.
	 */
	public void increase()
	{
		if (active)
		{
			counter++;
		}
	}

	/**
	Increase the counter by i.
	 */
	public void increase(int i)
	{
		if (active)
		{
			counter = counter + i;
		}
	}

	/**
	If the progress is active this method prints out status of the progress
	to standard output. Showing Items, Done, Percent Item/s and ETA.
	 */
	public void display()
	{
		if (active)
		{
			if (counter != lastOutputCounter)
			{
				printDate();
				int percent = (int) (((float) counter / (float) endCount) * 100f);
				int durationInSeconds = duration / 1000; // Should never be
															// zero
				float mean = (float) counter / (float) durationInSeconds;
				int itemsLeft = endCount - counter;
				System.out.printf(OUT_FORMAT, itemsLeft, counter, percent);
				System.out.printf("%9d", (int) mean);
				int secondsLeft = (int) ((float) itemsLeft / mean);
				int hh = (int) (secondsLeft / 3600);
				int mm = (int) ((secondsLeft % 3600) / 60);
				int ss = (secondsLeft % 3600) % 60;
				System.out.printf("   %02d:%02d:%02d\n", hh, mm, ss);
				lastOutputCounter = counter;
			}
		}
	}

	/**
	Stops the current progress. Stopped progress threads cannot be activated
	again.
	 */
	public void stop()
	{
		if (active)
		{
			display();
			printDate();
			System.out.println("-------------------------------------------------");
			active = false;
			updateInterval = 3600000; // A long time
		}
	}

	/**
	Prints date without linebreak.
	 */
	private void printDate()
	{
		System.out.print(new SimpleDateFormat("HH:mm:ss").format(new Date())
				+ " ");
	}
}
