/*
	$Id: ProtocolTypeTransfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005, 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.ProtocolType;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;

/**
Transfers protocol types that do not exist in BASE 2. If names in lowercase
match an existing protocol type it is skipped.

@author Gregory, Martin
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

public class ProtocolTypeTransfer extends Transfer
{
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.ProtocolTypeTransfer");
	
	HashMap<String, ProtocolType> existingProtocolTypes = null;

	ProtocolTypeTransfer()
	{
		existingProtocolTypes = new HashMap<String, ProtocolType>();
	}

	public void start()
	{
		try
		{
			log.info("Transferring ProtocolTypes...");
			int count = getInt("countProtocolTypes");
			initIdMap(count);
			// Get existing ProtocolTypes
			DbControl b2c = newDbControl();
			ItemResultList<ProtocolType> protocolTypes = ProtocolType.getQuery().list(
					b2c);
			for (ProtocolType protocolType : protocolTypes)
			{
				existingProtocolTypes.put(
						(protocolType.getName()).toLowerCase(), protocolType);
			}
			b2c.close();
			startProgress(count, "ProtocolTypes");
			run(prepareStatementFromFile("selectProtocolTypes"), 0, getBatchSize());
			done();
			log.info("Transferring ProtocolTypes...DONE");
		}
		catch (Exception e)
		{
			log.error("FAILED to transfer ProtocolTypes", e);
			stop();
		}
	}

	/**
	Creates a {@link ProtocolType} from the current position in
	<code>ResultSet</code>.
	
	@param b2c {@link DbControl} used to transfer this item
	@param rs <code>ResultSet</code> containing the data for this item.
	@return <code>null</code> the ProtocolType is already saved.
	@throws BaseException If a new ProtocolType cannot be created.
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	protected ProtocolType createItem(DbControl b2c, ResultSet rs)
			throws BaseException, SQLException
	{
		String name = (rs.getString(2)).toLowerCase();
		if (existingProtocolTypes.containsKey(name))
		{
			if (log.isDebugEnabled())log.debug("ProtocolType: " + rs.getString(2) + " already exists");
			return existingProtocolTypes.get(name);
		}
		else
		{
			// Create new ProtocolType if needed
			ProtocolType pt = ProtocolType.getNew(b2c);
			pt.setName(rs.getString(2));
			if (log.isDebugEnabled())
			{
				log.debug("Transfered PluginType: " + pt.getName() + "successfully");
				log.debug("--------------------------------------------------------------------------");
			}
			return pt;
		}
	}

	/**
	Returns a ProtocolTyp from the given base 1 protocol type id.
	
	@param dbc {@link DbControl} used to transfer this item
	@param base1ProtocolTypeId
	@return ProtocolType or null if a protocol with that id doesn't exist.
	@throws BaseException If a new ProtocolType cannot be created.
	 */
	public ProtocolType getProtocolTypeById(DbControl dbc,
			int base1ProtocolTypeId)
	{
		try
		{
			Integer id = getBase2Id(base1ProtocolTypeId);
			return id == null ? null : ProtocolType.getById(dbc, id);
		}
		catch (BaseException e)
		{
			log.warn("FAILED when getting protocol type by ID");
			if (log.isDebugEnabled()) log.debug("\t" + e);
		}
		return null;
	}
}
