/*
	$Id: Body.java 5256 2010-02-22 10:42:36Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.util.Values;

import java.util.Date;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;
import javax.servlet.jsp.tagext.BodyContent;

/**
	This tag defines the body section of a page. It must be located inside
	a {@link Page &lt;base:page&gt;} ag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:body
      clazz=...
      style=...
      onload=...
      onunload=...
      onkeypress=...
      attributes=...
   &gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>button</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>onload</td>
		<td>-</td>
		<td>no</td>
		<td>
			A JavaScript statement that is put into the <code>onload</code> attribute
			of the <code>body</code> tag.
		</td>
	</tr>
	<tr>
		<td>onunload</td>
		<td>-</td>
		<td>no</td>
		<td>
			A JavaScript statement that is put into the <code>onunload</code> attribute
			of the <code>body</code> tag.
		</td>
	</tr>
	<tr>
		<td>onukeypress</td>
		<td>-</td>
		<td>no</td>
		<td>
			A JavaScript statement that is put into the <code>onkeypress</code> attribute
			of the <code>body</code> tag.
		</td>
	</tr>
	<tr>
		<td>attributes</td>
		<td>-</td>
		<td>no</td>
		<td>Other attributes that is needed in the <code>body</code> tag.
		</td>
	</tr>
	</table>
	
	@author Nicklas
	@version 2.0
*/
public class Body
	extends TagSupport
{
	
	/**
		
	*/
	private static final long serialVersionUID = 6814973550614834311L;

	/**
		The parent &lt;base:page&gt; tag.
	*/
	private Page page;

	/**
		Optional <code>class</code> attribute
	*/
	private String clazz = null;
	
	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;


	/**
		A JavaScript statement that is put into the <code>onload</code> attribute
		of the <code>body</code> tag.
	*/
	private String onLoad = null;

	/**
		A JavaScript statement that is put into the <code>onunload</code> attribute
		of the <code>body</code> tag.
	*/
	private String onUnload = null;

	/**
		A JavaScript statement that is put into the <code>onkeypress</code> attribute
		of the <code>body</code> tag.
	*/
	private String onKeyPress = null;
	
	/**
		Any other attributes that is needed in the body tag.
	*/
	private String attributes = null;


	/*
		Taglib initialization methods
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}

	public void setStyle(String style)
	{
		this.style = style;
	}
	public String getStyle()
	{
		return style;
	}

	public void setOnload(String onLoad)
	{
		this.onLoad = onLoad;
	}
	public String getOnload()
	{
		return onLoad;
	}

	public void setOnunload(String onUnload)
	{
		this.onUnload = onUnload;
	}
	public String getOnunload()
	{
		return onUnload;
	}
	
	public void setOnkeypress(String onKeyPress)
	{
		this.onKeyPress = onKeyPress;
	}
	public String getOnkeypress()
	{
		return onKeyPress;
	}
	
	public void setAttributes(String attributes)
	{
		this.attributes = attributes;
	}
	public String getAttributes()
	{
		return attributes;
	}

	public int doStartTag()
		throws JspException
	{
		if (!(getParent() instanceof Page)) throw new JspTagException("Tag <base:body> must be inside a <base:page> tag");
		page = (Page)getParent();

		StringBuilder sb = new StringBuilder();
		switch (page.getTypeCode())
		{
		case Page.PAGE_TYPE_POPUP:
			String theClazz = getClazz() == null ? "popup" : getClazz();
			sb.append("<body class=\"").append(theClazz).append("\"");
			if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
			if (getOnload() != null) sb.append(" onload=\"").append(getOnload()).append("\"");
			if (getOnunload() != null) sb.append(" onunload=\"").append(getOnunload()).append("\"");
			if (getOnkeypress() != null) sb.append(" onkeypress=\"").append(getOnkeypress()).append("\"");
			if (getAttributes() != null) sb.append(" ").append(getAttributes());
			sb.append(">\n");
			break;

		case Page.PAGE_TYPE_INCLUDE:
			// Nothing should be generated
			break;
			
		default:
			theClazz = getClazz() == null ? "default" : getClazz();
			sb.append("<body class=\"").append(theClazz).append("\"");
			if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
			if (getOnload() != null) sb.append(" onload=\"").append(getOnload()).append("\"");
			if (getOnunload() != null) sb.append(" onunload=\"").append(getOnunload()).append("\"");
			if (getOnkeypress() != null) sb.append(" onkeypress=\"").append(getOnkeypress()).append("\"");
			if (getAttributes() != null) sb.append(" ").append(getAttributes());
			sb.append(">\n");
			SessionControl sc = page.getSessionControl();
			String menuName = page.getMenu();
			if ("auto".equals(menuName)) menuName = sc != null && sc.isLoggedIn() ? "standard" : "login";
			String menuHtml = null;
			if (sc != null && pageContext.getRequest().getParameter("refresh_menu") == null)
			{
				// The menu cache can be used, but check expiration date
				Date expirationDate = (Date)sc.getSessionSetting("menu."+menuName+".expire");
				if (expirationDate != null && expirationDate.after(new Date()))
				{
					menuHtml = (String)sc.getSessionSetting("menu."+menuName+".html");
				}
			}
			if (menuHtml == null)
			{
				try
				{
					/*
					We don't want the include() method to flush the pageContext.getOut() stream
					*/
					BodyContent menuBody = pageContext.pushBody();
					pageContext.include("/include/menu.jsp?name="+menuName);
					pageContext.popBody();
					menuHtml = menuBody.getString();
					if (sc != null) 
					{
						sc.setSessionSetting("menu."+menuName+".html", menuHtml);
						sc.setSessionSetting("menu."+menuName+".expire", new Date(System.currentTimeMillis()+60000*Application.permissionTimeout())); 
					}
				}
				catch (Exception ex)
				{
					throw new JspTagException(ex);
				}
			}
			if (menuHtml != null) sb.append(menuHtml);
			try
			{
				BodyContent menuBody = pageContext.pushBody();
				pageContext.include("/include/menu.jsp?name=recentItems");
				pageContext.include("/include/menu.jsp?name=projects");
				pageContext.popBody();
				sb.append(menuBody.getString());
			}
			catch (Exception ex)
			{
				throw new JspTagException(ex);				
			}
			if ((page.getTitle() != null) && !page.getTitle().equals("")) sb.append("\t<h1>").append(page.getTitle()).append("</h1>\n");
			break;
		}

		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex);
		}
		return EVAL_BODY_INCLUDE;
	}

	public int doEndTag() 
		throws JspException
	{
		if (page.getTypeCode() != Page.PAGE_TYPE_INCLUDE)
		{
			StringBuilder sb = new StringBuilder();
			sb.append("<script langage=\"JavaScript\" type=\"text/javascript\" defer=\"defer\">\n");
			if (pageContext.getRequest().getParameter("popmessage") != null)
			{
				sb.append("alert('").append(pageContext.getRequest().getParameter("popmessage")).append("');\n");
			}
			SessionControl sc = page.getSessionControl();
			String longTexts = Values.getString(sc != null ? sc.getUserClientSetting("text.long") : null, "display");
			if ("click".equals(longTexts))
			{
				sb.append("\t\twindow.textOverflowChecker = new TextOverflowChecker('div', 'constrained');\n");
				sb.append("\t\twindow.textOverflowChecker.checkAll();\n");
			}
			sb.append("</script>\n");
			sb.append("</body>\n");
			try
			{
				pageContext.getOut().print(sb.toString());
			}
			catch (Exception ex)
			{
				throw new JspTagException(ex.getMessage());
			}
		}
		return EVAL_PAGE;
	}
}



