/**
	$Id: SimpleSignalProgressReporter.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.web.util;

import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SimpleProgressReporter;
import net.sf.basedb.core.signal.Signal;
import net.sf.basedb.core.signal.SignalException;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalReceiver;
import net.sf.basedb.core.signal.SignalTransporter;

/**
  	This class can be used to report progress and hold signal receiver 
  	for those export plugins that download immediately.
  	There can only be one SignalHandler registerd at a time to this class.
 
  	@author Martin
  	@version 2.8
 	
 */
public class SimpleSignalProgressReporter 
	extends SimpleProgressReporter
	implements SignalReceiver
{

	/**
		Log signals processing.
	*/
	private static final org.apache.log4j.Logger logger = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.signal.SimpleSignalProgressReporter");
	
	private SignalHandler handler = null;
	private Thread notifyThread  = null;
	
	public SimpleSignalProgressReporter(ProgressReporter progress)
	{
		super(progress);
	}

	/* 	(non-Javadoc)
	  	@see net.sf.basedb.core.signal.SignalReceiver#close(int)
	 */
	@Override
	public void close(int wait)
	{
		if (handler == null) return;
		if (wait >= 0)
		{
			Thread abort = new Thread(
					new Runnable()
					{
						public void run()
						{
							sendToAll(Signal.ABORT);
						}
					}
			);
			abort.setDaemon(true);
			abort.start();
			if (wait > 0)
			{
				try
				{
					notifyThread = Thread.currentThread();
					Thread.sleep(wait);
					notifyThread = null;
				}
				catch (InterruptedException ex)
				{
					// Good, all handlers got the signal before timeout
				}
			}
		}
		handler = null;
	}
	
	/**
	 	This class does not need any signal transporter.
	  	@see net.sf.basedb.core.signal.SignalReceiver#getSignalTransporterClass()
	 */
	@Override
	public Class<? extends SignalTransporter> getSignalTransporterClass()
	{
		return null;
	}

	/**
	 	Nothing need to be initilized.
	 	@see net.sf.basedb.core.signal.SignalReceiver#init(java.lang.String)
	 */
	@Override
	public void init(String receiverId)
	{
		// Do nothing
	}

	/**
	  	This return null when there is no need for a global id.	  	
	  	@see net.sf.basedb.core.signal.SignalReceiver#registerSignalHandler(net.sf.basedb.core.signal.SignalHandler)
	 */
	@Override
	public String registerSignalHandler(SignalHandler handler)
	{
		if (notifyThread != null) throw new SignalException("Signal receiver is shutting down.");
		if (this.handler != null) 
		{	
			throw new SignalException("This receiver has already a handler registered");
		}
		this.handler = handler;
		return null;
	}

	/* 	(non-Javadoc)
	  	@see net.sf.basedb.core.signal.SignalReceiver#sendToAll(net.sf.basedb.core.signal.Signal)
	 */
	@Override
	public void sendToAll(Signal signal)
	{
		if (handler == null) return;
		logger.info("Sending " + signal);
		// Copy the signal handler to a temporary collection,
		// since sending a signal may cause a handler to get 
		// unregistered, which may fail or block due to synchronization
		SignalHandler tempHandler = null;
		synchronized (handler)
		{
			try
			{			
				tempHandler = handler.getClass().newInstance();
			}
			catch (Throwable th)
			{}
		}		
		if (tempHandler.supports(signal)) tempHandler.handleSignal(signal);
	}

	/* 	(non-Javadoc)
	  	@see net.sf.basedb.core.signal.SignalReceiver#unregisterSignalHandler(net.sf.basedb.core.signal.SignalHandler)
	 */
	@Override
	public void unregisterSignalHandler(SignalHandler handler)
	{
		logger.info("Unregister signal handler");
		if (this.handler != null) 
		{
			this.handler = null;
			// If called from the close() method() this is waiting for
			if (notifyThread != null && this.handler == null)
			{
				logger.info("Interrupting closedown thread");
				notifyThread.interrupt();
			}
		}
	}

}
