/*
	$Id: AbstractHqlQuery.java 5261 2010-03-05 13:26:11Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.HqlQuery;
import net.sf.basedb.core.query.QueryParameter;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.signal.SignalException;

import java.util.Set;
import java.util.Map;
import java.util.Collections;
import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.hibernate.Query;

/**
	An abstract implementation of the {@link HqlQuery} interface. The
	class can be used to build and manage just about any type of HQL
	query. It has no function for executing the query or returning
	the result. Subclasses may use {@link #getMainHqlQuery(DbControl)}
	to generate a Hibernate Query object that is ready to be executed.

	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-03-05 14:26:11 +0100 (Fri, 05 Mar 2010) $
*/
abstract class AbstractHqlQuery
	extends AbstractQuery
	implements HqlQuery
{

	/**
		Log all parameter bindings to prepared statement.
	*/
	private static final org.apache.log4j.Logger logParam = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.query.bind.AbstractHqlQuery");
		
	/**
		So we don't always have to call logParam.debug()
	*/
	private static final boolean debugEnabled = logParam.isDebugEnabled();

	/**
		The alias for the root table. This is a requirement
		in HQL that is not needed in SQL.
		SELECT ... FROM [rootTable] as [rootAlias]
	*/
	private final String rootAlias;
	
	/**
		If the query should be executed in a stateless session or not.
	*/
	private final boolean stateless;
	
	/**
		The most recent DbControl used to build/execute the main query.
	*/
	private DbControl lastMainDc;

	/**
		The most recent DbControl used to build/execute the count query.
	*/
	private DbControl lastCountDc;
	
	/**
		The most recent main query.
	*/
	private org.hibernate.Query lastMainQuery;
	
	/**
		The most recent count query.
	*/
	private org.hibernate.Query lastCountQuery;
	
	/**
		If the results should be cached or not.
	*/
	private boolean cacheResults;
	
	/**
		Create a new HQL query. The root table and alias must be given.
	*/
	AbstractHqlQuery(String root, String alias, boolean stateless)
	{
		super(root);
		this.rootAlias = alias;
		this.stateless = stateless;
	}
	
	/*
		From the Query interface
		-------------------------------------------
	*/
	/**
		@return {@link QueryType#HQL}
	*/
	@Override
	public QueryType getQueryType()
	{
		return QueryType.HQL;
	}
	
	/**
		The alias of the item that is the root of this query.
		@see Item#getAlias()
	*/
	@Override
	public String getRootAlias()
	{
		return rootAlias;
	}
	
	@Override
	public boolean isReadonly()
	{
		return super.isReadonly() || lastMainQuery != null;
	}
	
	/**
		Reset all non-permanent query elements of the query and 
		clear cached queries.
	*/
	@Override
	public void reset()
	{
		super.reset();
		lastMainQuery = null;
		lastCountQuery = null;
		lastMainDc = null;
		lastCountDc = null;
	}
	@Override
	public long count(DbControl dc)
		throws BaseException
	{
		long count = -1;
		try
		{
			Query query = getCountHqlQuery(dc);
			count = isStateless() ?
				QueryExecutor.loadData(Long.class, query, dc.getStatelessSession(), dc.getSessionControl()) :
				HibernateUtil.loadData(Long.class, query);
		}
		catch (InterruptedException ex)
		{
			throw new SignalException("Aborted by user", ex);
		}
		return count;
	}
	// -------------------------------------------
	/*
		From the HqlQuery interface
		-------------------------------------------
	*/
	@Override
	public void setCacheResult(boolean flag)
	{
		this.cacheResults = flag;
	}
	@Override
	public boolean isCachingResult()
	{
		return cacheResults;
	}
	// -------------------------------------------
	
	/**
		Build the main query and set parameter values for it. If the query has been 
		executed with the same <code>DbControl</code> before and hasn't been {@link #reset()} 
		the cached query is returned. If the parameters values have been changed the new
		values are used.
		
		@param dc The DbControl to use for executing the query
		@return A <code>org.hibernate.Query</code> object
	*/
	org.hibernate.Query getMainHqlQuery(DbControl dc)
	{
		if (lastMainQuery == null || lastMainDc != dc)
		{
			// If we are using the same DbControl as before we don't have to rebuild the query
			String mainHql = getMainQuery(dc, true);
			if (stateless)
			{
				lastMainQuery =  HibernateUtil.createQuery(dc.getStatelessSession(), mainHql);
			}
			else
			{
				lastMainQuery =  HibernateUtil.createQuery(dc.getHibernateSession(), mainHql);
			}
		}
		lastMainDc = dc;
		lastMainQuery.setCacheable(isCachingResult());
		setParameters(lastMainQuery, getParameters());
		if (getFirstResult() >= 0) lastMainQuery.setFirstResult(getFirstResult());
		if (getMaxResults() > 0) lastMainQuery.setMaxResults(getMaxResults());
		return lastMainQuery;
	}
	
	/**
		Build the count query and set parameter values for it. If the query has been 
		executed with the same <code>DbControl</code> before and hasn't been {@link #reset()} 
		the cached query is returned. If the parameters values have been changed the new
		values are used.
		
		@param dc The DbControl to use for executing the query
		@return A <code>org.hibernate.Query</code> object
	*/
	org.hibernate.Query getCountHqlQuery(DbControl dc)
	{
		if (lastCountQuery == null || lastCountDc != dc)
		{
			// If we are using the same DbControl as before we don't have to rebuild the query
			String countHql = getCountQuery(dc, true);
			if (stateless)
			{
				lastCountQuery =  HibernateUtil.createQuery(dc.getStatelessSession(), countHql);
			}
			else
			{
				lastCountQuery =  HibernateUtil.createQuery(dc.getHibernateSession(), countHql);
			}
		}
		lastCountDc = dc;
		lastCountQuery.setCacheable(true);
		setParameters(lastCountQuery, getParameters());
		return lastCountQuery;
	}
	
	/**
		If queries are using the stateless Hibernate session or the regular
		session.
		@return TRUE if the stateless session is used, FALSE otherwise
	*/
	protected boolean isStateless()
	{
		return stateless;
	}

	/**
		Processes the query so that joined paths that are found in other
		parts of the query are replaced with the join alias.
	*/
	@Override
	protected String postProcessQuery(String query) 
	{
		return replaceWithJoinAliases(query);
	}

	/**
		Replaces parts of the query with joined aliases. This is needed
		for Hibernate to generate correct SQL. Eg. 
		select f from Foo f join f.bar b where f.bar.name = :name
		needs to be changed to:
		select f from Foo f join f.bar b where b.name = :name
		<p>
		We implement this by looking for the joined path and replacing it
		with the joined alias.
	*/
	private String replaceWithJoinAliases(String query)
	{
		Pattern p = Pattern.compile("join\\s+([a-zA-Z\\.]+)\\s+(\\w+)", Pattern.CASE_INSENSITIVE);
		Matcher m = p.matcher(query);
		while (m.find())
		{
			String join = m.group(1);
			String alias = m.group(2);
			query = query.replace(join + ".", alias + ".");
		}
		return query;
	}
	
	
	private static final Set<Integer> ZERO_SET = Collections.singleton(0);

	/**
		Set parameter values on a query.
		@param query The Hibernate query object
		@param parameters A map containing parameter names and values
	*/
	private void setParameters(org.hibernate.Query query, Map<String, QueryParameter> parameters)
	{
		if (parameters == null) return;
		if (debugEnabled)
		{
			logParam.debug("Binding parameters to query: " + query.getQueryString());
		}
		for (Map.Entry<String, QueryParameter> entry : parameters.entrySet())
		{
			String name = entry.getKey();
			QueryParameter qp = entry.getValue();
			Type valueType = qp.getType();
			Object value = qp.getValue();
			if (value instanceof Collection)
			{
				Collection c = (Collection)value;
				if (debugEnabled)
				{
					logParam.debug("Binding collection '"+value+"' to parameter :"+name);
				}
				if (valueType == null)
				{
					query.setParameterList(name, c.size() == 0 ? ZERO_SET : c);
				}
				else
				{
					query.setParameterList(name, c.size() == 0 ? ZERO_SET : c, valueType.getHibernateType());
				}
			}
			else
			{
				if (debugEnabled)
				{
					logParam.debug("Binding '"+value+"' to parameter :"+name);
				}
				if (valueType == null)
				{
					query.setParameter(name, value);
				}
				else
				{
					query.setParameter(name, value, valueType.getHibernateType());
				}
			}
		}
	}

}
