/*
	$Id: DateUtil.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.text.SimpleDateFormat;
import java.util.Date;

/**
	Utility methods that will make it easier to handle date values.
	Especially, dates must be cloned whenver they are returned to
	or passed from client applications.

	@author Nicklas
	@version 2.0
*/
public class DateUtil
{

	/**
		Make a copy of a <code>Date</code> object. Null values are allowed.

		@param value The date to check
		@param name The name of the attribute to use if an exception is thrown
		@return A copy of the original date
	*/
	public static Date setNullableDate(Date value, String name)
	{
		return copy(value);
	}

	/**
		Make a copy of a <code>Date</code> object. Null values are not allowed.

		@param value The date to check. Can not be null.
		@param name The name of the attribute to use if an exception is thrown
		@return A copy of the original date
	 	@throws InvalidUseOfNullException If the value parameter is null.
	*/
	public static Date setNotNullDate(Date value, String name)
		throws InvalidUseOfNullException
	{
		if (value == null) throw new InvalidUseOfNullException(name);
		return copy(value);
	}


	/**
		Get a copy of the date. Null values are allowed.
	 	@param value Date value to copy 
	 	@return A cloned Date object or null if argument is null.
	*/
	public static Date copy(Date value)
	{
		return value == null ? null : (Date)value.clone();
	}

	private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");
	
	/**
	  	Parses a string to create a <code>Date</code>. This method supports
	  	date in yyyy-MM-dd format or as long timevalues.
	  	 
	  	@param value the <code>String</code> to be parsed
	  	@return a <code>Date</code> object
	  	@throws InvalidDataException if <code>value</code> isn't a valid date.
	*/
	public static Date parseString(String value)
		throws InvalidDataException
	{
		Date result = null;
		if (value != null)
		{
			try
			{
				result = DATE_FORMAT.parse(value);
			}
			catch (Exception ex)
			{
				try
				{
					result = new Date(new Long(value));
				}
				catch (Exception ex2)
				{
					throw new InvalidDataException("The value "+value+" is not a valid date.");
				}
			}
		}
		return result;
	}
	
	/**
		Formats a date in yyyy-MM-dd format.
	 	@param d The date to format. 
	 	@return Formated date as a String.
		@since 2.2
	*/
	public static String formatDate(Date d)
	{
		return DATE_FORMAT.format(d);
	}
	
}
