/*
	$Id: Extract.java 5279 2010-03-23 12:04:03Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.data.ExtractData;

import java.util.Set;
import java.util.HashSet;

/**
	This class is used to represent extract items. An extract is
	created from a {@link Sample} or by combining other extracts
	(pooling). It is also possible to create
	standalone extracts, with no reference to a sample or to other
	extracts.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2010-03-23 13:04:03 +0100 (Tue, 23 Mar 2010) $
*/
public class Extract
	extends MeasuredBioMaterial<ExtractData>
{

	/**
		The type of item represented by this class.
		@see Item#EXTRACT
		@see #getType()
	*/
	public static final Item TYPE = Item.EXTRACT;
	
	/**
		Create a new <code>Extract</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>Extract</code> item
		@throws BaseException If there is an error
	*/
	public static Extract getNew(DbControl dc)
		throws BaseException
	{
		Extract e = dc.newItem(Extract.class);
		e.setName("New extract");
		return e;
	}

	/**
		Get an <code>Extract</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Extract</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Extract getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Extract e = dc.loadItem(Extract.class, id);
		if (e == null) throw new ItemNotFoundException("Extract[id="+id+"]");
		return e;
	}

	/**
		Get a query that returns extracts.
		@return An {@link ItemQuery} object.
	*/
	public static ItemQuery<Extract> getQuery()
	{
		return new ItemQuery<Extract>(Extract.class);
	}
	
	
	Extract(ExtractData extractData)
	{
		super(extractData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		Get the sample or pooled extracts.
	*/
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		try
		{
			if (isPooled())
			{
				ItemQuery<? extends MeasuredBioMaterial> query = getCreationEvent().getSources();
				query.include(Include.ALL);
				annotatable.addAll(query.list(getDbControl()));
			}
			else if (getData().getParent() != null)
			{
				annotatable.add(getSample());
			}
		}
		catch (PermissionDeniedException ex)
		{}
		return annotatable;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>no {@link LabeledExtract}:s has been created from this item
		<li>no {@link Extract}:s has been created from this item
		</ul>
	*/
	public boolean isUsed()
		throws BaseException
	{		
		return countLabeledExtracts() > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link LabeledExtract}:s created from this extract
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_LABELED_EXTRACTS_FOR_EXTRACT", "le.id");
			/*
				SELECT {1}
				FROM LabeledExtractData le
				WHERE le.parent = :extract
			*/
		query.setEntity("extract", this.getData());
		addUsingItems(using, Item.LABELEDEXTRACT, query);
		return using;
	}
	// -------------------------------------------

	/**
		Get the {@link Sample} that is the parent of this extract.
		@return A <code>Sample</code> object or null if this is a pooled or
			standalone extract
		@throws PermissionDeniedException If the logged in user doesn't have read
			permission for the sample
		@throws BaseException If there is another error
	*/
	public Sample getSample()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Sample.class, getData().getParent());
	}

	/**
		Set the {@link Sample} item that is the parent
		of this extract.
		@param sample The new <code>Sample</code> item
		@param usedQuantity The quantity that was used from the sample when 
			the extract was created
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission for this extract or use permission for the sample
		@throws InvalidDataException If this is a pooled extract
		@throws BaseException If there is another error
	*/
	public void setSample(Sample sample, Float usedQuantity)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (sample != null) sample.checkPermission(Permission.USE);
		if (isPooled())
		{
			throw new InvalidDataException("A pooled extract can't have a parent sample");
		}
		getData().setParent(sample == null ? null : sample.getData());
		getCreationEvent().setSource(sample, usedQuantity);
	}
	
	/**
		Create a new {@link LabeledExtract} from this extract.
		@param label The labeling compound used to label the extract
		@param usedQuantity The quantity used from this extract or null
			if not known
		@return The new <code>LabeledExtract</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			use permission for this extract or use permission for the label
		@throws BaseException If there is another error
	*/
	public LabeledExtract newLabeledExtract(Label label, Float usedQuantity)
		throws PermissionDeniedException, BaseException
	{
		LabeledExtract le = LabeledExtract.getNew(getDbControl(), label);
		le.setExtract(this, usedQuantity);
		return le;
	}

	/**
		Get a query that returns all labeled extracts created from this extract.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<LabeledExtract> getLabeledExtracts()
	{
		ItemQuery<LabeledExtract> query = LabeledExtract.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("parent"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Creates a query that returns all child (pooled) extracts that
		has this extracts as one of it's sources.
		@return A query
		@since 2.10
	*/
	public ItemQuery<Extract> getChildExtracts()
	{
		ItemQuery<Extract> query = Extract.getQuery();
		query.joinPermanent(Hql.innerJoin("creationEvent", Item.BIOMATERIALEVENT.getAlias()));
		query.joinPermanent(Hql.innerJoin(Item.BIOMATERIALEVENT.getAlias(), "sources", "src"));
		query.restrictPermanent(Restrictions.eq(Hql.index("src", null), Hql.entity(this)));
		return query;
	}
	
	/**
	  	Gets the number of {@link LabeledExtract} that are create from this item.
	  	@return Number of <code>LabeledExtracts</code>
	  	@throws BaseException If there is any error.
	*/
	public long countLabeledExtracts()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_LABELED_EXTRACTS_FOR_EXTRACT", "count(*)");
		/*
			SELECT {1}
			FROM LabeledExtractData le
			WHERE le.parent = :extract
		*/
		query.setEntity("extract", this.getData());
		return HibernateUtil.loadData(Long.class, query);
	}
}

