/*
	$Id: Group.java 5077 2009-08-26 12:03:37Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.GroupData;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.Collections;

/**
	This class is used to represent groups. Groups are used in the
	permission system to make it possible for a user to share items
	with other users.
	<p>
	Groups are intended to represent the actual organisation of
	a company/organisation. For example a top-level group "The company"
	may contain the sub-groups "Sales", "Marketing" and "Research".
	Individual users should be added to the sub-groups only.
	<p>
	BASE comes with the predefined group <code>Everyone</code>,
	which cannot contain any members, except that the permission system
	works as if everyone is a member.

	@author Nicklas
	@version 2.0
	@see Role
	@see Project
	@base.modified $Date: 2009-08-26 14:03:37 +0200 (Wed, 26 Aug 2009) $
*/
public class Group
	extends BasicItem<GroupData>
	implements Nameable, Removable, SystemItem, Registered
{
	/**
		The type of item represented by this class.
		@see Item#GROUP
		@see #getType()
	*/
	public static final Item TYPE = Item.GROUP;
	
	/**
		The id of the </code>Everyone</code> group where all users
		are (virtual) members.
	*/
	public static final String EVERYONE = "net.sf.basedb.core.Group.EVERYONE";

	/**
		This filter will limit a query to only return groups
		where the logged in user is a member unless the logged in user
		has generic read permission.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();

	/**
		Create a new <code>Group</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@return The new <code>Group</code> item
		@throws BaseException If there is an error
	*/
	public static Group getNew(DbControl dc)
		throws BaseException
	{
		Group g = dc.newItem(Group.class);
		g.setName("New group");
		g.getData().setEntryDate(new Date());
		return g;
	}

	/**
		Get a <code>Group</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Group</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Group getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Group g = dc.loadItem(Group.class, id);
		if (g == null) throw new ItemNotFoundException("Group[id="+id+"]");
		return g;
	}

	/**
		Load all groups withing groups.
		@param dc The DbControl to use for database access
		@param groups The groups to start with
		@return A set containing the starting groups, the groups that are members of the 
			start groups and the groups that are members following subgroups down as
			far as possible
	*/
	public static Set<Integer> getGroupsRecursive(DbControl dc, Set<Integer> groups)
	{
		return getGroupsRecursive(dc.getHibernateSession(), groups);
	}

	static Set<Integer> getGroupsRecursive(org.hibernate.Session session, Collection<Integer> groups)
	{
		Set<Integer> allGroups = new HashSet<Integer>(groups);
		if (allGroups.size() > 0)
		{
			
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_CHILDGROUPS_IDS_FOR_GROUPS");
			/*
				SELECT DISTINCT gg.childId
				FROM GroupGroups gg
				WHERE gg.parentId IN (:groups)
				AND gg.childId NOT IN (:groups)
			*/
			do
			{
				query.setParameterList("groups", allGroups, org.hibernate.Hibernate.INTEGER);
			} while (allGroups.addAll(HibernateUtil.loadList(Integer.class, query, null)));
		}
		return allGroups;
	}

	
	/**
		Get a query configured to retrieve groups. 
		If the logged in user doesn't have generic permission to all groups, only groups
		where the logged in user is a member are returned.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Group> getQuery()
	{
		return new ItemQuery<Group>(Group.class, RUNTIME_FILTER);
	}

	Group(GroupData groupData)
	{
		super(groupData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	public String getSystemId()
	{
		return getData().getSystemId();
	}
	public boolean isSystemItem()
	{
		return getSystemId() != null;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Checks if:
		<ul>
		<li>A user has been assigned this group as a quota group
		</ul>
		There are also other items that can reference a group:
		<ul>
		<li>DiscUsage
		<li>Other groups
		<li>ItemKeys
		<li>Projects
		<li>Users (as group members)
		</ul>
		The links to all these items are automatically deleted if the
		group is deleted and aren't included in this check.
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_USERS_FOR_QUOTAGROUP", "count(*)");
		/*
			SELECT {1}
			FROM UserData usr
			WHERE usr.quotaGroup = :theGroup
		*/
		query.setEntity("theGroup", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link User}:s assigned to this group as a quota group
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_USERS_FOR_QUOTAGROUP", "usr.id");
			/*
				SELECT {1}
				FROM UserData usr
				WHERE usr.quotaGroup = :theGroup
			*/
		query.setEntity("theGroup", this.getData());
		addUsingItems(using, Item.USER, query);
		return using;
	}

	/**
		If the logged in user is a member of this group, read permission
		is granted. If this is a system group, delete and create permissions are
		revoked.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem()) 
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		if (getSessionControl().isMemberOf(this))
		{
			granted |= Permission.grant(Permission.READ);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Get the {@link Quota} that applies to the group.
		@return A <code>Quota</code> item, or null if no quota has
			been assigned to this group
		@throws PermissionDeniedException  If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
	 	@throws BaseException If there is some other error. 
	*/
	public Quota getQuota() 
		throws PermissionDeniedException, BaseException 
	{
		return getDbControl().getItem(Quota.class, getData().getQuota());
	}
	
	/**
		Set the quota for the group.
		@param quota The new <code>Quota</code>, or null to
			disable quota for this group
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#WRITE} permission for the group
			or {@link Permission#USE} permission for the quota
		@throws InvalidDataException If the quota is null
	*/
	public void setQuota(Quota quota)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (getId() == SystemItems.getId(EVERYONE))
		{
			throw new PermissionDeniedException("Not allowed to set quota for the EVERYONE group");
		}
		if (quota != null) quota.checkPermission(Permission.USE);
		getData().setQuota(quota == null ? null : quota.getData());
	}

	/**
		Get the used number of bytes for the specified quota type and location.
		@param quotaType The {@link QuotaType}. Null is not allowed.
		@param location The location. Null is not allowed
		@return The number of bytes that have been used
	 	@throws BaseException If there is some kind of error.  
	*/
	public long getDiskUsage(QuotaType quotaType, Location location)
		throws BaseException
	{
		if (quotaType == null) throw new InvalidUseOfNullException("quotaType");
		if (location == null) throw new InvalidUseOfNullException("location");
		return getDbControl().getDiskUsage(getData(), quotaType.getData(), location);
	}
	
	/**
		If this group should be assigned to new users by default or not.
		@since 2.4
	*/
	public boolean isDefault()
	{
		return getData().isDefault();
	}
	
	/**
		If this group should be assigned to new users by default or not.
		Note! The client must call {@link User#addToDefaultRolesAndGroups()}.
		@param isDefault The new setting
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#WRITE} permission for the group, or if
			this group is the {@link #EVERYONE} group
		@since 2.4
	*/
	public void setDefault(boolean isDefault)
	{
		checkPermission(Permission.WRITE);
		if (getId() == SystemItems.getId(EVERYONE))
		{
			throw new PermissionDeniedException("Not allowed to add members to the EVERYONE group.");
		}
		getData().setDefault(isDefault);
	}
	
	/**
		Does this group have hidden members? In a group with hidden members
		one user doesn't get read permission to other members of the same
		group.
		@return TRUE if the members are hidden
		@since 2.13
	*/
	public boolean hasHiddenMembers()
	{
		return getData().getHiddenMembers();
	}
	
	/**
		Set the hidden members flag for this group.
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#WRITE} permission for the group, or if
			this group is the {@link #EVERYONE} group
		@since 2.13
	*/
	public void setHiddenMembers(boolean hiddenMembers)
	{
		checkPermission(Permission.WRITE);
		if (getId() == SystemItems.getId(EVERYONE))
		{
			throw new PermissionDeniedException("Not allowed to add members to the EVERYONE group.");
		}
		getData().setHiddenMembers(hiddenMembers);
	}
	
	/**
		Add a user as a member to this group. It is not allowed to add
		members to the {@link #EVERYONE} group.
		@param user The user to add
		@throws PermissionDeniedException If the
			logged in user doesn't have {@link Permission#WRITE}
			permission for the group and {@link Permission#USE}
			for the user
		@throws InvalidDataException If the user is null
	*/
	public void addUser(User user)
		throws PermissionDeniedException, InvalidDataException
	{
		if (getId() == SystemItems.getId(EVERYONE))
		{
			throw new PermissionDeniedException("Not allowed to add members to the EVERYONE group.");
		}
		checkPermission(Permission.WRITE);
		if (user == null) throw new InvalidUseOfNullException("user");
		user.checkPermission(Permission.USE);
		getData().getUsers().add(user.getData());
	}
	/**
		Remove a user that is a member of this group.
		@param user The user to remove
		@throws PermissionDeniedException If the
			logged in user doesn't have {@link Permission#WRITE}
			permission for the group and {@link Permission#USE}
			for the user
		@throws InvalidDataException If the user is null
	*/
	public void removeUser(User user)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (user == null) throw new InvalidUseOfNullException("user");
		user.checkPermission(Permission.USE);
		getData().getUsers().remove(user.getData());
		if (getData().equals(user.getData().getQuotaGroup()))
		{
			user.getData().setQuotaGroup(null);
		}
	}

	/**
		Checks if a user is a member of this group. A useris counted as
		a member if it has been added to this group with the {@link #addUser(User)}
		method or has been assigned this group as the {@link User#getQuotaGroup()}.
		
		@param user The user to check
		@return TRUE if the user is a member, FALSE otherwise
	*/
	public boolean isMember(User user)
	{
		if (getData().equals(user.getData().getQuotaGroup())) return true;
		return getData().getUsers().contains(user.getData());
	}

	/**
		Get a query that returns the users that
		are members of this group. This query excludes users that the logged 
		in user doesn't have permission to read.
	 	@return A query to select users from database with. 
	 	@throws BaseException If there is an error when building the query.
		
		@see User#getQuery()
	*/
	public ItemQuery<User> getUsers()
		throws BaseException
	{
		Restriction thisGroup = Restrictions.eq(
				Hql.property(Item.GROUP.getAlias(), "id"), 
				Hql.entity(this)
			);
		Restriction thisQuotaGroup = Restrictions.eq(
				Hql.property("qgrp", "id"),
				Hql.entity(this)
			);
		ItemQuery<User> query = User.getQuery();
		query.joinPermanent(Hql.leftJoin(null, "groups", Item.GROUP.getAlias(), thisGroup, false));
		query.joinPermanent(Hql.leftJoin("quotaGroup", "qgrp"));
		query.restrictPermanent(Restrictions.or(thisQuotaGroup, thisGroup));
		query.setDistinct(true);
		return query;
	}

	/**
		Add a group as a member to this group. This method does nothing if 
		one tries to add a group to itself. It is not allowed to add
		members to the {@link #EVERYONE} group.
		@param group The group to add
		@throws PermissionDeniedException If the
			logged in user doesn't have {@link Permission#WRITE}
			permission for both groups
		@throws InvalidDataException If the group is null
	*/
	public void addGroup(Group group)
		throws PermissionDeniedException, InvalidDataException
	{
		if (this.equals(group)) return;
		if (group == null) throw new InvalidUseOfNullException("group");
		int everyoneId = SystemItems.getId(EVERYONE);
		if (getId() == everyoneId || group.getId() == everyoneId)
		{
			throw new PermissionDeniedException("Not allowed to add members to the EVERYONE group.");
		}
		checkPermission(Permission.WRITE);
		group.checkPermission(Permission.WRITE);
		getData().getChildren().add(group.getData());
	}
	/**
		Remove a group that is a member of this group.
		@param group The group to remove
		@throws PermissionDeniedException If the
			logged in user doesn't have {@link Permission#WRITE}
			permission for both groups
		@throws InvalidDataException If the group is null
	*/
	public void removeGroup(Group group)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (group == null) throw new InvalidUseOfNullException("group");
		group.checkPermission(Permission.WRITE);
		getData().getChildren().remove(group.getData());
	}

	/**
		Checks if a group is a member of this group.
		@param group The group to check
		@return TRUE if the group is a member, FALSE otherwise
	*/
	public boolean isMember(Group group)
	{
		return getData().getChildren().contains(group.getData());
	}

	/**
		Get a query that returns the groups that
		are members of this group. This query excludes groups that the logged 
		in user doesn't have permission to read.
	 	@return A restricted query to select groups from database with. 
		
		@see Group#getQuery()
	*/
	public ItemQuery<Group> getGroups()
	{
		ItemQuery<Group> query = Group.getQuery();
		query.joinPermanent(Hql.innerJoin("parents", "p"));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias("p"), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Get a query that returns the projects that
		this group is a member of. This query excludes projects that the logged 
		in user doesn't have permission to read.
	 	@return A restricted item query to select projects from database with.
		
		@see Project#getQuery()
	*/
	public ItemQuery<Project> getProjects()
	{
		ItemQuery<Project> query = Project.getQuery();
		query.joinPermanent(Hql.innerJoin("groups", Item.GROUP.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.index(Item.GROUP.getAlias(), null), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		A runtime filter implementation that limits a query to only
		return groups where the logged in user is a member unless the logged in user
		has generic read permission.
	*/
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			SessionControl sc = dc.getSessionControl();
			if (sc.isDenied(query.getItemPermission(), Item.GROUP))
			{
				manager.enableFilter("denyAll");
			}
			else if (!sc.hasPermission(query.getItemPermission(), Item.GROUP))
			{
				if (sc.isLoggedIn() && query.getItemPermission() == Permission.READ)
				{
					// Load groups where the logged in user is a member
					org.hibernate.Filter filter = manager.enableFilter("memberOf");
					if (filter != null)
					{
						Set<Integer> groups = sc.getGroups();
						if (groups == null || groups.size() == 0) groups = Collections.singleton(0);
						filter.setParameterList("items", groups);
						filter.setParameter("owner", sc.getLoggedInUserId());
					}
				}
				else
				{
					// Do not load any groups if not logged in
					manager.enableFilter("denyAll");
				}
			}
		}
	}
}



