/*
	$Id: GroupPermissions.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.core;

import net.sf.basedb.core.data.GroupData;

import java.util.Collections;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;

/**
	This class is used to specify permissions to share an item to
	groups. It is a three-step procedure:
	<ul>
	<li>Fill this object with permissions for specific groups
	<li>Create a new {@link ItemKey} using this object as one of the
		input parameters to {@link ItemKey#getNewOrExisting(DbControl, UserPermissions, GroupPermissions)}
	<li>Assign the key to the item: {@link Shareable#setItemKey(ItemKey)}
	</ul>

	Don't forget to commit the transaction.
	
	@see ItemKey
	@see UserPermissions
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class GroupPermissions
{

	private final Map<GroupData,Integer> groupPermissions;
	private final Map<GroupData,Integer> originalPermissions;

	/**
		Create a new empty object.
	*/
	public GroupPermissions()
	{
		groupPermissions = new HashMap<GroupData,Integer>();
		originalPermissions = null;
	}
	
	/**
		Create a new object, and initialise it with the permissions from a
		<code>ItemKey</code>.
		@param itemKey The item key from which we take the current permissions
		@throws InvalidDataException If the item key is null
	*/
	public GroupPermissions(ItemKey itemKey)
		throws InvalidDataException
	{
		if (itemKey == null) throw new InvalidUseOfNullException("itemKey");
		groupPermissions = new HashMap<GroupData,Integer>(itemKey.getData().getGroups());
		originalPermissions = new HashMap<GroupData,Integer>(groupPermissions);
	}

	/**
		Grant permissions to a group. Use an empty <code>Set</code> or null
		to remove all permissions for the group.

		@param group The <code>Group</code>
		@param permissions The permissions to grant
		@throws InvalidDataException If the group is null
		@see Permission
	*/
	public void setPermissions(Group group, Set<Permission> permissions)
		throws InvalidDataException
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		if (permissions == null || permissions.isEmpty())
		{
			groupPermissions.remove(group.getData());
		}
		else
		{
			groupPermissions.put(group.getData(), Permission.grant(permissions));
		}
	}
	
	/**
		Grant permissions to a group, keeping those that have already been set.
		
		@param group The <code>Group</code>
		@param permissions The permissions to grant
		@throws InvalidDataException If the user is null
		@see Permission
	*/
	public void addPermissions(Group group, Set<Permission> permissions)
		throws InvalidDataException
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		if (permissions != null && !permissions.isEmpty())
		{
			int current = IntegerUtil.getInt(groupPermissions.get(group.getData()));
			groupPermissions.put(group.getData(), current | Permission.grant(permissions));
		}
	}

	/**
		Get the permissions for a group.
		@param group The <code>Group</code> for which we want to get the permission
		@return An <code>Set</code> containing the granted permissions, or an
			empty set if no permissions have been granted
		@throws InvalidDataException If the group is null
		@see #getOriginalPermissions(Group)
		@see Permission
	*/
	public Set<Permission> getPermissions(Group group)
		throws InvalidDataException
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		return Permission.fromInt(IntegerUtil.getInt(groupPermissions.get(group.getData())));
	}

	/**
		Get the original permissions for a group if this object was created with the
		{@link #GroupPermissions(ItemKey)} constructor.
		@param group The <code>Group</code> for which we want to get the permission
		@return An <code>Set</code> containing the original permissions, or an
			empty set
		@throws InvalidDataException If the group is null
		@see #getPermissions(Group)
		@since 2.6
	*/
	public Set<Permission> getOriginalPermissions(Group group)
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		if (originalPermissions == null) return Collections.emptySet();
		return Permission.fromInt(IntegerUtil.getInt(originalPermissions.get(group.getData())));
	}
	
	/**
		Check if the current permissions for the group has been changed from the original
		permissions. This method will return false if the {@link #getPermissions(Group)}
		returns a set with the same permissions {@link #getOriginalPermissions(Group)}.
		@param group The group to check
		@return TRUE if the permissions have changed, FALSE otherwise
		@throws InvalidDataException If the group is null
		@since 2.6
	*/
	public boolean hasChanged(Group group)
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		int current = IntegerUtil.getInt(groupPermissions.get(group.getData()));
		int original = originalPermissions == null ?
			0 : IntegerUtil.getInt(originalPermissions.get(group.getData()));
		return current != original;
	}	
	
	/**
		Get the number of entries.
	*/
	public int size()
	{
		return groupPermissions.size();
	}

	/**
		Get the internal representation of the permissions.
	*/
	final Map<GroupData,Integer> getGroupPermissions()
	{
		return groupPermissions;
	}
	
	/**
		Get the internal representation of the original permissions.
		@since 2.6
	*/
	final Map<GroupData,Integer> getOriginalPermissions()
	{
		return originalPermissions;
	}

}
