/**
	$Id: HqlQueryJoiners.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.AutoJoiner;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.HqlQuery;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.ReporterListExpression;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;

/**
	This class collects <code>AutoJoiner</code>:s for <code>HqlQuery</code>
	queries. The implementations are thread-safe and can be re-used between threads.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
*/
public final class HqlQueryJoiners
{

	/**
		This instance joins reporter lists with a left join, unless the
		{@link ReporterListExpression#getAutoJoinType()} method overrides 
		the default.
	*/
	public static final ReporterListJoiner REPORTER_LIST_LEFT = 
		new ReporterListJoiner(JoinType.LEFT);
	
	/**
		This instance joins reporter lists with an inner join, unless the
		{@link ReporterListExpression#getAutoJoinType()} method overrides 
		the default.
	*/
	public static final ReporterListJoiner REPORTER_LIST_INNER = 
		new ReporterListJoiner(JoinType.INNER);
	
	/**
		Get a joiner instance for raw data for the specified join type.
		@param joinType The type of join
		@return {@link #REPORTER_LIST_LEFT} or {@link #REPORTER_LIST_INNER}
	 */
	public static ReporterListJoiner getReporterListJoiner(JoinType joinType)
	{
		return joinType == JoinType.LEFT ? REPORTER_LIST_LEFT : REPORTER_LIST_INNER;
	}

	/**
		Joiner for reporter lists.
		@author nicklas
		@version 2.8
		@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
	*/
	private static class ReporterListJoiner
		implements AutoJoiner<HqlQuery, ReporterListExpression>
	{
		private final JoinType joinType;
	
		private ReporterListJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		public Class<HqlQuery> getQueryClass()
		{
			return HqlQuery.class;
		}
		public Class<ReporterListExpression> getElementClass()
		{
			return ReporterListExpression.class;
		}
		public void joinIfNeeded(HqlQuery query, ReporterListExpression element)
		{
			JoinType jt = element.getAutoJoinType();
			String tableAlias = RealTable.REPORTER_LIST_SCORES.getAlias(element.getReporterList());
			Restriction on = Restrictions.eq(
					Hql.property(tableAlias, "reporterList.id"), 
					Expressions.integer(element.getReporterList().getId()
				)
			);
			if (jt == JoinType.LEFT)
			{
				query.join(Hql.leftJoin(element.getJoinAlias(), element.getJoinProperty(), 
					tableAlias, on, false));
			}
			else
			{
				query.join(Hql.innerJoin(element.getJoinAlias(), element.getJoinProperty(), 
					tableAlias, on, false));
			}
		}
		// -------------------------------------------
	}

}
