/*
	$Id: Install.java 5267 2010-03-12 07:59:58Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2008 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ColoringData;
import net.sf.basedb.core.data.DirectoryData;
import net.sf.basedb.core.data.DataFileTypeData;
import net.sf.basedb.core.data.FileTypeData;
import net.sf.basedb.core.data.FormulaData;
import net.sf.basedb.core.data.HardwareData;
import net.sf.basedb.core.data.HardwareTypeData;
import net.sf.basedb.core.data.MimeTypeData;
import net.sf.basedb.core.data.PlatformData;
import net.sf.basedb.core.data.PlatformFileTypeData;
import net.sf.basedb.core.data.PluginConfigurationData;
import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.data.ProtocolTypeData;
import net.sf.basedb.core.data.QuantityData;
import net.sf.basedb.core.data.QuotaData;
import net.sf.basedb.core.data.QuotaTypeData;
import net.sf.basedb.core.data.QuotaIndex;
import net.sf.basedb.core.data.RoleData;
import net.sf.basedb.core.data.SchemaVersionData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.SoftwareTypeData;
import net.sf.basedb.core.data.SoftwareData;
import net.sf.basedb.core.data.UnitData;
import net.sf.basedb.core.data.UnitSymbolData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.GroupData;
import net.sf.basedb.core.data.ClientData;
import net.sf.basedb.core.data.ItemKeyData;
import net.sf.basedb.core.data.RoleKeyData;
import net.sf.basedb.core.data.NewsData;
import net.sf.basedb.core.data.LabelData;
import net.sf.basedb.core.data.AnnotationTypeData;
import net.sf.basedb.core.data.ParameterValueData;
import net.sf.basedb.core.data.PlateGeometryData;
import net.sf.basedb.core.data.PlateMappingData;
import net.sf.basedb.core.data.MappingCoordinate;
import net.sf.basedb.core.data.PluginTypeData;

import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.DbControl;

import net.sf.basedb.util.MD5;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.XMLUtil;

import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.TimerTask;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;

/**
	This class contains static methods used to install items into an
	empty database.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2010-03-12 08:59:58 +0100 (Fri, 12 Mar 2010) $
*/
public final class Install
{
	/**
		Log core events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.Install");

	private static org.hibernate.Session session = null;
	
	private static SchemaVersionData schemaVersion = null;

	private static SessionControl sessionControl = null;

	/**
		The schema version to give to new installations. Already
		existing installations will be updated to this version number
		by the {@link Update#updateDatabase(ProgressReporter, String, String)}
		method.
	*/
	public static final int NEW_SCHEMA_VERSION = Integer.valueOf(79).intValue();
	
	public static synchronized void createTables(boolean update, final ProgressReporter progress)
		throws BaseException
	{
		String message = "";
		TimerTask timer = null;
		if (progress != null) 
		{
			progress.display(0, "Building database...");
			// The timer will append a dot to the progress reporter every second
			timer = new TimerTask()
				{
					public void run()
					{
						progress.append(".");
					}
				};
		}
		try
		{
			if (timer != null) Application.getCoreTimer().schedule(timer, 1000, 1000);
			Application.start(true, false, false);
			HibernateUtil.createStaticTables(update);
			message = "Database built successfully.";
		}
		catch (BaseException ex)
		{
			message = "Building database failed: " + ex.getMessage()+"\n";
			throw ex;
		}
		finally
		{
			if (timer != null) 
			{
				timer.cancel();
				timer = null;
			}
			if (progress != null)
			{
				progress.display(100, message);
			}
			Application.stop();
		}
	}
	
	/**
		Initialise the BASE database by creating all items that are
		required for BASE to work properly. Items that already exists
		will not be modified.
	 	
	 	@param update TRUE if the run is an update, FALSE otherwise. 
		@param progress An object implementing the {@link ProgressReporter}
			interface
	 	@param rootLogin The root user's login name.
	 	@param rootPassword The root user's password.
		@throws BaseException This exception is thrown if there is an error
	*/
	public static synchronized void initDatabase(boolean update, ProgressReporter progress, String rootLogin, String rootPassword)
		throws BaseException
	{
		final int total_progress_steps = 26;
		final float progress_factor = 100 / total_progress_steps;
		if (progress != null) progress.display(0, "Initialising database...");
		try
		{
			Application.start(false, false, false);
			session = HibernateUtil.newSession();
	
			// SchemaVersion
			schemaVersion = createSchemaVersion(update ? 1 : NEW_SCHEMA_VERSION);
	
			// QuotaTypes
			if (progress != null) progress.display((int)(1*progress_factor), "--Creating quota types...");
			createQuotaType(QuotaType.TOTAL, "Total", "Total quota", true);
			createQuotaType(QuotaType.FILE, "Files", "Quota for files", true);
			createQuotaType(QuotaType.RAW_DATA, "Raw data", "Quota for raw data", false);
			createQuotaType(QuotaType.EXPERIMENT, "Experiment", "Quota for experiment", false);
	
			// Quota
			if (progress != null) progress.display((int)(2*progress_factor), "--Creating quota...");
			QuotaData quotaDefault = createQuota(Quota.DEFAULT, "Default quota", "This quota is assigned to new users unless nothing else is specified.", 10*1024*1024);
			QuotaData quotaUnlimit = createQuota(Quota.UNRESTRICTED, "Unlimited quota", "This quota gives a user unlimited disk space.", (int)Quota.UNLIMITED);
			QuotaData quota100 = createQuota(null, "100 MB total", "100 MB total quota.", 100*1024*1024);
			QuotaData quota200 = createQuota(null, "200 MB total", "200 MB total quota.", 200*1024*1024);
			QuotaData quota500 = createQuota(null, "500 MB total", "500 MB total quota.", 500*1024*1024);
			QuotaData quota1000 = createQuota(null, "1 GB total", "1 GB total quota.", 1024*1024*1024);
			QuotaData quota0 = createQuota(null, "No quota", "May not consume any disk space at all", 0);
	
			// Roles
			if (progress != null) progress.display((int)(3*progress_factor), "--Creating roles...");
			RoleData roleAdmin = createRole(Role.ADMINISTRATOR, "Administrator", false, "Users with this role get full permission to BASE.");
			RoleData roleSuper = createRole(Role.SUPERVISOR, "Supervisor", false, "Users with this role get full READ permission to everything in BASE.");
			RoleData rolePower = createRole(Role.POWER_USER, "Power user", false, "Users with this role may do some things that an ordinary user may not.");
			RoleData roleUser = createRole(Role.USER, "User", true, "This permissions in this role should be sufficient for most regular users.");
			RoleData roleGuest = createRole(Role.GUEST, "Guest", false, "For users wanting to have a peek at BASE. Has very limited access to create new things.");
			RoleData roleJobAgent = createRole(Role.JOBAGENT, "Job agent", false, "This role is given to job agents and allows them to read and execute jobs.");
	
			// Users
			if (progress != null) progress.display((int)(4*progress_factor), "--Creating users...");
			UserData rootUser = createUser(User.ROOT, rootLogin, rootPassword, "Root", 
				"This is the root user account of BASE. It has full permission to everything.", 
				roleAdmin, quotaUnlimit, false, false);
			UserData jobAgentUser = createUser(null, "jobagent", "", "Job agent", 
				"This user account is used by the job agents to login and execute jobs. You must "+
				"activate it and set a password before it can be used.",
				roleJobAgent, quota0, true, true);
			// createUser(0, "demo", "demo", "Demo", "This account can be used for demonstration purposes.", Role.GUEST);
			// createUser(0, "power", "power", "Power user", "This account has power user privileges.", Role.POWER_USER);
	
			// Now that we have a root user let's create a session 
			sessionControl = Application.newSessionControl( null, null, null );
			sessionControl.login(rootLogin, rootPassword, "InitDBSessionId", false);
	
			if (progress != null) progress.display((int)(5*progress_factor), "--Creating groups...");
			GroupData groupEveryone = createGroup(Group.EVERYONE, "Everyone", false, true, "Everyone are members of this group.");
	
			if (progress != null) progress.display((int)(6*progress_factor), "--Creating keys...");
			int PERMISSION_ALL = Permission.grant(Permission.READ, Permission.USE, Permission.WRITE, 
				Permission.DELETE, Permission.CREATE, Permission.SET_OWNER, Permission.SET_PERMISSION);
			int PERMISSION_DELETE = Permission.grant(Permission.DELETE);
			int PERMISSION_WRITE = Permission.grant(Permission.WRITE);
			int PERMISSION_USE = Permission.grant(Permission.USE);
			int PERMISSION_READ = Permission.grant(Permission.READ);
			int PERMISSION_CREATE = Permission.grant(Permission.CREATE);
			
			// Administrators -> ALL
			HashMap<RoleData, Integer> administrators_all = new HashMap<RoleData, Integer>();
			administrators_all.put(roleAdmin, PERMISSION_ALL);
			administrators_all.put(roleSuper, PERMISSION_READ);
	
			// Administrators -> WRITE
			HashMap<RoleData, Integer> administrators_write = new HashMap<RoleData, Integer>();
			administrators_write.put(roleAdmin, PERMISSION_WRITE);
			administrators_write.put(roleSuper, PERMISSION_READ);
	
			// Administrators -> READ
			HashMap<RoleData, Integer> administrators_read = new HashMap<RoleData, Integer>();
			administrators_read.put(roleAdmin, PERMISSION_READ);
			administrators_read.put(roleSuper, PERMISSION_READ);
	
			// Power users -> CREATE; Administrators -> ALL
			HashMap<RoleData, Integer> power_users_create = new HashMap<RoleData, Integer>();
			power_users_create.put(roleAdmin, PERMISSION_ALL);
			power_users_create.put(rolePower, PERMISSION_CREATE);
			power_users_create.put(roleSuper, PERMISSION_READ);
	
			// Power users & Users -> CREATE; Administrators -> ALL
			HashMap<RoleData, Integer> users_create = new HashMap<RoleData, Integer>();
			users_create.put(roleAdmin, PERMISSION_ALL);
			users_create.put(rolePower, PERMISSION_CREATE);
			users_create.put(roleUser, PERMISSION_CREATE);
			users_create.put(roleSuper, PERMISSION_READ);
	
			// Power users & Users & Guests -> CREATE; Administrators -> ALL
			HashMap<RoleData, Integer> guests_create = new HashMap<RoleData, Integer>();
			guests_create.put(roleAdmin, PERMISSION_ALL);
			guests_create.put(rolePower, PERMISSION_CREATE);
			guests_create.put(roleUser, PERMISSION_CREATE);
			guests_create.put(roleGuest, PERMISSION_CREATE);
			guests_create.put(roleSuper, PERMISSION_READ);
	
			// All -> CREATE; Administrators -> ALL
			HashMap<RoleData, Integer> all_create = new HashMap<RoleData, Integer>();
			all_create.put(roleAdmin, PERMISSION_ALL);
			all_create.put(rolePower, PERMISSION_CREATE);
			all_create.put(roleUser, PERMISSION_CREATE);
			all_create.put(roleGuest, PERMISSION_CREATE);
			all_create.put(roleSuper, PERMISSION_CREATE);
			
			// All -> READ; Administrators --> WRITE
			HashMap<RoleData, Integer> all_use_administrators_write = new HashMap<RoleData, Integer>();
			all_use_administrators_write.put(roleAdmin, PERMISSION_WRITE);
			all_use_administrators_write.put(rolePower, PERMISSION_USE);
			all_use_administrators_write.put(roleUser, PERMISSION_USE);
			all_use_administrators_write.put(roleGuest, PERMISSION_USE);
			all_use_administrators_write.put(roleSuper, PERMISSION_USE);
	
			// Users & Guests & Powers users -> USE; Administrators -> ALL
			HashMap<RoleData, Integer> guests_use_administrators_all = new HashMap<RoleData, Integer>();
			guests_use_administrators_all.put(roleAdmin, PERMISSION_ALL);
			guests_use_administrators_all.put(rolePower, PERMISSION_USE);
			guests_use_administrators_all.put(roleUser, PERMISSION_USE);
			guests_use_administrators_all.put(roleGuest, PERMISSION_USE);
			guests_use_administrators_all.put(roleSuper, PERMISSION_READ);
	
			// Users & Guests -> USE; Powers users & Administrators -> ALL
			HashMap<RoleData, Integer> guests_use_power_users_all = new HashMap<RoleData, Integer>();
			guests_use_power_users_all.put(roleAdmin, PERMISSION_ALL);
			guests_use_power_users_all.put(rolePower, PERMISSION_ALL);
			guests_use_power_users_all.put(roleUser, PERMISSION_USE);
			guests_use_power_users_all.put(roleGuest, PERMISSION_USE);
			guests_use_power_users_all.put(roleSuper, PERMISSION_READ);
			
			HashMap<RoleData, Integer> systemPermissions = new HashMap<RoleData, Integer>();
			systemPermissions.put(roleAdmin, Permission.grant(Permission.SHARE_TO_EVERYONE, Permission.SELECT_JOBAGENT));
			systemPermissions.put(rolePower, Permission.grant(Permission.SHARE_TO_EVERYONE, Permission.SELECT_JOBAGENT));
			
			// System permissions
			RoleKeyData systemKey = createRoleKey(Item.SYSTEM, "System", "System-level permissions", systemPermissions);
			
			// User, group, role and project management
			createRoleKey(Item.GROUP, "Groups", "Gives access to groups.", administrators_all);
			createRoleKey(Item.ROLE, "Roles", "Gives access to roles", administrators_all);
			createRoleKey(Item.USER, "Users", "Gives access to users.", administrators_all);
			createRoleKey(Item.PROJECT, "Projects", "Gives access to projects.", users_create);
	
			// Files, directories and quota
			createRoleKey(Item.DIRECTORY, "Directories", "Gives access to directories", guests_create);
			createRoleKey(Item.FILETYPE, "File types", "Gives access to file types", all_use_administrators_write);
			createRoleKey(Item.FILE, "Files", "Gives access to files", guests_create);
			createRoleKey(Item.QUOTATYPE, "Quota types", "Gives access to quota types", all_use_administrators_write);
			createRoleKey(Item.QUOTA, "Quota", "Gives access to quota", administrators_all);
			createRoleKey(Item.DISKUSAGE, "Disk usage", "Gives access disk usage information", administrators_read);
	
			// Protocols, hardware, software
			createRoleKey(Item.PROTOCOLTYPE, "Protocol types", "Gives access to protocol types", guests_use_power_users_all);
			createRoleKey(Item.PROTOCOL, "Protocols", "Gives access to protocols", power_users_create);
			createRoleKey(Item.HARDWARETYPE, "Hardware types", "Gives access to hardware types", guests_use_power_users_all);
			createRoleKey(Item.HARDWARE, "Hardware", "Gives access to hardware", power_users_create);
			createRoleKey(Item.SOFTWARETYPE, "Software types", "Gives access to software types", all_use_administrators_write);
			createRoleKey(Item.SOFTWARE, "Software", "Gives access to software", power_users_create);
			
			// Platforms, platform variants, platform file types
			createRoleKey(Item.PLATFORM, "Platforms", "Gives access to platforms", guests_use_administrators_all);
			createRoleKey(Item.DATAFILETYPE, "Data file types", "Gives access to data file types", guests_use_administrators_all);

			// Annotations
			createRoleKey(Item.ANNOTATIONTYPE, "Annotation types", "Gives access to annotation types", power_users_create);
			createRoleKey(Item.ANNOTATIONTYPECATEGORY, "Annotation type categories", "Gives access to annotation type categories", power_users_create);
			createRoleKey(Item.QUANTITY, "Quantities", "Gives access to quantities", guests_use_administrators_all);
			createRoleKey(Item.UNIT, "Units", "Gives access to units", guests_use_administrators_all);
			
			// BioMaterials
			createRoleKey(Item.LABEL, "Labels", "Gives access to labels", power_users_create);
			createRoleKey(Item.BIOSOURCE, "BioSource", "Gives access to biosources", users_create);
			createRoleKey(Item.SAMPLE, "Samples", "Gives access to samples", users_create);
			createRoleKey(Item.EXTRACT, "Extracts", "Gives access to extracts", users_create);
			createRoleKey(Item.LABELEDEXTRACT, "Labeled extracts", "Gives access to labeled extracts", users_create);
			createRoleKey(Item.BIOPLATE, "Bio plates", "Gives access to bio plates", users_create);
			createRoleKey(Item.BIOMATERIALLIST, "Biomaterial lists", "Gives access to biomaterial lists", users_create);
	
			// Reporters
			createRoleKey(Item.REPORTERTYPE, "Reporter types", "Gives access to reporter types", guests_use_administrators_all);
			createRoleKey(Item.REPORTER, "Reporters", "Gives access to reporter", guests_use_administrators_all);
			createRoleKey(Item.REPORTERLIST, "Reporter lists", "Gives access to reporter lists", users_create);
	
			// Array LIMS - plates
			createRoleKey(Item.PLATEGEOMETRY, "Plate geometries", "Gives access to plate geometries", guests_use_administrators_all);
			createRoleKey(Item.PLATETYPE, "Plate types", "Gives access to plate types", power_users_create);
			createRoleKey(Item.PLATE, "Plates", "Gives access to plates", power_users_create);
			createRoleKey(Item.PLATEMAPPING, "Plate mappings", "Gives access to plate mappings", power_users_create);
	
			// Array LIMS - arrays
			createRoleKey(Item.ARRAYDESIGN, "Array design", "Gives access to array designs", power_users_create);
			createRoleKey(Item.ARRAYBATCH, "Array batches", "Gives access to array batches", power_users_create);
			createRoleKey(Item.ARRAYSLIDE, "Array slides", "Gives access to array slides", power_users_create);
	
			// Hybridization, raw data and experiments
			createRoleKey(Item.HYBRIDIZATION, "Hybridizations", "Gives access to hybridizations", users_create);
			createRoleKey(Item.SCAN, "Scans", "Gives access to scan", users_create);
			createRoleKey(Item.RAWBIOASSAY, "Raw bioassays", "Gives access to raw bioassays", users_create);
			createRoleKey(Item.EXPERIMENT, "Experiments", "Gives access to experiments", users_create);
			createRoleKey(Item.FORMULA, "Formulas", "Gives access to formulas", users_create);
			createRoleKey(Item.EXTRAVALUETYPE, "Extra value types", "Gives access to extra value types", guests_use_power_users_all);
	
			// Plugins, jobs
			createRoleKey(Item.PLUGINTYPE, "Plugin types", "Gives access to plugin types", administrators_all);
			createRoleKey(Item.PLUGINDEFINITION, "Plugin definitions", "Gives access to plugin definitions", administrators_all);
			createRoleKey(Item.PLUGINCONFIGURATION, "Plugin configurations", "Gives access to plugin configurations", power_users_create);
			createRoleKey(Item.JOB, "Jobs", "Gives access to jobs", guests_create);
			createRoleKey(Item.JOBAGENT, "Job agents", "Gives access to job agents", administrators_all);
	
			// Misc.
			createRoleKey(Item.CLIENT, "Client applications", "Gives access to client applications", administrators_all);
			createRoleKey(Item.NEWS, "News", "Gives access to news", administrators_all);
			createRoleKey(Item.MIMETYPE, "Mime types", "Gives access to mime types", administrators_all);
			createRoleKey(Item.SESSION, "Sessions", "Gives access to sessions", administrators_read);
			createRoleKey(Item.SETTING, "Settings", "Gives access to settings", administrators_all);
			createRoleKey(Item.MESSAGE, "Messages", "Gives access to messages", users_create);
			createRoleKey(Item.CHANGEHISTORY, "Change history", "Gives access to change history of items", administrators_read);
			
			// Permissions for job agents are only added for new installations and 
			// updates below schema version 13
			if (!update || schemaVersion.getSchemaVersion() < 13)
			{
				Map<Item, Integer> agentPermissions = new HashMap<Item, Integer>();
				agentPermissions.put(Item.JOB, PERMISSION_READ);
				agentPermissions.put(Item.JOBAGENT, PERMISSION_READ | PERMISSION_CREATE);
				agentPermissions.put(Item.MESSAGE, PERMISSION_CREATE);
				agentPermissions.put(Item.SYSTEM, Permission.grant(Permission.ACT_AS_ANOTHER_USER, Permission.SELECT_JOBAGENT));
				addRolePermissions(roleJobAgent, agentPermissions);
			}
			
			Map<GroupData, Integer> everyoneRead = new HashMap<GroupData, Integer>();
			everyoneRead.put(groupEveryone, PERMISSION_READ);
			
			Map<GroupData, Integer> everyoneUse = new HashMap<GroupData, Integer>();
			everyoneUse.put(groupEveryone, PERMISSION_USE);
			
			Map<UserData, Integer> jobAgentUse = new HashMap<UserData, Integer>();
			jobAgentUse.put(jobAgentUser, PERMISSION_USE);
			
			ItemKeyData keyEveryoneRead = createItemKey("EVERYONE=READ", null, everyoneRead);
			ItemKeyData keyEveryoneUse = createItemKey("EVERYONE=USE", null, everyoneUse);
			ItemKeyData keyJobAgentUse = createItemKey("JOBAGENT=USE", jobAgentUse, null);
	
			// Quantities and units
			if (progress != null) progress.display((int)(7*progress_factor), "--Creating units...");
			installUnits("/units.xml", update);
			
			// ProtocolTypes
			if (progress != null) progress.display((int)(8*progress_factor), "--Creating protocol types...");
			createProtocolType(ProtocolType.SAMPLING, "Sampling", "Protocols used for creating samples.");
			createProtocolType(ProtocolType.EXTRACTION, "Extraction", "Protocols used for creating extracts.");
			createProtocolType(ProtocolType.LABELING, "Labeling", "Protocols used for creating labeled extracts.");
			createProtocolType(ProtocolType.POOLING, "Pooling", "Protocols used for creating pooled biomaterails.");
			createProtocolType(ProtocolType.HYBRIDIZATION, "Hybridization", "Protocols used for creating hybridizations.");
			createProtocolType(ProtocolType.PRINTING, "Printing", "Protocols used for printing a bathc of array slides.");
			createProtocolType(ProtocolType.SCANNING, "Scanning", "Protocols used for scanning microarray images.");
			createProtocolType(ProtocolType.FEATURE_EXTRACTION, "Feature extraction", "Protocols used for extracting features from microarray images.");
	
			// HardwareTypes
			if (progress != null) progress.display((int)(9*progress_factor), "--Creating hardware types...");
			HardwareTypeData scanner = createHardwareType(HardwareType.SCANNER, "Scanner", "Scans the microarray images.");
			HardwareTypeData robot = createHardwareType(HardwareType.PRINT_ROBOT, "Print robot", "Spot the probe on the array.");
			HardwareTypeData hybStation = createHardwareType(HardwareType.HYBRIDIZATION_STATION, "Hybridization station", "Automate the hybridization of microarrays on slides.");
			HardwareTypeData freezer = createHardwareType(HardwareType.FREEZER, "Freezer", "Storage area for bioplates");
	
			// Hardware 
			if (progress != null) progress.display((int)(10*progress_factor), "--Creating hardware...");
			// Scanners
			createHardware("Agilent DNA Microarray Scanner", "G2565BA", null, scanner, rootUser, keyEveryoneUse);
			createHardware("GenePix Scanner", "4200A", null, scanner, rootUser, keyEveryoneUse);
			createHardware("ScanArray", null, null, scanner, rootUser, keyEveryoneUse);
			createHardware("GeneTAC", null, null, scanner, rootUser, keyEveryoneUse);
			createHardware("Affymetrix GeneChip Scanner", "3000", null, scanner, rootUser, keyEveryoneUse);
			// Print robots
			createHardware("BioRobotics", "MicroGrid II", "", robot, rootUser, keyEveryoneUse);
			createHardware("OmniGrid", "100", "", robot, rootUser, keyEveryoneUse);
			createHardware("Qarray", "Max", "", robot, rootUser, keyEveryoneUse);
	
			// SoftwareTypes
			if (progress != null) progress.display((int)(11*progress_factor), "--Creating software types...");
			SoftwareTypeData feature = createSoftwareType(SoftwareType.FEATURE_EXTRACTION, "Feature extraction", "Extracts features from microarray images.");
	
			// Software 
			if (progress != null) progress.display((int)(12*progress_factor), "--Creating software...");
			createSoftware("BZScan", null, null, feature, rootUser, keyEveryoneUse);
			createSoftware("AIDA", null, null, feature, rootUser, keyEveryoneUse);
			createSoftware("GenePix Pro", "6.0", null, feature, rootUser, keyEveryoneUse);
			createSoftware("ImaGene", null, null, feature, rootUser, keyEveryoneUse);
			createSoftware("TIGR Spotfinder", null, null, feature, rootUser, keyEveryoneUse);
			createSoftware("Agilent Feature Extraction Software", "8.5", null, feature, rootUser, keyEveryoneUse);
			createSoftware("ChipSkipper", null, null, feature, rootUser, keyEveryoneUse);
			createSoftware("GeneTAC", null, null, feature, rootUser, keyEveryoneUse);
			createSoftware("ScanArray", null, null, feature, rootUser, keyEveryoneUse);
			createSoftware("Affymetrix GeneChip Software", "1.2", null, feature, rootUser, keyEveryoneUse);
			createSoftware("Illumina BeadStudio", null, null, feature, rootUser, keyEveryoneUse);
	
			// Directory
			if (progress != null) progress.display((int)(13*progress_factor), "--Creating directories...");
			DirectoryData rootDirectory = createDirectory(rootUser, null, Directory.ROOT, "root", "This is the root directory", keyEveryoneRead, false);
			DirectoryData homeDirectory = createDirectory(rootUser, rootDirectory, Directory.HOME, "home", "This is the directory where the user's home directories are located", keyEveryoneRead, false);
			DirectoryData templateDirectory = createDirectory(rootUser, rootDirectory, Directory.TEMPLATE, "template", "The substructure of this directory is used to create the directory structure for new home directories", null, false);
			createDirectory(rootUser, templateDirectory, null, "images", "Place your images here", null, false);
			createDirectory(rootUser, templateDirectory, null, "raw data", "Place your raw data files here", null, true);
			createDirectory(rootUser, templateDirectory, null, "projects", "Create a subdirectory for project-related files in this directory.", null, false);
	
			// FileTypes
			if (progress != null) progress.display((int)(14*progress_factor), "--Creating file types...");
			createFileType(FileType.PROTOCOL, "Protocol", "Protocol files");
			FileTypeData rawDataType = createFileType(FileType.RAW_DATA, "Raw data", "Raw data files");
			createFileType(FileType.PLATE, "Plate", "Plate files");
			createFileType(FileType.PLATE_MAPPING, "Plate mapping", "Plate mapping files");
			FileTypeData printMapType = createFileType(FileType.PRINT_MAP, "Print map", "Print map files");
			FileTypeData reporterMapType = createFileType(FileType.REPORTER_MAP, "Reporter map", "Reporter map files");
			createFileType(FileType.REPORTER, "Reporter", "Reporter files");
			FileTypeData imageType = createFileType(FileType.IMAGE, "Image", "Image files");
			createFileType(FileType.SPOT_IMAGES, "Spot images", "A zip file containing generated spot images in png format.");
			FileTypeData spotDataType = createFileType(FileType.SPOT_DATA, "Spot data", "Analysed bioassay set data");
			FileTypeData extraValueType = createFileType(FileType.EXTRAVALUE_DATA, "Extra value data", "Analysed extra values");
	
			// MimeTypes
			if (progress != null) progress.display((int)(15*progress_factor), "--Creating MIME types...");
			createMimeType("text/css", "Cascading style sheet", "css", null, false);
			createMimeType("text/plain", "Text files", "txt", null, false);
			createMimeType("text/html", "Hypertext markup language files", "htm", null, false);
			createMimeType("text/html", "Hypertext markup language files", "html", null, false);
			createMimeType("image/jpeg", "JPEG image", "jpg", imageType, false);
			createMimeType("image/jpeg", "JPEG image","jpeg", imageType, false);
			createMimeType("image/jpeg", "JPEG image", "jpe", imageType, false);
			createMimeType("image/gif", "GIF image", "gif", imageType, false);
			createMimeType("image/png", "PNG image", "png", imageType, false);
			createMimeType("image/tiff", "TIFF image", "tiff", imageType, true);
			createMimeType("image/tiff", "TIFF image", "tif", imageType, true);
			createMimeType("image/rgb", "RGB image", "rgb", imageType, true);
			createMimeType("image/x-pict", "Macintosh PICT format", "pict", imageType, false);
			createMimeType("image/x-ms-bmp", "Microsoft Windows bitmap", "bmp", imageType, true);
			createMimeType("application/postscript", "PostScript", "ai", null, false);
			createMimeType("application/postscript", "PostScript","eps", null, false);
			createMimeType("application/postscript", "PostScript", "ps", null, false);
			createMimeType("application/pdf", "Adobe Acrobat PDF", "pdf", null, false);
			createMimeType("application/x-latex", "LaTeX document","latex", null, false);
			createMimeType("application/msword", "MS word document", "doc", null, true);
			createMimeType("application/rtf", "Microsoft Rich Text Format", "rtf", null, true);
			createMimeType("application/octet-stream", "PC executable", "exe", null, false);
			createMimeType("application/x-perl", "Perl program", "pl", null, false);
			createMimeType("application/xml", "Extended markup language", "xml", null, false);
			createMimeType("application/xml-dtd", "Document type definition for XML", "dtd", null, false);
			createMimeType("application/vnd.ms-powerpoint", "PowerPoint (Microsoft)", "ppt", null, true);
			createMimeType("application/vnd.ms-powerpoint", "PowerPoint (Microsoft)", "ppz", null, true);
			createMimeType("application/vnd.ms-powerpoint", "PowerPoint (Microsoft)", "pps", null, true);
			createMimeType("application/vnd.ms-powerpoint", "PowerPoint (Microsoft)", "pot", null, true);
			createMimeType("application/vnd.ms-excel", "Excel spreadsheet (Microsoft)", "xls", null, true);
			createMimeType("application/vnd.ms-excel", "Excel spreadsheet (Microsoft)", "xlc", null, true);
			createMimeType("application/vnd.ms-excel", "Excel spreadsheet (Microsoft)", "xll", null, true);
			createMimeType("application/vnd.ms-excel", "Excel spreadsheet (Microsoft)", "xlm", null, true);
			createMimeType("application/vnd.ms-excel", "Excel spreadsheet (Microsoft)", "xlw", null, true);
			createMimeType("application/x-compress", "Unix compressed file", "z", null, false);
			createMimeType("application/zip", "DOS/PC - Pkzipped archive", "zip", null, false);
			createMimeType("application/x-compressed", "Gzip compressed archive", "tgz", null, false);
			createMimeType("application/x-gtar", "Gzipped archive", "gtar", null, false);
			createMimeType("application/x-gzip", "Gzip compressed file", "gz", null, false);
	
			// Microarray/BASE-specific mime types
			createMimeType("text/plain", "GenePix Results file", "gpr", rawDataType, true);
			createMimeType("text/plain", "GenePix Array List file", "gal", reporterMapType, true);
			createMimeType("text/plain", "Spot file format", "spot", rawDataType, true);
			createMimeType("text/plain", "Spotfinder file format", "tav", rawDataType, true);
			createMimeType("text/plain", "Affymetrix CEL file", "cel", rawDataType, true);
			createMimeType("text/plain", "Affymetrix CDF file", "cdf", reporterMapType, true);
			createMimeType("text/plain", "FASTA formatted file", "fasta", null, true);
			createMimeType("text/plain", "BASEfile formatted file", "base", null, true);
			createMimeType("text/plain", "Biorobotics TAM format", "tam", printMapType, true);
			createMimeType("text/plain", "Molecularware MWBR format", "mvbr", printMapType, true);
	
			// Plate geometries
			if (progress != null) progress.display((int)(16*progress_factor), "--Creating plate geometries...");
			PlateGeometryData plate96 = createPlateGeometry("96-well (8 x 12)", "96-wells, 8 rows, 12 columns", 8, 12);
			PlateGeometryData plate384 = createPlateGeometry("384-well (16 x 24)", "384-wells, 16 rows, 24 columns", 16, 24);
			PlateGeometryData plate1536 = createPlateGeometry("1536-well (32 x 48)", "1536-wells, 32 rows, 48 columns", 32, 48);
	
			// Plate mappings
			if (progress != null) progress.display((int)(17*progress_factor), "--Creating plate mappings...");
			createPlateMapping("96 --> 96", "Maps 1:1 between 96-well (8 x 12) plates",
				plate96, 1, plate96, 1, null, rootUser, keyEveryoneUse);
			createPlateMapping("96 --> 384", "Maps 4:1 between 96-well (8 x 12) plates and 384 (16 x 24) plate",
				plate96, 4, plate384, 1, "pm1243.gif", rootUser, keyEveryoneUse);
			createPlateMapping("384 --> 384", "Maps 1:1 between 384-well (16 x 24) plates",
				plate384, 1, plate384, 1, null, rootUser, keyEveryoneUse);
			createPlateMapping("384 --> 1536", "Maps 4:1 between 384-well (16 x 24) plates and 1536 (32 x 48) plate",
				plate384, 4, plate1536, 1, "pm1243.gif", rootUser, keyEveryoneUse);
			createPlateMapping("1536 --> 1536", "Maps 1:1 between 1536-well (16 x 24) plates",
				plate1536, 1, plate1536, 1, null, rootUser, keyEveryoneUse);
	
	
			// Labels
			if (progress != null) progress.display((int)(18*progress_factor), "--Creating labels...");
			createLabel("cy3", "", rootUser, keyEveryoneUse);
			createLabel("cy5", "", rootUser, keyEveryoneUse);
			createLabel("Biotin", "", rootUser, keyEveryoneUse);
			
			// Annotation types
			Item[] wells = new Item[] { Item.WELL };
			Map<String, String> useRadio = Collections.singletonMap("displayAsList", "0");
			if (progress != null) progress.display((int)(19*progress_factor), "--Creating annotation types...");
			createAnnotationType("Sequence verified", "", Type.STRING, 1, new String[] { "Not done", "Good", "Bad", "Failed" }, 
				wells, useRadio, false, false, rootUser, keyEveryoneUse );
			createAnnotationType("Bands", "", Type.STRING, 1, new String[] { "Not done", "Good", "Missing", "Multiple" }, 
				wells, useRadio, false, false, rootUser, keyEveryoneUse );
			createAnnotationType("Bacterial growth", "", Type.STRING, 1, new String[] { "Not done", "Yes", "No" }, 
				wells, useRadio, false, false, rootUser, keyEveryoneUse );
			createAnnotationType("Comment", "", Type.STRING, 1, null, 
				wells, null, false, false, rootUser, keyEveryoneUse );
	
			// Clients
			if (progress != null) progress.display((int)(20*progress_factor), "--Creating clients...");
			createClient(rootUser, "net.sf.basedb.clients.test", "Test programs", "This client is used by test programs", keyEveryoneUse);
			createClient(rootUser, "net.sf.basedb.clients.migration", "Migration tool", "This client is used to migrate BASE 1.2.x data to BASE 2", null);
			createClient(rootUser, "net.sf.basedb.clients.jobagent", "Job agent client", "This client is used by all job agents", keyJobAgentUse);
			createClient(rootUser, "net.sf.basedb.clients.internal", "Internal", 
				"This client is used by the internal processes such as the job queue and secondary " +
				"storage controller. Do not remove!", keyEveryoneRead);
	
			// Plugins
			if (progress != null) progress.display((int)(21*progress_factor), "--Creating plugin definitions...");
			createPluginType("Autodetecting importer", "This interface should be implemented by all plugins that are capable of autodetect the file format by parsing an input stream.", 
				"net.sf.basedb.core.plugin.AutoDetectingImporter", null);
			createPluginType("Configure by example", "This interface should be implemented by all plugins that use the FlatFileParser and " +
				"wants to use an uploaded file to generate the regular expressions for parsing it.", 
				"net.sf.basedb.util.parser.ConfigureByExample", null);
			createPluginType("Configure by example (wrapped)", "This is an extension to the 'Configure by example' interface " +
					"that allows plug-ins to wrap the file with a special input stream when reading from it. " +
					"It makes it possible, for example, to parse compressed files on the fly.", 
					"net.sf.basedb.util.parser.WrappedConfigureByExample", null);
			createPluginType("File unpacker", "This interface should be implemented by all plugins that can unpack compressed files. " +
				"The file upload feature uses these plugins for automatic unpacking of compressed files.",
				"net.sf.basedb.util.zip.FileUnpacker", null);
			createPluginType("Immediate download exporter", "Export plugins that implement this interface supports " +
				"immediate downloading of the exported data without having to save it to a file in the BASE " +
				"file system  first. For this to work the plugin must also be given 'immediate exection' rights",
				"net.sf.basedb.core.plugin.ImmediateDownloadExporter", null);
			createPluginType("Analysis filter plug-in", "Analysis plug-ins that are pure filter plug-ins should " +
				"implement this interface. If they do they are linked with the 'Filter bioassayset' button and" +
				"not just the 'Run analysis plug-in' button.",
				"net.sf.basedb.core.plugin.AnalysisFilterPlugin", null);
			createPluginType("Annotation setter", "This interface should be implemented by plug-ins that " +
				"wants to hold annotations that should be copied to another items when it is executed. When " +
				"a plug-in implements this interface, it becomes possible to add annotation of the specified " +
				"types to the plugin definition and/or configuration. The plug-in must copy the annotation values " +
				"itself.",
				"net.sf.basedb.core.plugin.AnnotationSetterPlugin", null);
			createPluginType("External executor IO support", 
				"This is a tagging interface that should be implemented by import and export plug-ins " +
				"that wants to cooperate with the External Executor plug-in in either the export or import " +
				"step. This interface doesn't define any methods but requires that implementing classes " +
				"implement other methods (eg. configure(), run(), etc) in a special way. Consult the " +
				"API documentation for more details.",
				"net.sf.basedb.plugins.executor.IOSupport", null);
			
			createPluginDefinition("net.sf.basedb.plugins.Base1PluginExecuter", null, null, true, null, false);
			if (update && schemaVersion.getSchemaVersion() < 71)
			{
				// Update and unshare the bioassay set exporter plug-in
				PluginDefinition p = createPluginDefinition("net.sf.basedb.plugins.BioAssaySetExporter", null, keyEveryoneUse, true, null, true);
				setPermissions(Item.PLUGINDEFINITION, p.getId(), null, null);
			}
			
			createPluginDefinition("net.sf.basedb.plugins.BaseFileExporterPlugin", null, keyEveryoneUse, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.BfsExporterPlugin", null, keyEveryoneUse, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.FormulaFilter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.HelpExporter", null, null, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.HelpImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.IntensityCalculatorPlugin", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.JepExtraValueCalculator", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.JepIntensityTransformer", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.LowessNormalization", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.MedianRatioNormalization", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.PlateFlatFileImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.PlateMappingExporter", null, null, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.PlateMappingImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.PrintMapFlatFileImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.RawDataFlatFileImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.ReporterFlatFileImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.ReporterMapFlatFileImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.SpotImageCreator", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.ZipFileUnpacker", null, keyEveryoneUse, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.TarFileUnpacker", null, keyEveryoneUse, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.PluginConfigurationExporter", null, keyEveryoneUse, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.PluginConfigurationImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.CdfFileReporterImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.PackedFileExporter", null, keyEveryoneUse, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.AnnotationFlatFileImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.IlluminaRawDataImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.GalExporter", null, keyEveryoneUse, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.ManualTransformCreator", null, keyEveryoneUse, true, null, true);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.BioSourceImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.SampleImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.ExtractImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.LabeledExtractImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.HybridizationImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.ScanImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.RawBioAssayImporter", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.ArrayDesignImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.ArrayBatchImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.batchimport.ArraySlideImporter", null, null, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.executor.ExternalProgramExecutor", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.executor.BfsExporterSupport", null, keyEveryoneUse, true, null, false);
			createPluginDefinition("net.sf.basedb.plugins.executor.BfsImporterSupport", null, keyEveryoneUse, true, null, false);
			
			// Plugin configurations
			if (progress != null) progress.display((int)(22*progress_factor), "--Creating example plugin configurations...");
			createPluginConfigurations("/plugin_configfile.xml", update);
			createPluginConfigurations("/illumina.configurations.xml", update);
			
	
			// Formulas
			ColoringData NO_COLORING = new ColoringData(false, false, null, null, null);
			ColoringData RATIO_COLORING = new ColoringData(true, true, -1f, 0f, 1f);
			if (progress != null) progress.display((int)(23*progress_factor), "--Creating formulas...");
			createFormula(Formula.Type.COLUMN_EXPRESSION, Type.FLOAT, Formula.Parser.JEP, 
				Formula.AverageMethod.GEOMETRIC_MEAN, null, null,
				"Ratio, ch1 / ch2", "The ratio of channel 1 to channel 2", null, 2, 
				new String[] { "ch(1) / ch(2)"}, RATIO_COLORING, rootUser, keyEveryoneUse);
			createFormula(Formula.Type.COLUMN_EXPRESSION, Type.FLOAT, Formula.Parser.JEP, 
				Formula.AverageMethod.ARITHMETIC_MEAN, null, null,
				"M, log2(ch1 / ch2)", "The log of the ratio of channel 1 to channel 2" , null, 2, 
				new String[] { "log2(ch(1) / ch(2))" }, NO_COLORING, rootUser, keyEveryoneUse);
			createFormula(Formula.Type.COLUMN_EXPRESSION, Type.FLOAT, Formula.Parser.JEP, 
				Formula.AverageMethod.ARITHMETIC_MEAN, null, null,
				"A, log10(ch1 * ch2) / 2", "The log of the geometric mean of the two channels", null, 2,
				new String[] { "log(ch(1) * ch(2)) / 2" }, NO_COLORING, rootUser, keyEveryoneUse);
			createFormula(Formula.Type.INTENSITY_TRANSFORMATION, Type.FLOAT, Formula.Parser.JEP, 
				Formula.AverageMethod.NONE, null, null,
				"Swap channel 1 and 2", "Swaps the intensities in channel 1 and 2", null, 2,
				new String[] { "rawCh(2)", "rawCh(1)" }, NO_COLORING, rootUser, keyEveryoneUse);
	
			// Intensity expression formulas defined by raw data types
			int maxChannels = 0;
			for (RawDataType rdt : RawDataTypes.getRawDataTypes())
			{
				if (rdt.getChannels() > maxChannels) maxChannels = rdt.getChannels();
				for (IntensityFormula f : rdt.getIntensityFormulas())
				{
					createFormula(Formula.Type.INTENSITY_EXPRESSION, Type.FLOAT, Formula.Parser.JEP, 
						Formula.AverageMethod.NONE, null, null,
						f.getTitle(), f.getDescription(), rdt.getId(), rdt.getChannels(), 
						f.getExpressions(), NO_COLORING, rootUser, keyEveryoneUse);
				}
			}
			
			// Filter formulas based on channels
			String expression = "";
			String deliminator = "";
			for (int channel = 1; channel <= maxChannels; ++channel)
			{
				expression += deliminator + "ch(" + channel + ") > 0";
				deliminator = " && ";
				createFormula(Formula.Type.COLUMN_RESTRICTION, Type.BOOLEAN, Formula.Parser.JEP, 
					Formula.AverageMethod.NONE, null, null,
					"All intensities > 0", "Remove spots with intensities less than zero", 
					null, channel, new String[] { expression }, NO_COLORING, rootUser, keyEveryoneUse);
			}
	
			// Logarithmic intensity transformations
			List<String> expressions = new ArrayList<String>(maxChannels);
			for (int channel = 1; channel <= maxChannels; ++channel)
			{
				expressions.add("log2(rawCh(" + channel + "))");
				createFormula(Formula.Type.INTENSITY_TRANSFORMATION, Type.FLOAT, Formula.Parser.JEP, 
					Formula.AverageMethod.GEOMETRIC_MEAN, IntensityTransform.NONE, IntensityTransform.LOG2,
					"Log2", "Transform all intensity values to log2 scale", 
					null, channel, expressions.subList(0, channel).toArray(new String[0]), 
					NO_COLORING, rootUser, keyEveryoneUse);
			}
			
			// File set member types
			if (progress != null) progress.display((int)(24*progress_factor), "--Creating data file types...");
			DataFileTypeData celFile = createDataFileType(
				DataFileType.AFFYMETRIX_CEL, "CEL file", "Affymetrix CEL file", 
				Item.RAWBIOASSAY, "cel", rawDataType,  
				"net.sf.basedb.core.filehandler.CelFileHandler", 
				"net.sf.basedb.core.filehandler.CelFileHandler");
			DataFileTypeData cdfFile = createDataFileType(
				DataFileType.AFFYMETRIX_CDF, "CDF file", "Affymetrix CDF file", 
				Item.ARRAYDESIGN, "cdf", reporterMapType,  
				"net.sf.basedb.core.filehandler.CdfFileHandler", 
				"net.sf.basedb.core.filehandler.CdfFileHandler");
			DataFileTypeData reporterMapFile = createDataFileType(
				DataFileType.GENERIC_REPORTER_MAP, "Reporter map", 
				"Generic file that contains information about the reporters present on an array design",
				Item.ARRAYDESIGN, null, reporterMapType, null, null);
			DataFileTypeData printMapFile = createDataFileType(
				DataFileType.GENERIC_PRINT_MAP, "Print map", 
				"Generic file that contains information about how the plates was printed on the array design",
				Item.ARRAYDESIGN, null, printMapType, null, null);
			DataFileTypeData rawDataFile = createDataFileType(
				DataFileType.GENERIC_RAW_DATA, "Generic raw data", 
				"Generic file that contains the raw data of this raw bioassay", 
				Item.RAWBIOASSAY, null, rawDataType, null, null);
			DataFileTypeData spotDataFile = createDataFileType(
				DataFileType.GENERIC_SPOT_DATA, "Generic spot data",
				"Generic file that contains analysed spot data",
				Item.BIOASSAYSET, null, spotDataType, null, null);
			DataFileTypeData extraValueFile = createDataFileType(
				DataFileType.GENERIC_EXTRAVALUE_DATA, "Generic extra value data",
				"Generic file that contains extra value data",
				Item.EXTRAVALUE, null, extraValueType, null, null);
			DataFileTypeData baseFileSerial = createDataFileType(
				DataFileType.BASEFILE_SERIAL, "Serial BASEFile", 
				"A serial BASEFile with spot data",
				Item.BIOASSAYSET, null, spotDataType, null, null);
			DataFileTypeData baseFileMatrix = createDataFileType(
				DataFileType.BASEFILE_MATRIX, "Matrix BASEFile", 
				"A matrix BASEFile with spot data",
				Item.BIOASSAYSET, null, spotDataType, null, null);
			DataFileTypeData bfsMetadataFile = createDataFileType(
					DataFileType.BFS_METADATA, "BFS metadata file", 
					"The metadata file of a BASE file set. The metadata " +
					"file may reference additional files in the [files] section.",
					Item.BIOASSAYSET, null, spotDataType, null, null);
			DataFileTypeData mevTdmsFile = createDataFileType(
				DataFileType.MEV_TDMS, "MeV TDMS", 
				"A MeV Tab-Delimited Multiple Sample file",
				Item.BIOASSAYSET, null, spotDataType, null, null);
			
			// Platforms and variants
			if (progress != null) progress.display((int)(25*progress_factor), "--Creating platforms...");
			createPlatform("generic", "Generic", "Generic platform which expects data to be imported into the database",
				false, null, 0, 
				new PlatformFT(reporterMapFile, false),
				new PlatformFT(printMapFile, false),
				new PlatformFT(rawDataFile, false)
			);
			createPlatform("affymetrix", "Affymetrix", "Affymetrix platform", true, null, 1, 
				new PlatformFT(celFile, true), 
				new PlatformFT(cdfFile, true)
			);
			
			// News
			if (progress != null) progress.display((int)(26*progress_factor), "--Creating news...");
			String version = "BASE " + Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance();
			if (update)
			{
				createNews("Server upgraded to " + version, "Welcome to your upgraded BASE server");
			}
			else
			{
				createNews(version + " installed", "Welcome to your new BASE server.");
			}

			
			if (progress != null) progress.display(100, "Database initialised successfully.");
		}
		catch (BaseException ex)
		{
			if (progress != null) progress.display(100, "Database initialisation failed: " + ex.getMessage()+"\n");
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			session = null;
			Application.stop();
		}
	}
	
	/**
		Create the schema version.
	*/
	private static SchemaVersionData createSchemaVersion(int sv)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		SchemaVersionData schemaVersion = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);

			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SCHEMA_VERSION");
			schemaVersion = HibernateUtil.loadData(SchemaVersionData.class, query);
			if (schemaVersion != null)
			{
				log.info("createSchemaVersion: EXISTS");
				HibernateUtil.commit(tx);
			}
			else
			{
				schemaVersion = new SchemaVersionData();
				schemaVersion.setBuild(Application.getBuild());
				schemaVersion.setSchemaVersion(sv);
				HibernateUtil.saveData(session, schemaVersion);
				HibernateUtil.commit(tx);
				log.info("createSchemaVersion: OK");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
		Create a {@link Directory}.
	*/
	private static DirectoryData createDirectory(UserData owner, DirectoryData parent, String systemId, String name, String description, ItemKeyData itemKey, boolean autoCompress) 
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		DirectoryData directory = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0) 
			{
				directory = HibernateUtil.loadData(session, DirectoryData.class, id);
			}
			if (directory == null && parent != null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SUBDIRECTORY_WITH_NAME");
				query.setString("name", name);
				query.setEntity("parent", parent);
				directory = HibernateUtil.loadData(DirectoryData.class, query);
			}
			
			if (directory != null)
			{
				if (schemaVersion.getSchemaVersion() < 41) directory.setAutoCompress(autoCompress);
				log.info("createDirectory: EXISTS [parent="+parent+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				directory = new DirectoryData();
				if (systemId != null) directory.setSystemId(systemId);
				directory.setParent(parent);
				directory.setName(name);
				directory.setDescription(description);
				directory.setOwner(owner);
				directory.setItemKey(itemKey);
				directory.setAutoCompress(autoCompress);
				HibernateUtil.saveData(session, directory);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(directory);
				log.info("createDirectory: OK [SYSTEM ID="+systemId+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createDirectory: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return directory;
	}

	/**
		Create a {@link MimeType}.
	*/
	private static MimeTypeData createMimeType(String name, String description, String extension, FileTypeData fileType, boolean autoCompress)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		MimeTypeData mimeType = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (mimeType == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_MIME_TYPE_FROM_EXTENSION");
				query.setString("extension", extension);
				mimeType = HibernateUtil.loadData(MimeTypeData.class, query);
			}
			
			if (mimeType != null)
			{
				if (schemaVersion.getSchemaVersion() < 17) mimeType.setFileType(fileType);
				if (schemaVersion.getSchemaVersion() < 41) mimeType.setAutoCompress(autoCompress);
				log.info("createMimeType: EXISTS [extension="+extension+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				mimeType = new MimeTypeData();
				mimeType.setEntryDate(new Date());
				mimeType.setName(name);
				mimeType.setDescription(description);
				mimeType.setExtension(extension);
				mimeType.setFileType(fileType);
				mimeType.setAutoCompress(autoCompress);
				HibernateUtil.saveData(session, mimeType);
				HibernateUtil.commit(tx);
				log.info("createMimeType: OK [extension="+extension+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createMimeType: FAILED [extension="+extension+"; name="+name+"]", ex);
			throw ex;
		}
		return mimeType;
	}
	
	/**
		Create a {@link FileType}.
	*/
	private static FileTypeData createFileType(String systemId, String name, String description)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		FileTypeData fileType = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0) 
			{
				fileType = HibernateUtil.loadData(session, FileTypeData.class, id);
			}
			if (fileType == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_FILE_TYPE_WITH_NAME");
				query.setString("name", name);
				fileType = HibernateUtil.loadData(FileTypeData.class, query);
			}

			if (fileType != null)
			{
				log.info("createFileType: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				fileType = new FileTypeData();
				fileType.setEntryDate(new Date());
				if (systemId != null) fileType.setSystemId(systemId);
				fileType.setName(name);
				fileType.setDescription(description);
				HibernateUtil.saveData(session, fileType);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(fileType);
				log.info("createFileType: OK [SYSTEM ID="+systemId+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createFileType: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return fileType;
	}

	/**
		Create a {@link QuotaType}.
	*/
	private static QuotaTypeData createQuotaType(String systemId, String name, String description, boolean secondaryLocation)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		QuotaTypeData quotaType = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				quotaType = HibernateUtil.loadData(session, QuotaTypeData.class, id);
			}
			if (quotaType == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_QUOTA_TYPE_WITH_NAME");
				query.setString("name", name);
				quotaType = HibernateUtil.loadData(QuotaTypeData.class, query);
			}
			
			if (quotaType != null)
			{
				log.info("createQuotaType: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				quotaType = new QuotaTypeData();
				quotaType.setEntryDate(new Date());
				if (systemId != null) quotaType.setSystemId(systemId);
				quotaType.setName(name);
				quotaType.setDescription(description);
				quotaType.setSecondaryLocation(secondaryLocation);
				HibernateUtil.saveData(session, quotaType);
				HibernateUtil.commit(tx);
				if(systemId != null) SystemItems.add(quotaType);
				log.info("createQuotaType: OK [SYSTEM ID="+systemId+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createQuotaType: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return quotaType;
	}

	/**
		Create a {@link Quota}.
	*/
	private static QuotaData createQuota(String systemId, String name, String description, long totalPrimaryQuota)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		QuotaData quota = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				quota = HibernateUtil.loadData(session, QuotaData.class, id);
			}
			
			if (quota == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_QUOTA_WITH_NAME");
				query.setString("name", name);
				quota = HibernateUtil.loadData(QuotaData.class, query);
			}
			
			if (quota != null)
			{
				log.info("createQuota: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				if (schemaVersion.getSchemaVersion() <= 63)
				{
					// Fixes spelling error
					quota.setDescription(description);
				}
				HibernateUtil.commit(tx);
			}
			else
			{
				quota = new QuotaData();
				if (systemId != null) quota.setSystemId(systemId);
				quota.setName(name);
				quota.setDescription(description);
				QuotaTypeData total = HibernateUtil.loadData(session, QuotaTypeData.class, SystemItems.getId(QuotaType.TOTAL));
				quota.getQuotaValues().put(new QuotaIndex(total, Location.PRIMARY.getValue()), totalPrimaryQuota);
				HibernateUtil.saveData(session, quota);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(quota);
				log.info("createQuota: OK [SYSTEM ID="+systemId+"; name="+name+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createQuota: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return quota;
	}
	/**
		Create a {@link User}.
	*/
	private static UserData createUser(String systemId, String login, String password, 
		String name, String description, RoleData role, QuotaData quota, 
		boolean disabled, boolean multiUserAccount)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		UserData user = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				user = HibernateUtil.loadData(session, UserData.class, id);
			}
			
			if (user == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_USER_FOR_LOGIN");
				query.setString("login", login);
				user = HibernateUtil.loadData(UserData.class, query);
			}
			
			if (user != null)
			{
				log.info("createUser: EXISTS [SYSTEM ID="+systemId+"; login="+login+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				user = new UserData();
				user.setEntryDate(new Date());
				if (systemId != null) user.setSystemId(systemId);
				user.setLogin(login);
				user.getPassword().setMd5Password(MD5.getHashString(password));
				user.setName(name);
				user.setDescription(description);
				user.setQuota(quota);
				user.setDisabled(disabled);
				user.setMultiuserAccount(multiUserAccount);
				if (role != null)
				{
					role.getUsers().add(user);
				}
				HibernateUtil.saveData(session, user);
				HibernateUtil.commit(tx);
				if(systemId != null) SystemItems.add(user);
				log.info("createUser: OK [SYSTEM ID="+systemId+"; login="+login+"]");
			}
		}
		catch(BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createUser: FAILED [SYSTEM ID="+systemId+"; login="+login+"]", ex);
			throw ex;
		}
		return user;
	}

	/**
		Create a {@link Group}.
	*/
	private static GroupData createGroup(String systemId, String name, boolean isDefault, 
		boolean hiddenMembers, String description)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		GroupData group = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				group = HibernateUtil.loadData(session, GroupData.class, id);
			}
			if (group == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_GROUP_WITH_NAME");
				query.setString("name", name);
				group = HibernateUtil.loadData(GroupData.class, query);
			}
			
			if (group != null)
			{
				if (schemaVersion.getSchemaVersion() < 32)
				{
					// Set default property if upgrading
					group.setDefault(isDefault);
				}
				if (schemaVersion.getSchemaVersion() < 75)
				{
					// Set default property if upgrading
					group.setHiddenMembers(hiddenMembers);
				}
				log.info("createGroup: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				group = new GroupData();
				group.setEntryDate(new Date());
				if(systemId != null) group.setSystemId(systemId);
				group.setName(name);
				group.setDescription(description);
				group.setDefault(isDefault);
				group.setHiddenMembers(hiddenMembers);
				HibernateUtil.saveData(session, group);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(group);
				log.info("createGroup: OK [SYSTEM ID="+systemId+" name="+name+"]");
			}
		}
		catch(BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createGroup: FAILED [SYSTEM ID="+systemId+" name="+name+"]", ex);
			throw ex;
		}
		return group;
	}

	/**
		Create a {@link Role}.
	*/
	private static RoleData createRole(String systemId, String name, boolean isDefault, String description)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		RoleData role = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				role = HibernateUtil.loadData(session, RoleData.class, id);
			}
			if (role == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_ROLE_WITH_NAME");
				query.setString("name", name);
				role = HibernateUtil.loadData(RoleData.class, query);
			}

			if (role != null)
			{
				if (schemaVersion.getSchemaVersion() < 32)
				{
					// Set default property if upgrading
					role.setDefault(isDefault);
				}
				log.info("createRole: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				role = new RoleData();
				role.setEntryDate(new Date());
				if (systemId != null) role.setSystemId(systemId);
				role.setName(name);
				role.setDescription(description);
				role.setDefault(isDefault);
				HibernateUtil.saveData(session, role);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(role);
				log.info("createRole: OK [SYSTEM ID="+systemId+"; name="+name+"]");
			}
		}
		catch(BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createRole: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return role;
	}
	
	@SuppressWarnings("unchecked")
	private static void installUnits(String xmlFile, boolean update)
	{
		String schemaFileURL = Install.class.getResource("/net/sf/basedb/core/xsd/units.xsd").toString();
		String namespace = "http://base.thep.lu.se/units.xsd";
		URL fileURL = Install.class.getResource(xmlFile);
		Document doc = null;
		
		try
		{
			doc = XMLUtil.getSchemaValidatedXML(fileURL, namespace, schemaFileURL);
		}
		catch (Exception ex)
		{
			log.error("installUnits: FAILED [file=" + xmlFile + "]", ex);
			throw new BaseException(ex);
		}

		Element root = doc.getRootElement();
		Namespace ns = root.getNamespace();
		
		for (Element quantityElement : (List<Element>)root.getChildren("quantity", ns))
		{
			String systemId = quantityElement.getAttributeValue("system-id");
			String name = quantityElement.getChildTextTrim("name", ns);
			String referenceUnit = quantityElement.getChildTextTrim("reference-unit", ns);
			String description = quantityElement.getChildTextTrim("description", ns);
			QuantityData quantity = createQuantity(systemId, name, description, referenceUnit);
			
			for (Element unitElement : (List<Element>)quantityElement.getChildren("unit", ns))
			{
				String unitName = unitElement.getChildTextTrim("name", ns);
				String symbol = unitElement.getChildTextTrim("symbol", ns);
				double factor = Values.getDouble(unitElement.getChildTextTrim("factor", ns), 1.0);
				double offset = Values.getDouble(unitElement.getChildTextTrim("offset", ns), 0.0);
				String unitDescription = unitElement.getChildTextTrim("description", ns);
				
				List<String> aliases = new ArrayList<String>();
				for (Element aliasElement : (List<Element>)unitElement.getChildren("alias", ns))
				{
					aliases.add(aliasElement.getTextTrim());
				}
				createUnit(quantity, unitName, unitDescription, factor, offset, symbol, 
						aliases.toArray(new String[aliases.size()]));
			}
		}
	}
	
	/**
		Create a {@link Quantity}.
	*/
	private static QuantityData createQuantity(String systemId, String name, 
			String description, String referenceUnit)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		QuantityData q = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				q = HibernateUtil.loadData(session, QuantityData.class, id);
			}
			if (q != null)
			{
				log.info("createQuantity: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				q = new QuantityData();
				q.setSystemId(systemId);
				q.setName(name);
				q.setDescription(description);
				q.setReferenceUnit(referenceUnit);
				HibernateUtil.saveData(session, q);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(q);
				log.info("createQuantity: OK [SYSTEM ID="+systemId+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createQuantity: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return q;
	}

	/**
		Create a {@link Unit}.
	*/
	private static UnitData createUnit(QuantityData quantity, String name, 
			String description, double factor, double offset, String symbol,
			String... aliasSymbols)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		UnitData u = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (u == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"GET_UNIT_WITH_NAME_FOR_QUANTITY");
				query.setString("name", name);
				query.setEntity("quantity", quantity);
				u = HibernateUtil.loadData(UnitData.class, query);
			}
			
			Set<String> allSymbols = new HashSet<String>();
			allSymbols.add(symbol);
			if (aliasSymbols != null)
			{
				allSymbols.addAll(Arrays.asList(aliasSymbols));
			}
			
			if (u != null)
			{
				log.info("createUnit: EXISTS [name="+name+"]");
				
				Set<String> existingAliases = new HashSet<String>();
				for (UnitSymbolData alias : u.getSymbols())
				{
					existingAliases.add(alias.getSymbol());
				}
				for (String aliasSymbol : allSymbols)
				{
					if (!existingAliases.contains(aliasSymbol))
					{
						UnitSymbolData alias = new UnitSymbolData();
						alias.setQuantity(quantity);
						alias.setUnit(u);
						alias.setSymbol(aliasSymbol);
						u.getSymbols().add(alias);
					}
				}
				HibernateUtil.commit(tx);
			}
			else
			{
				u = new UnitData();
				u.setQuantity(quantity);
				u.setName(name);
				u.setDescription(description);
				u.setDisplaySymbol(symbol);
				u.setReferenceFactor(factor);
				u.setReferenceOffset(offset);
				HibernateUtil.saveData(session, u);
			
				for (String aliasSymbol : allSymbols)
				{
					UnitSymbolData alias = new UnitSymbolData();
					alias.setQuantity(quantity);
					alias.setUnit(u);
					alias.setSymbol(aliasSymbol);
					u.getSymbols().add(alias);
				}
				HibernateUtil.commit(tx);
				log.info("createUnit: OK [name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createUnit: FAILED [name="+name+"]", ex);
			throw ex;
		}
		return u;
	}

	/**
		Create a {@link ProtocolType}.
	*/
	private static ProtocolTypeData createProtocolType(String systemId, String name, String description)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		ProtocolTypeData pt = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				pt = HibernateUtil.loadData(session, ProtocolTypeData.class, id);
			}
			if (pt == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_PROTOCOL_TYPE_WITH_NAME");
				query.setString("name", name);
				pt = HibernateUtil.loadData(ProtocolTypeData.class, query);
			}
			
			if (pt != null)
			{
				log.info("createProtocolType: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				pt = new ProtocolTypeData();
				pt.setEntryDate(new Date());
				if (systemId != null) pt.setSystemId(systemId);
				pt.setName(name);
				pt.setDescription(description);
				HibernateUtil.saveData(session, pt);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(pt);
				log.info("createProtocolType: OK [SYSTEM ID="+systemId+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createProtocolType: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return pt;
	}
	/**
		Create a {@link SoftwareType}.
	*/
	private static SoftwareTypeData createSoftwareType(String systemId, String name, String description)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		SoftwareTypeData st = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				st = HibernateUtil.loadData(session, SoftwareTypeData.class, id);
			}
			if (st == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SOFTWARE_TYPE_WITH_NAME");
				query.setString("name", name);
				st = HibernateUtil.loadData(SoftwareTypeData.class, query);
			}

			if (st != null)
			{
				log.info("createSoftwareType: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				st = new SoftwareTypeData();
				st.setEntryDate(new Date());
				if (systemId != null) st.setSystemId(systemId);
				st.setName(name);
				st.setDescription(description);
				HibernateUtil.saveData(session, st);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(st);
				log.info("createSoftwareType: OK [SYSTEM ID="+systemId+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createSoftwareType: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return st;
	}

	/**
		Create a {@link Software}.
	*/
	private static SoftwareData createSoftware(String name, String version, String description, SoftwareTypeData type, UserData owner, ItemKeyData shareTo)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		SoftwareData sw = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
		
			if (sw == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SOFTWARE_WITH_NAME");
				query.setString("name", name);
				sw = HibernateUtil.loadData(SoftwareData.class, query);
			}
			
			if (sw != null)
			{
				log.info("createSoftware: EXISTS [name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				sw = new SoftwareData();
				sw.setEntryDate(new Date());
				sw.setName(name);
				sw.setDescription(description);
				sw.setVersionString(version);
				sw.setSoftwareType(type);
				sw.setOwner(owner);
				sw.setItemKey(shareTo);
				HibernateUtil.saveData(session, sw);
				HibernateUtil.commit(tx);
				log.info("createSoftware: OK [name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createSoftware: FAILED [name="+name+"]", ex);
			throw ex;
		}
		return sw;
	}

	/**
		Create a {@link HardwareType}.
	*/
	private static HardwareTypeData createHardwareType(String systemId, String name, String description)
		throws BaseException
	{
		int id = SystemItems.getId(systemId);
		org.hibernate.Transaction tx = null;
		HardwareTypeData ht = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				ht = HibernateUtil.loadData(session, HardwareTypeData.class, id);
			}
			if (ht == null && systemId == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_HARDWARE_TYPE_WITH_NAME");
				query.setString("name", name);
				ht = HibernateUtil.loadData(HardwareTypeData.class, query);
			}
			
			if (ht != null)
			{
				log.info("createHardwareType: EXISTS [SYSTEM ID="+systemId+"; name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				ht = new HardwareTypeData();
				ht.setEntryDate(new Date());
				if (systemId != null) ht.setSystemId(systemId);
				ht.setName(name);
				ht.setDescription(description);
				HibernateUtil.saveData(session, ht);
				HibernateUtil.commit(tx);
				if (systemId != null) SystemItems.add(ht);
				log.info("createHardwareType: OK [SYSTEM ID="+systemId+"; name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createHardwareType: FAILED [SYSTEM ID="+systemId+"; name="+name+"]", ex);
			throw ex;
		}
		return ht;
	}
	
	/**
		Create a {@link HardwareData}.
	*/
	private static HardwareData createHardware(String name, String version, String description, HardwareTypeData type, UserData owner, ItemKeyData shareTo)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		HardwareData hw = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			if (hw == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_HARDWARE_WITH_NAME");
				query.setString("name", name);
				hw = HibernateUtil.loadData(HardwareData.class, query);
			}
			
			if (hw != null)
			{
				log.info("createHardware: EXISTS [name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				hw = new HardwareData();
				hw.setEntryDate(new Date());
				hw.setName(name);
				hw.setDescription(description);
				hw.setVersionString(version);
				hw.setHardwareType(type);
				hw.setOwner(owner);
				hw.setItemKey(shareTo);
				HibernateUtil.saveData(session, hw);
				HibernateUtil.commit(tx);
				log.info("createHardware: OK [name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createHardware: FAILED [name="+name+"]", ex);
			throw ex;
		}
		return hw;
	}

	/**
		Create a {@link PlateGeometry}.
	*/
	private static PlateGeometryData createPlateGeometry(String name, String description, int rows, int columns)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		PlateGeometryData pg = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			if (pg == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_PLATE_GEOMETRY_WITH_NAME");
				query.setString("name", name);
				pg = HibernateUtil.loadData(PlateGeometryData.class, query);
			}

			if (pg != null)
			{
				log.info("createPlateGeometry: EXISTS [name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				pg = new PlateGeometryData();
				pg.setEntryDate(new Date());
				pg.setName(name);
				pg.setDescription(description);
				pg.setRows(rows);
				pg.setColumns(columns);
				HibernateUtil.saveData(session, pg);
				HibernateUtil.commit(tx);
				log.info("createPlateGeometry: OK [NAME="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createPlateGeometry: FAILED [NAME="+name+"]", ex);
			throw ex;
		}
		return pg;
	}

	/**
		Create a {@link PlateMapping}. At the moment only 1:1 and 4:1 mappings are alloed. If the 4:1 mapping
		is choosen it creates a 1243 mapping as defines in the "Merge 96-well -> 384-well" in base 1.2.14. 
		
		@param source the PlateGeometry of the source plate(s). 
		@param sourceCount the number of source plates. 1 or 4 at the moment.
		@param destination the PlateGeometry of the destination plate(s).
		@param destinationCount the number of destination plate. Only 1 at the moment.
		@throws BaseException
	 */
	private static PlateMappingData createPlateMapping(String name, String description, PlateGeometryData source, int sourceCount, PlateGeometryData destination, int destinationCount, String image, UserData owner, ItemKeyData shareTo ) 
		throws BaseException
	{
		if((sourceCount != 1 || sourceCount != 4) && (destinationCount != 1))
		{
			log.error("createPlateMapping: FAILED: Can't create mapping " + sourceCount + ":" + destinationCount);
			return null;
		}
		
		org.hibernate.Transaction tx = null;
		PlateMappingData pm = null;
		try 
		{
			tx = HibernateUtil.newTransaction(session);
			if (pm == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_PLATE_MAPPING_WITH_NAME");
				query.setString("name", name);
				pm = HibernateUtil.loadData(PlateMappingData.class, query);
			}
		
			if (pm != null)
			{
				log.info("createPlateMapping: EXISTS [name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				pm = new PlateMappingData();
				pm.setEntryDate(new Date());
				pm.setName(name);
				pm.setDescription(description);
				pm.setImage(image);
				pm.setSourceCount(sourceCount);
				pm.setSourceGeometry(source);
				pm.setDestinationCount(destinationCount);
				pm.setDestinationGeometry(destination);
				pm.setOwner(owner);
				pm.setItemKey(shareTo);
				
				Map<MappingCoordinate,MappingCoordinate> details = pm.getDetails();
				if (sourceCount == 1 && destinationCount == 1)
				{
					if(source.getRows() != destination.getRows() || source.getColumns() != destination.getColumns())
					{
						throw new BaseException("The number of rows and columns must be the same in the destination and source PlateGeometires to create a 1:1 mapping.");
					}
					for (int col = 0; col < source.getColumns(); col++)
					{
						for (int row = 0; row < source.getRows(); row++)
						{
							MappingCoordinate from = new MappingCoordinate(0, row, col);
							MappingCoordinate to = new MappingCoordinate(0, row, col);
							details.put(to, from);
						}
					}
				}
				else if (sourceCount == 4 && destinationCount == 1)
				{
					int sourceRows = source.getRows();
					int sourceColumns = source.getColumns(); 
					
					if (sourceColumns * 2 != destination.getColumns() || sourceRows * 2 != destination.getRows())
					{
						throw new BaseException("The source PlateGeometry must be one quarter of the destination PlateGeometry to create a 4:1 mapping.");
					}
					for (int col = 0; col < source.getColumns(); col++)
					{
						for (int row = 0; row < source.getRows(); row++)
						{
							MappingCoordinate from1 = new MappingCoordinate(0, row, col);
							MappingCoordinate from2 = new MappingCoordinate(1, row, col);
							MappingCoordinate from3 = new MappingCoordinate(2, row, col);
							MappingCoordinate from4 = new MappingCoordinate(3, row, col);
	
							MappingCoordinate to1 = new MappingCoordinate(0, 2 * row, 2 * col);
							MappingCoordinate to2 = new MappingCoordinate(0, 2 * row, 2 * col + 1);
							MappingCoordinate to3 = new MappingCoordinate(0, 2 * row + 1, 2 * col + 1);
							MappingCoordinate to4 = new MappingCoordinate(0, 2 * row + 1, 2 * col);
							
							details.put(to1, from1);
							details.put(to2, from2);
							details.put(to3, from3);
							details.put(to4, from4);
						}
					}
				}
				HibernateUtil.saveData(session, pm);
				HibernateUtil.commit(tx);
				log.info("createPlateMapping: OK [NAME="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createPlateMapping: FAILED [NAME="+name+"]", ex);
			throw ex;
		}
		return pm;
	}

	/**
		Create a {@link Label}.
	*/
	private static LabelData createLabel(String name, String description, UserData owner, ItemKeyData shareTo)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		LabelData label = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (label == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_LABEL_WITH_NAME");
				query.setString("name", name);
				label = HibernateUtil.loadData(LabelData.class, query);
			}
			
			if (label != null)
			{
				log.info("createLabel: EXISITS [NAME="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				label = new LabelData();
				label.setEntryDate(new Date());
				label.setName(name);
				label.setDescription(description);
				label.setOwner(owner);
				label.setItemKey(shareTo);
				HibernateUtil.saveData(session, label);
				HibernateUtil.commit(tx);
				log.info("createLabel: OK [NAME="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.info("createLabel: FAILED [NAME="+name+"]", ex);
			throw ex;
		}
		return label;
	}
	
	/**
		Create an {@link AnnotationType}.
	*/
	private static AnnotationTypeData createAnnotationType(String name, String description, Type type,
		int multiplicity, Object[] enumeration, Item[] itemTypes, 
		Map<String, String> options, boolean requiredForMiame, boolean isProtocolParameter, 
		UserData owner, ItemKeyData shareTo)
	{
		org.hibernate.Transaction tx = null;
		AnnotationTypeData at = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			if (at == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_ANNOTATION_TYPE_WITH_NAME");
				query.setString("name", name);
				at = HibernateUtil.loadData(AnnotationTypeData.class, query);
			}
			
			if (at != null)
			{
				log.info("createAnnotationType: OK [EXISTS="+name+"; TYPE="+type);
				HibernateUtil.commit(tx);
			}
			else
			{
				at = new AnnotationTypeData();
				at.setEntryDate(new Date());
				at.setName(name);
				at.setDescription(description);
				at.setOwner(owner);
				at.setItemKey(shareTo);
				at.setValueType(type.getValue());
				at.setMultiplicity(multiplicity);
				at.setWidth(30);
				at.setHeight(1);
				at.setRequiredForMiame(requiredForMiame);
				at.setProtocolParameter(isProtocolParameter);
				if (options != null) at.getOptions().putAll(options);
				
				if (itemTypes != null)
				{
					Set<Integer> itemCodes = at.getItemTypes();
					for (Item item : itemTypes)
					{
						itemCodes.add(item.getValue());
					}
				}
				
				if (enumeration != null)
				{
					ParameterValueData<?> enumValues = type.newParameterValueData();
					enumValues.replaceValues(Arrays.asList(enumeration));
					at.setEnumeration(true);
					at.setEnumerationValues(enumValues);
				}
				
				HibernateUtil.saveData(session, at);
				HibernateUtil.commit(tx);
				log.info("createAnnotationType: OK [NAME="+name+"; TYPE="+type);
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.info("createAnnotationType: FAILED [NAME="+name+"; TYPE="+type, ex);
			throw ex;
		}
		return at;
	}
	

	/**
		Create a {@link RoleKey}.
		@param permissions A map with RoleData as keys and permission codes as values
	*/
	private static RoleKeyData createRoleKey(Item itemType, String name, String description, Map<RoleData, Integer> permissions)
		throws BaseException
	{
		int id = SystemItems.getRoleKeyId(itemType);
		org.hibernate.Transaction tx = null;
		RoleKeyData key = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (id != 0)
			{
				log.info("createRoleKey: EXISTS [NAME="+name+"]");
				key = HibernateUtil.loadData(session, RoleKeyData.class, id);
				if (schemaVersion.getSchemaVersion() <= 63)
				{
					// Fixes spelling error
					key.setName(name);
					key.setDescription(description);
				}
				HibernateUtil.commit(tx);
			}
			else
			{
				key = new RoleKeyData();
				key.setItemType(itemType.getValue());
				key.setName(name);
				key.setDescription(description);
				Map<RoleData, Integer> roleKeyPermissions = key.getRoles();
				int definedPermissions = Permission.grant(itemType.getDefinedPermissions());
				for (Map.Entry<RoleData, Integer> entry : permissions.entrySet())
				{
					roleKeyPermissions.put(entry.getKey(), entry.getValue() & definedPermissions);
				}
				HibernateUtil.saveData(session, key);
				HibernateUtil.commit(tx);
				SystemItems.addRoleKey(key);
				log.info("createRoleKey: OK [NAME="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createRoleKey: FAILED [NAME="+name+"]", ex);
			throw ex;
		}
		return key;
	}
	
	/**
		Add permissions for a role for specific item types. The role keys for
		all item types must exist.
		@param role The role to add the permissions to
		@param permissions The permissions to add
	*/
	private static void addRolePermissions(RoleData role, Map<Item, Integer> permissions)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			for (Map.Entry<Item, Integer> entry : permissions.entrySet())
			{
				Item itemType = entry.getKey();
				RoleKeyData key = HibernateUtil.loadData(session, RoleKeyData.class, 
						SystemItems.getRoleKeyId(itemType));
				if (key == null)
				{
					log.warn("addRolePermissions: MISSING KEY [ROLE="+role.getName()+"; ITEM="+itemType.name()+"]");
				}
				else
				{
					int definedPermissions = Permission.grant(itemType.getDefinedPermissions());
					Map<RoleData, Integer> roleKeyPermissions = key.getRoles();
					roleKeyPermissions.put(role, entry.getValue() & definedPermissions);
				}
			}
			HibernateUtil.commit(tx);
			log.info("addRolePermissions: OK [ROLE="+role.getName()+"]");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("addRolePermissions: FAILED [ROLE="+role.getName()+"]", ex);
			throw ex;
		}
	}
	
	/**
		Create an {@link ItemKey}.
		@param message Message will be written to the log
		@param userPermissions A map with UserData as keys and permission codes as values
		@param groupPermissions A map with GroupData as keys and permission codes as values
	*/
	private static ItemKeyData createItemKey(String message, Map<UserData,Integer> userPermissions, Map<GroupData,Integer> groupPermissions)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			ItemKeyData key = new ItemKeyData();
			if (userPermissions != null) key.getUsers().putAll(userPermissions);
			if (groupPermissions != null) key.getGroups().putAll(groupPermissions);
			HibernateUtil.saveData(session, key);
			HibernateUtil.commit(tx);
			log.info("createItemKey: OK ["+message+"]");
			return key;
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createItemKey: FAILED ["+message+"]", ex);
			throw ex;
		}
	}

	/**
		Create a {@link Client}.
	*/
	private static ClientData createClient(UserData owner, String externalId, String name, String description, ItemKeyData key)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		ClientData client = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
		
			if (client == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_CLIENT_FOR_EXTERNAL_ID");
				query.setString("externalId", externalId);
				client = HibernateUtil.loadData(ClientData.class, query);
			}
			
			if (client != null)
			{
				log.info("createClient: EXISTS [externalId="+externalId+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				client = new ClientData();
				client.setEntryDate(new Date());
				client.setOwner(owner);
				client.setExternalId(externalId);
				client.setName(name);
				client.setDescription(description);
				client.setItemKey(key);
	
				HibernateUtil.saveData(session, client);
				HibernateUtil.commit(tx);
				log.info("createClient: OK [externalId="+externalId+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createClient: FAILED [externalId="+externalId+"]", ex);
			throw ex;
		}
		return client;
	}

	/**
		Create a {@link News}.
	*/
	private static NewsData createNews(String name, String description)
		throws BaseException
	{

		NewsData news = null;
		org.hibernate.Transaction tx = null;
		
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			if (news == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_NEWS_WITH_NAME");
				query.setString("name", name);
				news = HibernateUtil.loadData(NewsData.class, query);
			}
			if (news != null)
			{
				log.info("createNews: EXISTS [name="+name+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				news = new NewsData();
				news.setEntryDate(new Date());
				news.setName(name);
				news.setDescription(description);
				Date today = new Date();
				news.setStartDate(today);
				news.setNewsDate(today);
				news.setEndDate(new Date(System.currentTimeMillis()+1000*3600*24*7));  // 1 week from now
				HibernateUtil.saveData(session, news);
				HibernateUtil.commit(tx);
				log.info("createNews: OK [name="+name+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createNews: FAILED [name="+name+"]", ex);
			throw ex;
		}
		return news;
	}
	/**
		Create a {@link PluginType}.
	*/
	private static PluginTypeData createPluginType(String name, String description, String interfaceName, String jarPath)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		PluginTypeData pluginType = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (pluginType == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_PLUGINTYPE_FOR_INTERFACENAME");
				query.setString("interfaceName", interfaceName);
				pluginType = HibernateUtil.loadData(PluginTypeData.class, query);
			}
			
			if (pluginType != null)
			{
				log.info("createPluginType: EXISTS [interfaceName="+interfaceName+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				pluginType = new PluginTypeData();
				pluginType.setEntryDate(new Date());
				pluginType.setName(name);
				pluginType.setDescription(description);
				pluginType.setInterfaceName(interfaceName);
				pluginType.setJarPath(jarPath);
				HibernateUtil.saveData(session, pluginType);
				HibernateUtil.commit(tx);
				log.info("createPluginType: OK [interfaceName="+interfaceName+"]");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("createPluginType: FAILED [interfaceName="+interfaceName+"]", ex);
			throw ex;
		}
		return pluginType;

	}
	/**
		Create a {@link PluginDefinition}.
	*/
	private static PluginDefinition createPluginDefinition(String className, String jarPath, 
		ItemKeyData shareTo, boolean trusted, Long maxMemory, boolean allowImmediateExecution)
		throws BaseException
	{
		PluginDefinition pd = null;
		DbControl dc = null;
		try
		{
			dc = sessionControl.newDbControl();
			
			try
			{
				pd = PluginDefinition.getByClassName(dc, className);
			}
			catch (ItemNotFoundException ex)
			{}
			
			if (pd != null)
			{
				int version = schemaVersion.getSchemaVersion();
				pd.loadPluginInformation(jarPath, className, true);
				if (version < 21) pd.setAllowImmediateExecution(allowImmediateExecution);
				dc.commit();
				log.info("createPluginDefinition: UPDATED [className="+className+"; jarPath="+jarPath+"]");
			}
			else
			{
				pd = PluginDefinition.getNew(dc, className, jarPath, true);
				pd.setTrusted(trusted);
				pd.setMaxMemory(maxMemory);
				pd.setAllowImmediateExecution(allowImmediateExecution);
				pd.getData().setItemKey(shareTo);
				dc.saveItem(pd);
				dc.commit();
				log.info("createPluginDefinition: OK [className="+className+"; jarPath="+jarPath+"]");
			}
		}
		catch (BaseException ex)
		{
			log.error("createPluginDefinition: FAILED [className="+className+"; jarPath="+jarPath+"]", ex);
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return pd;
	}
		
	/**
		Create {@link PluginConfiguration}s from a XML file.
	*/
	private static void createPluginConfigurations(String filePath, boolean update)
		throws BaseException
	{
		DbControl dc = null;
		
		try
		{
			dc = sessionControl.newDbControl();
			
			org.hibernate.Query query = null;
			if (update)
			{
				// If we are updating, we must check if a configuration with a 
				// given name already exists
				query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
					"GET_PLUGINCONFIGURATION_FOR_PLUGIN_WITH_NAME");
				/*
					SELECT pc
					FROM PluginConfigurationData pc
					WHERE pc.pluginDefinition.className = :className
					AND pc.name = :name
				*/
			}
			
			URL dtdURL = Install.class.getResource("/net/sf/basedb/core/dtd/plugin-configuration-file.dtd");
			URL fileURL = Install.class.getResource(filePath);
			Document doc = XMLUtil.getValidatedXml(fileURL, dtdURL);
			Element rootElement = doc.getRootElement();		
			
			List configurations = rootElement.getChildren();
			
			for (Object obj : configurations)
			{
				Element configuration = (Element)obj;
				String name = configuration.getChildText("configname");
				String description = configuration.getChildText("description");
				String pluginClassName = configuration.getAttributeValue("pluginClassName");
				
				if (update)
				{
					// Check if it already exists
					query.setString("className", pluginClassName);
					query.setString("name", name);
					PluginConfigurationData config = HibernateUtil.loadData(PluginConfigurationData.class, query);
					
					if (config != null)
					{
						log.info("createPluginConfigurations: EXISTS [plugin="+pluginClassName+
								"; name=" + name + "]");
						/* #### CONTINUE-STATEMENT #### */
						continue;
						// Continue with next configuration
					}
				}
				
				try
				{
					PluginDefinition pluginDefinition = PluginDefinition.getByClassName(dc, pluginClassName);
															
					// Create a new plugin configurationj
					PluginConfiguration pluginConfig = PluginConfiguration.getNew(dc, pluginDefinition);	
					pluginConfig.setItemKey(pluginDefinition.getItemKey());
					pluginConfig.setProjectKey(pluginDefinition.getProjectKey());
					dc.saveItem(pluginConfig);
					pluginConfig.setName(name);
					pluginConfig.setDescription(description);
					
					// Get the confiuration parameters from file
					List parameters = configuration.getChildren("parameter");
					for (Object elementObj : parameters)
					{	
						Element parameter = (Element)elementObj;
						String parametername = parameter.getChildText("name");								
						String cl = parameter.getChildText("class"); 
						String label = parameter.getChildText("label").length() > 0 ? parameter.getChildText("label") : name; 				
						Class clazz = null;
						List<Object> values = new ArrayList<Object>() ;
						
						if (cl.length() > 0)
						{	
							clazz = Class.forName(cl);
							List children = parameter.getChildren("value");
							// Get the parameter values
							for (Object ch : children)
							{
								String text = ((Element)ch).getText();
								if (clazz.equals(Boolean.class)) values.add(new Boolean(text));
								else if (clazz.equals(Date.class)) values.add(DateUtil.parseString(text));
								else if (clazz.equals(Double.class)) values.add(new Double(text));
								else if (clazz.equals(Float.class)) values.add(new Float(text));
								else if (clazz.equals(Long.class)) values.add(new Long(text));
								else if (clazz.equals(String.class)) values.add(text);
							}		
							// Get the parametertype
							ParameterType pType;
							if (clazz.equals(Boolean.class)) pType = new BooleanParameterType();
							else if (clazz.equals(Date.class)) pType =  new DateParameterType();
							else if (clazz.equals(Double.class)) pType =  new DoubleParameterType();		
							else if (clazz.equals(Float.class)) pType =  new FloatParameterType();
							else if (clazz.equals(Integer.class)) pType =  new IntegerParameterType();		
							else if (clazz.equals(Long.class)) pType =  new LongParameterType();
							else if (clazz.equals(String.class)) pType =  new StringParameterType();
							else pType = null;
							pluginConfig.setParameterValues(parametername,label, "", pType, values);
						}
					}					
				}				
				catch(Throwable ex)
				{
					log.error("createPluginConfiguration: FAILED[Configuration: " + name, ex);					
				}
			}				
			dc.commit();
		}
		catch(Throwable ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	/**
		Create a {@link Formula}.
	*/
	private static FormulaData createFormula(Formula.Type type, Type valueType, Formula.Parser parser, 
		Formula.AverageMethod avgMethod, IntensityTransform source, IntensityTransform result,
		String name, String description, 
		String rawDataType, int channels, 
		String[] formulas, ColoringData coloring, 
		UserData owner, ItemKeyData shareTo)
		throws BaseException
	{
		FormulaData formula = null;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			if (formula == null)
			{
				org.hibernate.Query query = null;
				if (rawDataType == null)
				{
					query = HibernateUtil.getPredefinedQuery(session, "GET_FORMULA_WITH_NAME_AND_CHANNELS");			
					query.setInteger("channels", channels);
				}
				else
				{
					query = HibernateUtil.getPredefinedQuery(session, "GET_FORMULA_WITH_NAME_AND_TYPE");
					query.setString("rawDataType", rawDataType);
				}
				query.setString("name", name);
				formula = HibernateUtil.loadData(FormulaData.class, query);
			}
			if (formula != null)
			{
				int sv = schemaVersion.getSchemaVersion();
				if (sv < 20)
				{
					// Add coloring information if upgrading
					formula.setColoring(coloring);
				}
				if (sv < 36)
				{
					// Add average method if upgrading
					formula.setAverageMethod(avgMethod.getValue());
				}
				if (sv < 70)
				{
					// Add intensity transformation information if upgrading
					formula.setSourceIntensityTransform(source == null ? null : source.getValue());
					formula.setResultIntensityTransform(result == null ? null : result.getValue());
					if (name.startsWith("Swap"))
					{
						// The "Swap intensities" transformation needs to be updated
						formula.getFormulas().clear();
						formula.getFormulas().addAll(Arrays.asList(formulas));
					}
				}
				if (sv < 77)
				{
					// Add value type information if upgrading
					formula.setValueType(valueType.getValue());
				}
				log.info("createFormula: EXISTS [name="+name+"; rawDataType="+rawDataType+"; channels="+channels+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				formula = new FormulaData();
				formula.setEntryDate(new Date());
				formula.setType(type.getValue());
				formula.setValueType(valueType.getValue());
				formula.setParser(parser.getValue());
				formula.setName(name);
				formula.setDescription(description);
				formula.setRawDataType(rawDataType);
				formula.setChannels(channels);
				formula.setOwner(owner);
				formula.setItemKey(shareTo);
				formula.getFormulas().addAll(Arrays.asList(formulas));
				formula.setColoring(coloring);
				formula.setAverageMethod(avgMethod.getValue());
				formula.setSourceIntensityTransform(source == null ? null : source.getValue());
				formula.setResultIntensityTransform(result == null ? null : result.getValue());

				HibernateUtil.saveData(session, formula);
				HibernateUtil.commit(tx);
				log.info("createFormula: OK [name="+name+"; rawDataType="+rawDataType+"; channels="+channels+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createFormula: FAILED [name="+name+"; rawDataType="+rawDataType+"; channels="+channels+"]", ex);
			throw ex;
		}
		return formula;
	}
	
	/**
		Create a {@link Platform}.
	*/
	private static PlatformData createPlatform(String externalId, String name, String description,
		boolean fileOnly, String rawDataType, int channels, PlatformFT... fileTypes)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		PlatformData platform = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
		
			if (platform == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
					"GET_PLATFORM_FOR_EXTERNAL_ID");
				query.setString("externalId", externalId);
				platform = HibernateUtil.loadData(PlatformData.class, query);
			}
			
			if (platform != null)
			{
				log.info("createPlatform: EXISTS [externalId="+externalId+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				platform = new PlatformData();
				platform.setExternalId(externalId);
				platform.setName(name);
				platform.setDescription(description);
				platform.setFileOnly(fileOnly);
				if (fileOnly)
				{
					platform.setChannels(channels);
					platform.setRawDataType("platform." + externalId);
				}
				else
				{
					platform.setRawDataType(rawDataType);
				}
				HibernateUtil.saveData(session, platform);
	
				if (fileTypes != null)
				{
					for (PlatformFT pft : fileTypes)
					{
						PlatformFileTypeData platformFile = new PlatformFileTypeData();
						platformFile.setPlatform(platform);
						platformFile.setDataFileType(pft.fileType);
						platformFile.setRequired(pft.required);
						HibernateUtil.saveData(session, platformFile);
					}
				}
				
				HibernateUtil.commit(tx);
				log.info("createPlatform: OK [externalId="+externalId+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createPlatform: FAILED [externalId="+externalId+"]", ex);
			throw ex;
		}
		return platform;
	}

	/**
		Create a {@link DataFileType}.
	*/
	private static DataFileTypeData createDataFileType(
		String externalId, String name, String description,
		Item itemType, String extension, FileTypeData genericType, 
		String validatorClass, String metadataReaderClass)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		DataFileTypeData fileType = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
		
			if (fileType == null)
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
					"GET_DATAFILETYPE_FOR_EXTERNAL_ID");
				query.setString("externalId", externalId);
				fileType = HibernateUtil.loadData(DataFileTypeData.class, query);
			}
			
			if (fileType != null)
			{
				log.info("createDataFileType: EXISTS [externalId="+externalId+"]");
				HibernateUtil.commit(tx);
			}
			else
			{
				fileType = new DataFileTypeData();
				fileType.setExternalId(externalId);
				fileType.setName(name);
				fileType.setDescription(description);
				fileType.setItemType(itemType.getValue());
				fileType.setExtension(extension);
				fileType.setGenericType(genericType);
				fileType.setValidatorClass(validatorClass);
				fileType.setMetadataReaderClass(metadataReaderClass);
	
				HibernateUtil.saveData(session, fileType);
				HibernateUtil.commit(tx);
				log.info("createDataFileType: OK [externalId="+externalId+"]");
			}
		}
		catch (BaseException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("createDataFileType: FAILED [externalId="+externalId+"]", ex);
			throw ex;
		}
		return fileType;
	}	
	
	private static void setPermissions(Item itemType, int itemId, ItemKeyData itemKey, ProjectKeyData projectKey)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			ShareableData data = (ShareableData)session.get(itemType.getDataClass(), itemId);
			data.setItemKey(itemKey);
			data.setProjectKey(projectKey);
			
			HibernateUtil.commit(tx);
			log.info("setPermissions: OK " + itemType.name() + "(" +  itemId + ")");
		}
		catch (RuntimeException ex)
		{
			if(tx != null) HibernateUtil.rollback(tx);
			log.error("setPermissions: FAILED " + itemType.name() + "(" +  itemId + ")");
			throw ex;
		}
	}
	
	private static class PlatformFT
	{
		private final DataFileTypeData fileType;
		private final boolean required;
		
		private PlatformFT(DataFileTypeData fileType, boolean required)
		{
			this.fileType = fileType;
			this.required = required;
		}
	}
	
}
