/**
	$Id: LogControl.java 5057 2009-08-18 12:46:43Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.lang.ref.WeakReference;

import net.sf.basedb.core.data.ChangeHistoryData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.log.TransactionDetails;

/**
	Gives information about the current transaction and gives log
	implementations limited access to the database.

	@author Nicklas
	@version 2.13
	@base.modified $Date: 2009-08-18 14:46:43 +0200 (Tue, 18 Aug 2009) $
*/
public class LogControl
{

	private final WeakReference<DbControl> dc;
	private TransactionDetails transactionDetails;
	private org.hibernate.StatelessSession session;
	
	LogControl(DbControl dc)
	{
		this.dc = new WeakReference<DbControl>(dc);
	}
	
	/**
		Get information about the current transaction.
		@return A TransactionDetails object
	*/
	public TransactionDetails getTransactionDetails()
	{
		if (transactionDetails == null) 
		{
			transactionDetails = new TransactionDetails(dc.get().getSessionControl());
		}
		return transactionDetails;
	}
	
	/**
		Add a main entry the change history table. Each transaction
		usually results in one such entry. Changes to individual entities
		are logged with {@link #log(ChangeHistoryDetailData)}.
		
		@param history A fully initialized ChangeHistoryData object
	*/
	public void log(ChangeHistoryData history)
	{
		getSession().insert(history);
	}
	
	/**
		Adds information about a modified entity to the change history
		table. 		
		@param detail A fully initialized ChangeHistoryDetailData object
			including a link to {@link ChangeHistoryData} that has 
			been logged with {@link #log(ChangeHistoryData)}
	*/
	public void log(ChangeHistoryDetailData detail)
	{
		getSession().insert(detail);
	}
	
	/**
		Creates a Hibernate HQL query. Use this to get information
		from the database that is relevent for the logging implementation.
		The query uses the stateless session, which means that there
		is no first- or second-level cache and on support for initialization
		of proxies. We recommend that queries are used to return
		scalar values only, not entities.
		@return A Query
	*/
	public org.hibernate.Query createHqlQuery(String hql)
	{
		return HibernateUtil.createQuery(getSession(), hql);
	}
	
	/**
		Creates a Hibernate SQL query. Use this to get information
		from the database that is relevent for the logging 
		implementation. The query uses the stateless session.
		@return A query
	*/
	public org.hibernate.Query createSqlQuery(String sql)
	{
		return HibernateUtil.createSqlQuery(session, sql);
	}
	
	private org.hibernate.StatelessSession getSession()
	{
		if (session == null) session = dc.get().getStatelessSession();
		return session;
	}
	
	
}
