/*
	$Id: ShareableUtil.java 5060 2009-08-19 07:02:11Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.LinkedList;
import java.util.List;

import net.sf.basedb.core.data.GroupData;
import net.sf.basedb.core.data.ItemKeyData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.signal.ThreadSignalHandler;

/**
	Utility methods that will make it easier to implement
	the {@link Shareable} interface, including data validation.
	
	@author Nicklas
	@version 2.0
	@see Shareable
	@base.modified $Date: 2009-08-19 09:02:11 +0200 (Wed, 19 Aug 2009) $
*/
public class ShareableUtil
{

	/**
		Set the item key of a {@link ShareableData} object.
	 	@param shareableData The data object to set the item key on. Null is not allowed. 
	 	@param itemKey Item key to be set.
	 	@deprecated Use {@link #setItemKey(DbControl, ShareableData, ItemKey)} instead
	*/
	public static void setItemKey(ShareableData shareableData, ItemKey itemKey)
	{
		shareableData.setItemKey(itemKey == null ? null : itemKey.getData());
	}
	
	/**
		Set the project key of a {@link ShareableData} object.
	 	@param shareableData The data object to set project key to. Null is not allowed. 
	 	@param projectKey Project key to be set.
	*/
	public static void setProjectKey(ShareableData shareableData, ProjectKey projectKey)
	{
		shareableData.setProjectKey(projectKey == null ? null : projectKey.getData());
	}
	
	/**
		Set the item key of a {@link ShareableData} object.
		If the logged in user doesn't have {@link Permission#SHARE_TO_EVERYONE} permission
		then new item key must, if it includes sharing to the everyone group, have exactly
		the same permissions as there is in the existing item key.
		
		@param dc A DbControl
		@param shareableData The item to share
		@param itemKey The new item key
		@throws PermissionDeniedException If the logged in user doesn't have permission 
			and the new item key contains a permission that is different from the
			existing permission
		@since 2.6
	*/
	public static void setItemKey(DbControl dc, ShareableData shareableData, ItemKey itemKey)
	{
		if (itemKey != null && !dc.getSessionControl().hasSystemPermission(Permission.SHARE_TO_EVERYONE))
		{
			GroupData everyone = HibernateUtil.loadData(dc.getHibernateSession(), GroupData.class, 
				SystemItems.getId(Group.EVERYONE));
			
			// Check permissions for Everyone in the new key
			ItemKeyData newKey = itemKey.getData();
			int newPermission = IntegerUtil.getInt(newKey.getGroups().get(everyone));
			if (newPermission != 0)
			{
				// Check permission for Everyone in the old key
				ItemKeyData oldKey = shareableData.getItemKey();
				int oldPermission = IntegerUtil.getInt(oldKey == null ? 0 : oldKey.getGroups().get(everyone));
				if (newPermission != oldPermission)
				{
					throw new PermissionDeniedException("Not allowed to change permissions for the to the EVERYONE group.");
				}
			}
		}
		shareableData.setItemKey(itemKey == null ? null : itemKey.getData());
	}
	
	/**
		Share all files and sub-directories in the given directory with the
		same permissions. This method will of course only change the
		permissions on items that the current user has {@link Permission#SET_PERMISSION}
		on.

		@param dc A DbControl for database access
		@param directory The directory
		@return The number of files and directories processed
		@since 2.8
	*/
	public static int shareRecursively(DbControl dc, Directory directory)
	{
		ItemKeyData itemKey = directory.getData().getItemKey();
		ProjectKeyData projectKey = directory.getData().getProjectKey();

		List<Directory> toProcess = new LinkedList<Directory>();
		toProcess.add(directory);
		int numItems = 0;
		
		while (toProcess.size() > 0)
		{
			ThreadSignalHandler.checkInterrupted();
			Directory currentDir = toProcess.remove(0);
			
			// Share files
			ItemQuery<File> fileQuery = currentDir.getFiles();
			fileQuery.include(Include.ALL);
			fileQuery.setItemPermission(Permission.SET_PERMISSION);
			for (File f : fileQuery.list(dc))
			{
				f.getData().setItemKey(itemKey);
				f.getData().setProjectKey(projectKey);
				numItems++;
			}
			
			// Share directories
			ItemQuery<Directory> dirQuery = currentDir.getSubDirectories();
			dirQuery.include(Include.ALL);
			dirQuery.setItemPermission(Permission.SET_PERMISSION);
			for (Directory d : dirQuery.list(dc))
			{
				d.getData().setItemKey(itemKey);
				d.getData().setProjectKey(projectKey);
				numItems++;
				
				// Make sure that we go into this subdirectory
				toProcess.add(d);
			}
		}
		return numItems;
	}
	
	/**
		Copy permissions from one item to another. If the 'from' item
		is shared to the 'EVERYONE' group, but the logged in user lacks
		the {@link Permission#SHARE_TO_EVERYONE} permission, the share to
		everyone is removed. If the 'to' item is a new item, a project 
		is active and the 'from' item isn't already shared to the project,
		the 'to' item will be shared to the project.

		@param from The item to copy permissions from
		@param to The item to copy permissions to
		@since 2.8
	*/
	public static void copyPermissions(DbControl dc, Shareable from, Shareable to)
	{
		SessionControl sc = dc.getSessionControl();
		
		BasicItem fromItem = (BasicItem)from;
		if (fromItem.isDetached()) dc.reattachItem(fromItem, false);
		
		ItemKey itemKey = from.getItemKey();
		ProjectKey projectKey = from.getProjectKey();
		
		// Check if we need to remove permissions for the EVERYONE group
		if (itemKey != null && !sc.hasSystemPermission(Permission.SHARE_TO_EVERYONE))
		{
			GroupData everyone = HibernateUtil.loadData(dc.getHibernateSession(), GroupData.class, 
				SystemItems.getId(Group.EVERYONE));

			if (itemKey.getData().getGroups().containsKey(everyone))
			{
				// Remove the EVERYONE grup
				GroupPermissions gp = new GroupPermissions(itemKey);
				gp.getGroupPermissions().remove(everyone);
				
				UserPermissions up = new UserPermissions(itemKey);
				if (up.size() == 0 && gp.size() == 0)
				{
					itemKey = null;
				}
				else
				{
					itemKey = ItemKey.getNewOrExisting(dc, up, gp);
				}
			}
		}

		// Check if we need to add permissions for the current project
		if (projectKey != null && sc.getProjectKeyId() != 0 && to.getId() == 0)
		{
			ProjectData project = HibernateUtil.loadData(dc.getHibernateSession(), 
					ProjectData.class, sc.getActiveProjectId());
			
			if (!projectKey.getData().getProjects().containsKey(project))
			{
				// Add the current project
				ProjectKey activeKey = ProjectKey.getById(dc, sc.getProjectKeyId());
				ProjectPermissions pp = new ProjectPermissions(projectKey);
				pp.getProjectPermissions().put(project, activeKey.getData().getProjects().get(project));
				projectKey = ProjectKey.getNewOrExisting(dc, pp);
			}
		}
		
		if (itemKey != null) to.setItemKey(itemKey);
		if (projectKey != null) to.setProjectKey(projectKey);
	}
	
}
