/*
	$Id: SpotBatcher.java 4992 2009-06-23 09:47:46Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.hibernate.ExecuteUpdateWork;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Selects;

import java.util.List;
import java.util.LinkedList;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Types;

/**
	Batcher class for per-spot intensity values for bioassaysets. This batcher 
	is normally used when a new bioassayset has been created that uses a different 
	cube and/or layer to store it's data than the source bioassayset.
	<p>
	Code example: Swap the intensities of channel 1 and 2
<pre class="code">
// Get source bioassayset
DbControl dc = ...
BioAssaySet source = ...

// Create new transformation and bioassayset to hold the
// new intensities, we use the same data cube as the source
// but a different layer
Transformation t = source.newTransformation();
dc.saveItem(t);
BioAssaySet bas = t.newProduct(null, "new", true);
dc.saveItem(bas);

// Create a query that selects the data of
// the source bioassayset
DynamicSpotQuery query = source.getSpotData();

// Execute the query, and use the batcher to swap the
// intensities
SpotBatcher batcher = bas.getSpotBatcher();
DynamicResultIterator result = query.iterate(dc);
int column = result.getIndex(VirtualColumn.COLUMN.getName());
int position = result.getIndex(VirtualColumn.POSITION.getName());
int ch1 = result.getIndex(VirtualColumn.channel(1).getName());
int ch2 = result.getIndex(VirtualColumn.channel(2).getName());
while (result.hasNext())
{
   SqlResult r = result.next();
   batcher.insert(r.getShort(column), r.getInt(position), 
      r.getFloat(ch2), r.getFloat(ch1));
}

// Close and commit
batcher.close();
dc.commit();
</pre>

	@author Nicklas
	@version 2.0
	@see BioAssaySet#getSpotBatcher()
	@base.modified $Date: 2009-06-23 11:47:46 +0200 (Tue, 23 Jun 2009) $
*/
public class SpotBatcher
	extends AbstractBatcher
{

	/**
		The bioassayset this batcher inserts data for.
	*/
	private final BioAssaySet bioAssaySet;
	
	/**
		The number of channels used in the raw data type of the 
		experiment where the bioassayset belongs.
	*/
	private final int numChannels;
	
	/**
		The number of queued inserts. Used to check when a flush
		is needed.
	*/
	private int queuedInsertCount = 0;
	
	/**
		The total number of inserts done by this batcher.
	*/
	private int totalInsertCount = 0;

	/**
		The number of bytes this batcher has added to the experiment.
	*/
	private long bytes = 0;

	/**
		The number of bytes used by a single row.
	*/
	private final long bytesPerRow;

	/**
		The statement that inserts spot data.
<pre class="code">
INSERT INTO Dynamic#PerSpot (cube, layer, column, position, ch1, ch2, ...)
VALUES (cube, layer, ?, ?, ?, ?, ...)
</pre>
	*/
	private PreparedStatement spotSql = null;

	/**
	 	The SQL string for the {@link #spotSql} statement.
	*/
	private String spotSqlStatement;
	
	/**
		Create a new spot batcher for a bioassayset. It is assumed that
		the calling code has checked that the bioassayset is allowed
		to receive more data (ie. it's data cube layer hasn't been committed 
		to the database).
	*/
	SpotBatcher(DbControl dc, BioAssaySet bioAssaySet)
	{
		super();
		this.bioAssaySet = bioAssaySet;
		this.numChannels = bioAssaySet.getRawDataType().getChannels();
		this.bytesPerRow = VirtualTable.SPOT.getBytesPerRow(bioAssaySet.getRawDataType());
		setDbControl(dc);
	}

	/*
		From the Batcher interface
		-------------------------------------------
	*/
	public void flush()
		throws BaseException
	{
		try
		{
			if (spotSql != null) spotSql.executeBatch();
		}
		catch (SQLException ex)
		{
			logSql.error(spotSqlStatement, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
		queuedInsertCount = 0;
	}
	// -------------------------------------------
	/*
		From the AbstractBatcher class
		-------------------------------------------
	*/
	/**
		Update disk usage and close open SQL statements
	*/
	@Override 
	void onBeforeClose()
		throws BaseException
	{
		bioAssaySet.getDataCubeLayer().addBytes(bytes);
		try
		{
			if (spotSql != null) 
			{
				spotSql.close();
				spotSql = null;
			}
		}
		catch (SQLException ex)
		{
			logSql.error(ex.getMessage(), ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
		}
		if (totalInsertCount > 0 && !getDbControl().isRolledBack())
		{
			String catalog = Application.getDynamicCatalog();
			String schema = Application.getDynamicSchema();
			String table = VirtualTable.SPOT.getTableName(bioAssaySet.getVirtualDb());
			analyzeTable(catalog, schema, table);
		}
	}
	// -------------------------------------------

	/**
		Insert intensities for a given column/position coordinate.
		@param column The column coordinate
		@param position The position coordinate
		@param intensities An array holding the intensities for the coordinate,
			the length must match the number of channels of the raw data type
		@throws InvalidDataException If the intensities array is null or
			of the incorrect length
		@throws BaseException If there is another error
	*/
	public void insert(short column, int position, float... intensities)
		throws BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		updateLastAccess();
		if (intensities == null || intensities.length != numChannels)
		{
			throw new InvalidDataException("Incorrect number of intensities for raw data type. Expected "+
				numChannels+"; found "+(intensities == null ? 0 : intensities.length));
		}
		try
		{
			if (spotSql == null) buildSpotSql();
			spotSql.setInt(1, column);
			spotSql.setInt(2, position);
			int i = 3;
			for (float intensity : intensities)
			{
				if (Float.isNaN(intensity))
				{
					spotSql.setNull(i, Types.FLOAT);
				}
				else
				{
					spotSql.setFloat(i, intensity);
				}
				i++;
			}
			spotSql.addBatch();
			bytes += bytesPerRow;
			totalInsertCount++;
		}
		catch (SQLException ex)
		{
			logSql.error(spotSqlStatement, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
		if (++queuedInsertCount == getBatchSize())
		{
			flush();
		}
	}
	
	/**
		Insert spot intensities using a query. The query MUST select data in
		the following order: column, position, ch1, ch2, ...
		<p>
		Internally this method converts the query into a 
		<code>INSERT INTO ... SELECT ...</code> query, making it very fast,
		since the data for each spot doesn't have to be retreived, processed and
		then inserted one row at a time. It is possible to call
		this method multiple times with the same or different query. 
		<p>
		Here is an alternative to the example above:
		<pre class="code">
// Get source bioassayset
DbControl dc = ...
BioAssaySet source = ...

// Create new transformation and bioassayset to hold the
// new intensities, we use the same data cube as the source
// but a different layer
Transformation t = source.newTransformation();
dc.saveItem(t);
BioAssaySet bas = t.newProduct(null, "new", true);
dc.saveItem(bas);

// Create a query that selects the data of
// the source bioassayset, swap the order of the channels
DynamicSpotQuery query = source.getSpotData();
query.select(Dynamic.select(VirtualColumn.COLUMN));
query.select(Dynamic.select(VirtualColumn.POSITION));
query.select(Dynamic.select(VirtualColumn.channel(2)));
query.select(Dynamic.select(VirtualColumn.channel(1)));

// Execute the query, and use the batcher to insert the result
SpotBatcher batcher = bas.getSpotBatcher();
batcher.insert(query);

// Close and commit
batcher.close();
dc.commit();
</pre>
		
		@param query The query that selects the data to be inserted
		@return int The number of spots inserted
		@throws BaseException If there is an error
	*/
	public int insert(final AbstractSqlQuery query)
		throws BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		updateLastAccess();
		if (query == null) throw new InvalidUseOfNullException("query");
		
		query.resetTemporary();
		
		// Add static values cubeNo and layerNo to the selection
		query.temporarySelect(
			Selects.expression(
				Expressions.integer(bioAssaySet.getDataCubeNo()), 
				VirtualColumn.CUBE.getName()
			)
		);
		query.temporarySelect(
			Selects.expression(
				Expressions.integer(bioAssaySet.getDataCubeLayerNo()), 
				VirtualColumn.LAYER.getName()
			)
		);
		
		String selectSql = query.getMainQuery(getDbControl(), true);
		
		// Holds the names of the parameters in the query in the order they appear
		final List<String> parameterOrder = new LinkedList<String>();
		
		// Parse out named parameters
		selectSql = query.parseParameters(selectSql, parameterOrder);
		String insertSql = buildInsertSelectSql(selectSql);

		int rowsInserted = 0;
		try
		{
			rowsInserted = HibernateUtil.doJdbcWork(getDbControl().getHibernateSession(), 
				new ExecuteUpdateWork(getSessionControl(), insertSql)
				{
					@Override
					public void setParameters(PreparedStatement ps)
						throws SQLException
					{
						query.setParameters(ps, parameterOrder);
					}
				}
			);
			bytes += rowsInserted * bytesPerRow;
			totalInsertCount += rowsInserted;
		}
		catch (SQLException ex)
		{
			logSql.error(insertSql, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
		return rowsInserted;
	}
	
	/**
		Builds the insert SQL statement. Also creates the Dynamic#PerSpot table
		if it doesn't exists.
		@see BatchUtil#buildInsertSql(VirtualDb, VirtualTable, VirtualColumn[], Object[])
	*/
	private void buildSpotSql()
		throws SQLException, BaseException
	{
		RawDataType rdt = bioAssaySet.getRawDataType();
		VirtualDb vdb = bioAssaySet.getVirtualDb();
		if (!vdb.hasTable(VirtualTable.SPOT))
		{
			vdb.createTables(VirtualTable.SPOT);
		}
		short cube = bioAssaySet.getDataCubeNo();
		short layer = bioAssaySet.getDataCubeLayerNo();
		spotSqlStatement =  BatchUtil.buildInsertSql(
			vdb,
			VirtualTable.SPOT,
			VirtualTable.SPOT.getColumns(rdt), 
			new Object[] {cube, layer}
		);
		spotSql = HibernateUtil.getConnection(getDbControl().getHibernateSession()).prepareStatement(spotSqlStatement);
	}
	
	/**
		Build the INSERT INTO ... SELECT ... statment. Also creates the Dynamic#PerSpot table
		if it doesn't exists.
		@param selectSql The SELECT part of the query
		@see BatchUtil#buildInsertSelectSql(VirtualDb, VirtualTable, VirtualColumn[], String)
	*/
	private String buildInsertSelectSql(String selectSql)
		throws BaseException
	{
		RawDataType rdt = bioAssaySet.getRawDataType();
		VirtualDb vdb = bioAssaySet.getVirtualDb();
		if (!vdb.hasTable(VirtualTable.SPOT))
		{
			vdb.createTables(VirtualTable.SPOT);
		}
		String sql = BatchUtil.buildInsertSelectSql(vdb, VirtualTable.SPOT, VirtualTable.SPOT.getColumns(rdt), selectSql);
		return sql;
	}
	
}
