/*
	$Id: StringUtil.java 5283 2010-03-24 10:42:29Z nicklas $

	Copyright (C) 2005, 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;

/**
	Utility methods that will make it easier to implement
	data validation for string attributes.
	
	@author Nicklas
	@version 2.0
*/
public class StringUtil
{

	/**
		Trim and check the length of a string. Null values are allowed.

		@param value The string to check
		@param name The name of the attribute to use if an exception is thrown
		@param maxLength The maximum length of the string
		@return The original string, trimmed from white-space at the 
			beginning and end
		@throws StringTooLongException If the string is too long
	*/
	public static String setNullableString(String value, String name, int maxLength)
		throws StringTooLongException
	{
		if (value == null) return null;
		value = value.trim();
		if (value.length() > maxLength) throw new StringTooLongException(name, value, maxLength);
		return value;
	}
	
	/**
		Trim and check the length of a string. Null values are not allowed.

		@param value The string to check
		@param name The name of the attribute to use if an exception is thrown
		@param maxLength The maximum length of the string
		@return The original string, trimmed from white-space at the 
			beginning and end
		@throws StringTooLongException If the string is too long
		@throws InvalidUseOfNullException If the string is null
	*/
	public static String setNotNullString(String value, String name, int maxLength)
		throws InvalidUseOfNullException, StringTooLongException
	{
		if (value == null) throw new InvalidUseOfNullException(name);
		value = value.trim();
		if (value.length() > maxLength) throw new StringTooLongException(name, value, maxLength);
		return value;
	}
	
	/**
		Check if two strings are equal or both are null.
		@param s1 The first string
		@param s2 The second string
		@return TRUE if the strings are equal or both are null, FALSE otherwise
	*/
	public static boolean isEqualOrNull(String s1, String s2)
	{
		return (s1 == s2) || (s1 != null &&  s1.equals(s2));
	}
	
	/**
		Check if two strings are equal or both are null.
		@param s1 The first string
		@param s2 The second string
		@param caseInsensitive TRUE if the comparison should ignore case, FALSE
			otherwise
		@return TRUE if the strings are equal or both are null, FALSE otherwise
		@since 2.4
	*/
	public static boolean isEqualOrNull(String s1, String s2, boolean caseInsensitive)
	{
		if (!caseInsensitive)
		{
			return isEqualOrNull(s1, s2);
		}
		else
		{
			return (s1 == s2) || (s1 != null &&  s1.equalsIgnoreCase(s2));
		}
	}
	
	/**
	 	Join a collection of values to a string, using the specified delimiator.
		@param values The values to join
		@param deliminator The deliminator between the values
		@param skipNull If null values should be skipper or not
		@return The resulting string
	*/
	public static final String join(Collection<?> values, String deliminator, boolean skipNull)
	{
		StringBuilder sb = new StringBuilder();
		boolean firstElement = true;
		if (values != null)
		{
			for (Object value : values)
			{
				if (value == null)
				{					
					if (skipNull)
					{
						/* #### CONTINUE-STATEMENT #### */
						continue;
					}
					else
					{
						value = "";
					}
				}
				if (!firstElement) sb.append(deliminator);
				sb.append(value);
				firstElement = false;
			}
		}
		return sb.toString();
	}

	/**
		Trim a string to a maximum length. String longer than the specified length
		will be cut off and three dots are appended. If no trimming is needed the
		original string is returned.
		@param s The string to trim
		@param length The maximum length of the string
		@return The trimmed string
	*/
	public static final String trimString(String s, int length)
	{
		if (s != null && s.length() > length)
		{
			s = s.substring(0, length - 3) + "...";
		}
		return s;
	}
	
	/**
		Trim a string to a maximum length. Strings longer than the specified 
		length will have text cut-out from the middle and replaced by three 
		dots. If no trimming is needed the original string is returned.
		@param s The string to trim
		@param length The maximum length of the string
		@return The trimmed string
		@since 2.14
	*/
	public static final String trimStringMiddle(String s, int length)
	{
		if (s != null && s.length() > length)
		{
			int charactersToCut = s.length() + 3 - length;
			int beginAt = (s.length() - charactersToCut) / 2;
			s = s.substring(0, beginAt) + "..." + s.substring(beginAt + charactersToCut);
		}
		return s;
	}
	
	/**
		Tokenize a string by splitting it at positions matching the
		regular expression, optionally ignoring matches that are inside
		quotes.
		@param s The string to tokenize
		@param regexp The regular expression for matching split sites
		@return An array with the elements of the array
		@since 2.15
	*/
	public static final String[] tokenize(String s, String regexp, boolean notInsideQoutes)
	{
		String[] result = null;
		if (s != null && regexp != null)
		{
			if (notInsideQoutes) regexp += "(?=(?:[^\"]*\"[^\"]*\")*(?![^\"]*\"))";
			result = s.split(regexp);
		}
		return result;
	}
	

}
