/**
	$Id: ExecuteUpdateWork.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;

import net.sf.basedb.core.QueryExecutor;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.signal.SignalException;

/**
	A JDBC work implementation that executes an update/insert/delete 
	query and returns the number of rows that was affected by the query.
	By subclassing this class and overridding the {@link #setParameters(PreparedStatement)}
	method it is possible to set parameters on the prepared statement before 
	the SQL is executed.

	@author nicklas
	@version 2.9
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
	@see PreparedStatement#executeUpdate()
*/
public class ExecuteUpdateWork
	implements JdbcWork<Integer>
{

	private SessionControl sc;
	private String sql;
	private int affectedRows;
	
	/**
		Create a new work item. 
		@param sc The session control to keep alive during the execution
			of the sql, or null
		@param sql The sql to execute
	 */
	public ExecuteUpdateWork(SessionControl sc, String sql)
	{
		this.sc = sc;
		this.sql = sql;
	}
	
	/**
		From the Work interface
		-----------------------
	*/
	@Override
	public void execute(Connection connection) 
		throws SQLException
	{
		PreparedStatement ps = null;
		try
		{
			ps = connection.prepareStatement(sql);
			setParameters(ps);
			affectedRows = QueryExecutor.executeUpdate(ps, sc); 
		}
		catch (InterruptedException ex)
		{
			throw new SignalException("Aborted by user", ex);
		}
		finally
		{
			try
			{
				if (ps != null) ps.close();
			}
			catch (Throwable t)
			{}
		}
	}
	// --------------------------
	
	/*
		From the JdbcWork interface
		---------------------------
	*/
	/**
		@return The number of rows affected by the update/insert/delete.
	*/
	@Override
	public Integer getResult()
	{
		return affectedRows;
	}
	// ----------------------------

	/**
		This method should be overridden by subclasses that needs to
		set parameters on the statement that is going to be executed.
		The default implementation does nothing. This method is called
		after the statement has been prepared with 
		{@link Connection#prepareStatement(String)} and before it is executed
		with {@link PreparedStatement#executeUpdate()}.
		@param ps The prepared statement
	*/
	protected void setParameters(PreparedStatement ps)
		throws SQLException
	{}

}
