/**
	$Id: PooledParentsTransformer.java 4723 2009-01-13 08:20:26Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.biomaterial;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.collections.BasicItemToIdTransformer;
import net.sf.basedb.util.collections.CollectionTransformer;

/**
	Collection transformer that given a collection of (pooled) 
	biomaterials finds all parents (of the same type).

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-01-13 09:20:26 +0100 (Tue, 13 Jan 2009) $
*/
public class PooledParentsTransformer<I extends MeasuredBioMaterial>
	implements CollectionTransformer<I, I>
{

	private final DbControl dc;
	private final ItemQuery<I> query;
	private final boolean copySources;
	
	public PooledParentsTransformer(DbControl dc, boolean copySources, ItemQuery<I> query)
	{
		this.dc = dc;
		this.copySources = copySources;
		this.query = query;
		query.include(Include.ALL);
		query.join(Hql.innerJoin("sourceEvents", "se"));
		query.join(Hql.innerJoin("se", "event", "evt"));
		query.restrict(
			Restrictions.in(
				Hql.property("evt", "bioMaterial"), 
				Expressions.parameter("sources")
			)
		);
	}
	
	/*
		From the CollectionTransformer interface
		----------------------------------------
	*/
	/**
		
	*/
	@Override
	public int transform(Collection<? extends I> source,
			Collection<? super I> destination)
	{
		// Copy source ID:s so we can use them as a parameter in the query
		Set<Integer> sourceIds = new HashSet<Integer>();
		BasicItemToIdTransformer idTransformer = new BasicItemToIdTransformer();
		idTransformer.transform(source, sourceIds);
		int originalSize = destination.size();
		
		// Copy source biomaterials
		if (copySources) destination.addAll(source);
		
		// We need to loop so we can move down as many levels as we need
		while (sourceIds.size() > 0)
		{
			query.setParameter("sources", sourceIds, Type.INT);
			List<I> parents = query.list(dc);
			// Add the parents to the destination; if they all already exists we break to 
			// loop (to avoid problematic circular references)
			if (!destination.addAll(parents)) break;
			sourceIds.clear();
			// Next iteration will find the parents of the parents, and so on...
			idTransformer.transform(parents, sourceIds);
		}
		return destination.size() - originalSize;
	}
	// --------------------------------------------------

}
