/**
	$Id: BfsExporter.java 5309 2010-04-13 18:22:28Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.DatabaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicPositionQuery;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.util.ChainedProgressReporter;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.bfs.AnnotationWriter;
import net.sf.basedb.util.bfs.DataWriter;
import net.sf.basedb.util.bfs.DataWriterFactory;
import net.sf.basedb.util.bfs.MetadataWriter;
import net.sf.basedb.util.bfs.MetadataModel.SectionEntry;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

/**
	Base class for exporting spotdata to BFS format. This class will
	do most of the hard work. A subclass is required to:
	<ul>
	<li>
	<li>create a list of data writers that will output the actual 
		spot data (see {@link #createDataWriters()}).  
	<li>rearrange the data that is loaded by the query to the 
		correct data writer (see {@link #arrangeData(int, short, Object[][], Object[])}).
	</ul>
	
	<p>
	Before the exporter can be used it must be configured. The most important options are:
	<ul>
	<li>{@link #setDbControl(DbControl)} (required): Sets the DbControl that is used for database access.
	<li>{@link #setSource(net.sf.basedb.core.BioAssaySet)} (required): Sets the bioassay set that data should
		be exported from.
	<li>{@link #setMetadataWriter(MetadataWriter)} (required): Sets the writer stream were BFS metadata is
		written.
	<li>{@link #setReporterAnnotationsWriter(AnnotationWriter)} and {@link #addReporterField(DynamicField)}
		(optional): Sets the writer stream were reporter annoatations are written and selects which 
		reporter annotations to export. If no writer is specified the reporter annotations file is not
		created.
	<li>{@link #setAssayAnnotationsWriter(AnnotationWriter)} and {@link #addAssayField(AssayField)}
		(optional): Sets the writer stream were assay annotations is written and selects which
		assay annotations to export. If no writer is specified the assay annotations file is not
		created.
	<li>{@link #setDataWriterFactory(DataWriterFactory)} and {@link #addSpotField(DynamicField)} 
		(optional): Sets the factory that should be used to create output files and selects which
		spot data to export. The number of files and the data arrangement depends on the subtype.
	<li>{@link #setParameter(String, String...)} (optional): Additional parameters that are needed
		by, for example, a plug-in that is about to process the exported data files.
	</ul>
	
	
	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-04-13 20:22:28 +0200 (Tue, 13 Apr 2010) $
*/
public abstract class BfsExporter
	extends AbstractBioAssaySetExporter
{

	private boolean autoCloseWriters;
	private DataWriterFactory dataWriterFactory;
	private Map<String, String[]> parameters;
	private MetadataWriter metadataWriter;
	private AnnotationWriter reporterWriter;
	private AnnotationWriter assayWriter;
	private List<DataWriter> dataWriters;
	private List<SectionEntry> extraFiles;
	private ChainedProgressReporter chainedProgress;
	private long spotCount;
	private Set<Integer> exportedPositions;
	
	protected BfsExporter()
	{
		this.parameters = new LinkedHashMap<String, String[]>();
		this.extraFiles = new ArrayList<SectionEntry>();
		this.autoCloseWriters = true;
	}
	
	/*
		Configuration properties
		------------------------
	*/
	/**
		If this option is set then all writers are automatically closed
		when all data has been writted to them. This setting is enabled by default.
	*/
	public void setAutoCloseWriters(boolean autoClose)
	{
		this.autoCloseWriters = autoClose;
	}
	
	/**
		Set the stream were the metadata should be written. This 
		stream is required before it is possible to start the 
		export. It is expected that the given writer is a fresh writer 
		and that no data has been written to it yet.
	*/
	public void setMetadataWriter(MetadataWriter metadataWriter)
	{
		this.metadataWriter = metadataWriter;
	}
	
	/**
		Set the stream were reporter annotations should be written. This is optional
		and if not given, no reporter annotations file is generated. The
		writer must return a filename from {@link AnnotationWriter#getFilename()}
		since this is required in the metadata file. 
		<p>
		Use {@link #addReporterField(DynamicField)} to register reporter
		fields to export. The position number is always exported.
		
		@param reporterWriter An annotation writer, or null if no reporter
			annotation file should be created
	*/
	public void setReporterAnnotationsWriter(AnnotationWriter reporterWriter)
	{
		this.reporterWriter = reporterWriter;
	}
	
	/**
		Set the stream were assay annotations should be written. This is optional
		and if not given, no assay annotations file is generated. The
		writer must return a filename from {@link AnnotationWriter#getFilename()}
		since this is required in the metadata file. 
		<p>
		Use {@link #addAssayField(AssayField)} to register fields to export. 
		The assay id is always exported.
		
		@param assayWriter An annotation writer, or null if no assay
			annotations file should be created
	*/
	public void setAssayAnnotationsWriter(AnnotationWriter assayWriter)
	{
		this.assayWriter = assayWriter;
	}
	
	/**
		Set the data writer factory to use for generating spot data files.
		What files that are required depends on the BFS subtype. Eg. 
		the serial format uses a file per assay and the matrix format
		uses a file per spot field. A data writer factory is optional
		and if not given, no spot data is exported.
	
		@param dataWriterFactory A data writer factory, or null if no spot
			data files should be created
	*/
	public void setDataWriterFactory(DataWriterFactory<?> dataWriterFactory)
	{
		this.dataWriterFactory = dataWriterFactory;
	}
	
	/**
		Add a parameter that is exported in the [parameters] section. 
		There can be one or more values for a given key, which are then
		separated by tab characters. The parameters are written to the file 
		in the order they are registered by this method.
		<p>
		NOTE! Calling this method twice with the same key, overwrites the
		existing parameters, but doesn't change the order they are written
		to the file.
		
		@param key The parameter key (if null this method call is ignored)
		@param values The parameter value (may be null)
	*/
	public void setParameter(String key, String... values)
	{
		if (key == null) return;
		parameters.put(key, values);
	}
	
	/**
		Get the current value of a parameter.
		@param key The parameter key
		@return The value, or null if no values has been registered for
			the key (or if the registered value is null)
	*/
	public String[] getParameter(String key)
	{
		return parameters.get(key);
	}


	@Override
	public void addReporterField(DynamicField field)
	{
		super.addReporterField(field);
	}

	/**
		Add multiple reporter fields in one go. A null collection is ignored and
		so are null values in the collection.
		@param fields A collection with the fields that should be added
	*/
	public void addReporterFields(Collection<? extends DynamicField> fields)
	{
		if (fields == null) return;
		for (DynamicField f : fields)
		{
			 addReporterField(f);
		}
	}
	
	@Override
	public List<DynamicField> getReporterFields()
	{
		return super.getReporterFields();
	}
	
	@Override
	public void addAssayField(AssayField field)
	{
		super.addAssayField(field);
	}
	
	/**
		Add multiple assay fields in one go. A null collection is ignored and
		so are null values in the collection.
		@param fields A collection with the fields that should be added
	*/
	public void addAssayFields(Collection<? extends AssayField> fields)
	{
		if (fields == null) return;
		for (AssayField f : fields)
		{
			 addAssayField(f);
		}
	}
	
	@Override
	public List<AssayField> getAssayFields()
	{
		return super.getAssayFields();
	}
	
	/**
		Add information about a spot field that should be exported.
		@param spotField The spot field information
	*/
	@Override
	public void addSpotField(DynamicField spotField)
	{
		super.addSpotField(spotField);
	}

	/**
		Add multiple spot fields in one go. A null collection is ignored and
		so are null values in the collection.
		@param fields A collection with the fields that should be added
	*/
	public void addSpotFields(Collection<? extends DynamicField> fields)
	{
		if (fields == null) return;
		for (DynamicField f : fields)
		{
			 addSpotField(f);
		}
	}

	@Override
	public List<DynamicField> getSpotFields()
	{
		return super.getSpotFields();
	}

	/**
		Add information about an extra data file that is included
		as part of the BFS. This exporter will register this information
		as an entry in the [files] section in the metadata file.
		It will not check that the file actually exists or do any
		other action with the file. 
		<p>
		File entries are written in the order they are registered by
		this method. Multiple entries with the same key are allowed
		(eg. the previous entry is not replaced).
		<p>

		@param key The key for the file
		@param filename The name of the file
	*/
	public void addExtraFile(String key, String filename)
	{
		extraFiles.add(new SectionEntry(key, filename));
	}
	
	// --------------------------------------

	/*
		From the AbstractBioAssaySetExporter class
		------------------------------------------
	*/
	@Override
	protected void validate()
	{
		super.validate();
		if (getMetadataWriter() == null) throw new NullPointerException("getMetadataWriter()");
		AnnotationWriter aw = getReporterAnnotationsWriter();
		if (aw != null && aw.getFilename() == null)
		{
			throw new NullPointerException("getReporterAnnotationsWriter().getFilename()");
		}
		aw = getAssayAnnotationsWriter();
		if (aw != null && aw.getFilename() == null)
		{
			throw new NullPointerException("getAssayAnnotationsWriter().getFilename()");
		}
	}
	
	/**
		Validates that all required properties has been set and
		initialises other things that are needed.
	*/
	@Override
	protected void beginExport()
	{
		super.beginExport();
		setProgress(0, "Initializing...");
		ProgressReporter p = getProgressReporter();
		if (p != null) chainedProgress = new ChainedProgressReporter(p);
		try
		{
			if (getDataWriterFactory() != null) dataWriters = createDataWriters();
			if (dataWriters == null) dataWriters = Collections.emptyList();
		}
		catch (IOException ex)
		{
			throw new RuntimeException(ex);
		}
	}
	
	/**
		Export the metadata and assay annotations.
	*/
	@Override
	protected boolean exportGlobalHeader()
	{
		exportMetadata();
		exportAssayAnnotations();
		return true;
	}
	
	/**
		Export all spot data.
	*/
	@Override
	protected void exportSectionData()
	{
		if (chainedProgress != null) chainedProgress.setRange(15, 80);
		spotCount = exportSpotData(chainedProgress, getSource().getNumSpots());
	}

	/**
		Exports reporter annotations.
	*/
	@Override
	protected void exportGlobalFooter()
	{
		if (chainedProgress != null) chainedProgress.setRange(80, 95);
		exportReporterAnnotations(chainedProgress, exportedPositions.size());
	}

	/**
		Clean up.
	*/
	@Override
	protected void endExport(RuntimeException e)
	{
		if (e == null) setProgress(100, "Export complete. " + spotCount + " spots done");
		if (exportedPositions != null) exportedPositions.clear();
		exportedPositions = null;
		super.endExport(e);
	}
	// -------------------------------------------	

	/**
		Get the BFS subtype. This method may be overridden by subclasses.
		The default implementation returns null. The subtype is written to
		the metadata file: {@link MetadataWriter#setSubtype(String)}.
	*/
	public String getBFSSubtype()
	{
		return null;
	}

	/**
		Get the writer were metadata should be printed.
	*/
	protected MetadataWriter getMetadataWriter()
	{
		return metadataWriter;
	}

	/**
		Get the writer were reporter annotations should be printed.
		Can be null.
	*/
	protected AnnotationWriter getReporterAnnotationsWriter()
	{
		return reporterWriter;
	}

	/**
		Get the writer were assay annotations should be printed.
		Can be null.
	*/
	protected AnnotationWriter getAssayAnnotationsWriter()
	{
		return assayWriter;
	}
	
	/**
		Get the factory that creates data writers.
		Can be null.
	*/
	protected DataWriterFactory getDataWriterFactory()
	{
		return dataWriterFactory;
	}
	
	/**
		Export the metadata file which contains information about the
		other files (reporter annotations, assay annotations and spot data),
		plug-in parameters, etc. Plug-in parameters are registered with
		{@link #setParameter(String, String[])} and are written to
		the [parameters] section. If no parameters has been registered this 
		section is skipped.
	*/
	protected void exportMetadata()
	{
		setProgress(5, "Exporting metadata...");
		MetadataWriter out = getMetadataWriter();
		out.setSubtype(getBFSSubtype());
		

		// [files]
		out.bfsPrintSection("files", true, true);
		// Reporter and assay annotation files
		AnnotationWriter aw = getReporterAnnotationsWriter();
		if (aw != null) out.bfsPrintValue("rdata", aw.getFilename());
		aw = getAssayAnnotationsWriter();
		if (aw != null) out.bfsPrintValue("pdata", aw.getFilename());
		// Spot data files are labeled 'sdataN' with N = 1, 2, ...
		int index = 1;
		for (DataWriter dataWriter : dataWriters)
		{
			out.bfsPrintValue("sdata" + index, dataWriter.getFilename());
			index++;
		}
		// Custom extra data files
		for (SectionEntry extraFile : extraFiles)
		{
			out.bfsPrintValue(extraFile.getKey(), extraFile.getValues());
		}
	
		// [parameters]
		if (parameters.size() > 0)
		{
			out.bfsPrintSection("parameters", true, true);
			for (Map.Entry<String, String[]> entry : parameters.entrySet())
			{
				String key = entry.getKey();
				String[] values = entry.getValue();
				out.bfsPrintValue(key, values);
			}
		}
		
		// [sdata]
		List<DynamicField> spotFields = getSpotFields();
		if (spotFields.size() > 0)
		{
			out.bfsPrintSection("sdata", true, true);
			for (DynamicField spotField : spotFields)
			{
				Type spotType = spotField.getType();
				out.bfsPrintValue(spotField.getTitle(), spotType == null ? "" : spotType.getStringValue());
			}
		}
		out.flush();
		if (autoCloseWriters) out.close();
	}

	/**
		Export the reporter annotations. This method is ignored if
		{@link #getReporterAnnotationsWriter()} returns null. The 
		position is used as the value for the id column. Other columns
		are exported as given by {@link #getReporterFields()}.
		<p>
		If a progress reporter is given, progress is updated between 
		0 and 100%.
		<p>
		This method needs to be executed after {@link #exportSpotData(ProgressReporter, long)}
		since we need to know which positions that was exported.
		
		@param progress An optional progress reporter
		@param count The total number of reporters, or 0 if not known 
	*/
	@SuppressWarnings("unchecked")
	protected void exportReporterAnnotations(ProgressReporter progress, long count)
	{
		AnnotationWriter out = getReporterAnnotationsWriter();
		if (out == null) return;

		List<DynamicField> fields = getReporterFields();
		int numReporterFields = fields.size();
		DbControl dc = getDbControl();
		int posIndex = numReporterFields + 1;
		
		// Initialize headers and formatters
		String[] headers = new String[numReporterFields];
		Formatter[] formatters = new Formatter[numReporterFields];
		int index = 0;
		for (DynamicField df : fields)
		{
			headers[index] = df.getTitle();
			Formatter f = df.getFormatter();
			if (f == null) f = new ToStringFormatter();
			formatters[index] = f;
			index++;
		}
		out.bfsPrintHeaders(headers);
		
		if (progress != null) progress.display(0, "Loading reporter annotations...");
		DynamicPositionQuery query = getReporterQuery();
		DynamicResultIterator it = query.iterate(dc);
		Object[] data = new Object[numReporterFields];
		try
		{
			long numDone = 0;
			long progressInterval = Math.max(count > 0 ? 1 + count / 100 : 100, 100);
			while (it.hasNext())
			{
				checkInterrupted();
				SqlResult result = it.next();
				
				int position = result.getInt(posIndex);
				if (exportedPositions.contains(position))
				{
					index = 0;
					for (int i = 1; i <= numReporterFields; ++i)
					{
						data[index] = formatters[index].format(result.getObject(i));
						++index;
					}
					out.bfsPrintData(position, data);
					++numDone;
					if (progress != null && numDone % progressInterval == 0) 
					{
						int percent = count == 0 ? 0 : 10+(int)((90L * numDone) / count);
						progress.display(percent, "Exporting reporter annotations: " + 
							numDone + " of " + (count == 0 ? "unknown" : count) + " done");
					}
				}
			}
			out.flush();
			if (progress != null) 
			{
				progress.display(100, numDone + " reporter annotations exported.");
			}
		}
		catch (SQLException e)
		{
			throw new DatabaseException(e);
		}
		finally
		{
			if (it != null) it.close();
			if (out != null && autoCloseWriters) out.close();
		}
	}
	
	
	/**
		Export the assay annotations. This method is ignored if
		{@link #getAssayAnnotationsWriter()} returns null. The 
		assay ID is used as the value for the id column. Other columns
		are exported as given by {@link #getAssayFields()}.
	*/
	@SuppressWarnings("unchecked")
	protected void exportAssayAnnotations()
	{
		AnnotationWriter out = getAssayAnnotationsWriter();
		if (out == null) return;
		
		setProgress(10, "Exporting assay annotations...");
		List<AssayField> fields = getAssayFields();
		int numFields = fields.size();
		DbControl dc = getDbControl();

		int idIndex = numFields + 1;
		String[] headers = new String[numFields];
		Formatter[] formatters = new Formatter[numFields];
		
		int index = 0;
		for (AssayField df : fields)
		{
			headers[index] = df.getTitle();
			Formatter f = df.getFormatter();
			if (f == null) f = new ToStringFormatter();
			formatters[index] = f;
			index++;
		}
		out.bfsPrintHeaders(headers);
		
		List<BioAssay> assays = getBioAssays();
		Object[] data = new Object[numFields];

		for (BioAssay ba : assays)
		{
			checkInterrupted();
			index = 0;
			for (int i = 1; i <= numFields; ++i)
			{
				Collection<?> values = fields.get(index).getAssayValue(dc, ba);
				data[index] = values == null || values.size() == 0 ? 
						null : Values.getString(values, ", ", true, formatters[index]);
				++index;
			}
			out.bfsPrintData(ba.getId(), data);
		}
		out.flush();
		if (autoCloseWriters) out.close();
	}
	
	/**
		Export spot data. This method needs to be executed even if no spot
		fields has been selected. The reason is that we need to find out
		the positions that are used in the bioassay set when exporting
		reporter annotations. The only exception is if also no reporter
		annotations are exported.

		@param progress An optional progress reporter
		@param count The expected number of spots or 0 if not known
		@return The number of spots processed
	*/
	@SuppressWarnings("unchecked")
	protected long exportSpotData(ProgressReporter progress, long count)
	{
		exportedPositions = new HashSet<Integer>();
		DbControl dc = getDbControl();
		List<DynamicField> spotFields = getSpotFields();
		int numSpotFields = spotFields.size();
		if (numSpotFields == 0 && getReporterFields().size() == 0) return 0;

		Formatter[] formatters = new Formatter[numSpotFields];
		Object[] spotData = new Object[numSpotFields];
		for (int i = 0; i < numSpotFields; ++i)
		{
			Formatter f = spotFields.get(i).getFormatter();
			if (f == null) f = new ToStringFormatter();
			formatters[i] = f; 
		}
		
		Object[][] data = new Object[dataWriters.size()][];
		int index = 0;
		for (DataWriter dw : dataWriters)
		{
			data[index] = new Object[dw.getColumnCount()];
			++index;
		}
		
		int posIndex = numSpotFields + 1;
		int colIndex = numSpotFields + 2;
		int currentPosition = -1;

		// Execute the query
		if (progress != null) progress.display(0, "Loading spot data...");
		DynamicSpotQuery spotQuery = getSpotQuery(false);
		DynamicResultIterator it = spotQuery.iterate(dc);
		long numDone = 0;
		try
		{
			long progressInterval = Math.max(count > 0 ? 1 + count / 100 : 100, 100);
			while (it.hasNext())
			{
				checkInterrupted();
				SqlResult result = it.next();
				
				int position = result.getInt(posIndex);
				short column = result.getShort(colIndex);
				if (position != currentPosition)
				{
					if (currentPosition != -1) 
					{
						// print out data for the old position
						index = 0;
						for (DataWriter writer : dataWriters)
						{
							writer.bfsPrintData(data[index]);
							Arrays.fill(data[index], null);
							index++;
						}
						exportedPositions.add(currentPosition);
					}
					currentPosition = position;
				}
				
				// Copy formatted spot field values into the spotData array...
				for (int i = 0; i < spotData.length; ++i)
				{
					spotData[i] = formatters[i].format(result.getObject(i+1));
				}
				// ... and let the subclass arrange the data into the output array
				if (numSpotFields > 0) arrangeData(position, column, data, spotData);
				
				++numDone;
				if (progress != null && numDone % progressInterval == 0) 
				{
					int percent = count == 0 ? 0 : 10+(int)((90L * numDone) / count);
					progress.display(percent, "Exporting spot data: " + 
						numDone + " of " + (count == 0 ? "unknown" : count) + " done");
				}
			}
			
			// Print out the final data line
			exportedPositions.add(currentPosition);
			index = 0;
			for (DataWriter writer : dataWriters)
			{
				writer.bfsPrintData(data[index]);
				writer.flush();
				if (autoCloseWriters) writer.close();
				index++;
			}
		}
		catch (SQLException ex)
		{
			throw new DatabaseException(ex);
		}
		finally
		{
			if (it != null) it.close();
		}
		if (progress != null) progress.display(100, "Exported spot data: " + numDone + " spots");
		return numDone;
	}
	
	/**
		Utility method for calling creating a data writer. Calls
		{@link DataWriterFactory#createDataWriter(Object, String)}.
		@return A data writer, or null if no factory has been specified
	*/
	@SuppressWarnings("unchecked")
	protected DataWriter createDataWriter(Object owner, String suggestedFilename)
		throws IOException
	{
		DataWriter writer = null;
		if (dataWriterFactory != null) 
		{
			writer = dataWriterFactory.createDataWriter(owner, suggestedFilename);
		}
		return writer;
	}
	
	/**
		Create the exact number of data writers that is needed to export the
		desired data. Each data writer needs to specify a filename 
		({@link DataWriter#getFilename()} and the number of data columns 
		({@link DataWriter#getColumnCount()}) it is going to write.
		<p>
		
		This method must be implemented by a subclass and is called once from
		the {@link #beginExport()} method unless no data writer factory has been
		set up.
		
		@return A list with data writers
		@throws IOException If there is a problem with creating the writers
	*/
	protected abstract List<DataWriter> createDataWriters()
		throws IOException;
	
	/**
		Arrange the data that was returned by the query into the data writers
		array. The data array is a two-demensional array. The first index
		represents a datawriter and goes from 0 to the one below the number
		of data writers returned by {@link #createDataWriters()}. The size of
		the second index is determined by the number of columns for the
		data writer (eg. {@link DataWriter#getColumnCount()}.
		<p>
		
		This method must be implemented by a subclass which should copy
		data from the spotdata array into the data array. The data array
		is flushed to the data writers every time the position coordinate
		changes. Thus, this method can be called multiple times with
		different column (=assay) coordinates.
		
		@param position The current position of the spot data
		@param column The current column of the spot data
		@param data The data array to copy the data to
		@param spotData The current row of spot data, in the same order
			as {@link #getSpotFields()}
	*/
	protected abstract void arrangeData(int position, short column, 
		Object[][] data, Object[] spotData);
	
}
