/*
	$Id: ChannelFunction.java 5258 2010-02-26 11:05:49Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.util.BioAssaySetUtil;

import java.sql.SQLException;
import java.util.Stack;

import org.nfunk.jep.Node;
import org.nfunk.jep.ParseException;

/**
	A JEP function class that adds a <code>ch(int)</code> function to a 
	JEP expression parser. The function will look up the intensity value for the
	given channel. For example: <code>ch(1) / ch(2)</code>
	<p>
	Note! If the current bioassay set has stored transformed intensity values,
	the values are un-transformed before returned by the this function.
	Use {@link RawChannelFunction} if you need the raw values exactly 
	as they are stored in the database.	
	<p>
	To be able to use this function it must be registered with the JEP
	parser and, before the expression is evaluated, a {@link SqlResult} object must be set.
	For example we can evaluate an expression for every position in a bioassay:
	
	<pre class="code">
DbControl dc = ...
BioAssay assay = ...
String expression = "ch(1) / ch(2)";
ChannelFunction ch = new ChannelFunction({3, 4});
JEP jep = JepUtil.newJep(expression, ch);
DynamicResultIterator result = 
   assay.getSpotData().iterate(dc);
while (result.hasNext())
{
   ch.setSqlResult(result.next());
   double value = jep.getValue();
   // Do something with the value
}
result.close();
</pre>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2010-02-26 12:05:49 +0100 (Fri, 26 Feb 2010) $
	@see Jep
	@see BioAssaySetUtil#createJepExpression(DbControl, String, DynamicQuery)
*/
public class ChannelFunction
	implements JepExpressionFunction
{

	private int[] channelToIndex;
	private IntensityTransform transform;
	private SqlResult result;
	private int numParameters;

	/**
		Create a new instance of this function. The new instance cannot be used
		to dynamically evaluate expressions. It should only be used for converting
		JEP formulas to {@link Expression}:s.
		@see Jep#formulaToExpression(String, JepFunction[])	
	*/
	public ChannelFunction()
	{}
	
	/**
		@deprecated In 2.12. Use {@link ChannelFunction#ChannelFunction(int[], IntensityTransform)}
			instead. Using this constructor will disable the untransformation
			of transformed values when JEP dynamic evalution is used.
	*/
	public ChannelFunction(int[] channelToIndex)
	{
		this.channelToIndex = channelToIndex;
	}
	
	/**
		Create a new instance of this function which can be used
		to dynamically evaluate expressions.
		@param channelToIndex An array that maps channel numbers to column indexes in
			the SqlResult, array position 0 maps the index of channel number 1, etc.
		@see #setSqlResult(SqlResult)
	*/
	public ChannelFunction(int[] channelToIndex, IntensityTransform transform)
	{
		this.channelToIndex = channelToIndex;
		this.transform = transform;
	}
	
	/*
		From the JepFunction interface
		-------------------------------------------
	*/
	/**
		@return The string "ch"
	*/
	public String getFunctionName()
	{
		return "ch";
	}
	// -------------------------------------------
	/*
		From the JepExpressionFunction interface
		-------------------------------------------
	*/
	/**
		Use the {@link Dynamic#column(VirtualColumn)} method to create an
		expression referencing a channel intensity.
	*/
	public Expression toExpression(Node node)
	{
		int numChildren = node.jjtGetNumChildren();
		if (numChildren != 1)
		{
			throw new BaseException("Invalid number of expressions for 'ch' function: " + numChildren);
		}
		return Dynamic.column(VirtualColumn.channelIntensity(Jep.nodeToInt(node.jjtGetChild(0))));
	}
	// -------------------------------------------
	/*
		From the PostfixMathCommandI interface
		-------------------------------------------
	*/
	/**
		@return Always 1
	*/
	public int getNumberOfParameters()
	{
		return 1;
	}
	public void setCurNumberOfParameters(int n)
	{
		this.numParameters = n;
	}
	public boolean checkNumberOfParameters(int n)
	{
		return n == 1;
	}
	@SuppressWarnings({"unchecked"})
	public void run(Stack stack)
		throws ParseException
	{
		if (stack == null || stack.empty()) 
		{
			throw new ParseException("Stack is empty");
		}
		Object channel = stack.pop();
		if (channel instanceof Number)
		{
			stack.push(channel(((Number)channel).intValue()));
		}
		else
		{
			throw new ParseException("Invalid parameter type: " + channel + "; expected integer");
		}
	}
	// -------------------------------------------

	/**
		Set a new {@link SqlResult} object that will be used the next time the
		JEP expression is evaluated.
		@param result The result object
	*/
	public void setSqlResult(SqlResult result)
	{
		this.result = result;
	}
	/**
		Get the value of the specified channel of the current sql result.
		@param channel The channel number
	 	@throws ParseException Either if it's an invalid channel number or 
	 		if no result object has been specified.
	*/
	public Object channel(int channel)
		throws ParseException
	{
		if (result == null)
		{
			throw new ParseException("No result object has been specified for function ch("+channel+")");
		}
		if (channelToIndex == null || channel < 1 || channel > channelToIndex.length)
		{
			throw new ParseException("Invalid channel number for function ch("+channel+")");
		}
		try
		{
			Number n = (Number)result.getObject(channelToIndex[channel-1]);
			if (transform != null) n = transform.unTransform(n.doubleValue());
			return n;
		}
		catch (SQLException ex)
		{
			throw new ParseException(ex.getMessage());
		}
	}
}
