/*
	$Id: ExtraValueFunction.java 5258 2010-02-26 11:05:49Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.ExtraValue;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.util.BioAssaySetUtil;
import net.sf.basedb.util.Values;

import java.sql.SQLException;
import java.util.Map;
import java.util.Stack;

import org.nfunk.jep.Node;
import org.nfunk.jep.ParseException;


/**
	A JEP function class that adds a <code>xtra(int)</code> function to a 
	JEP expression parser. The function will look up the value of the extra value
	with the given ID. For example: <code>xtra(1)</code>
	<p>
	To be able to use this function it must be registered with the JEP
	parser.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2010-02-26 12:05:49 +0100 (Fri, 26 Feb 2010) $
	@see Jep
	@see BioAssaySetUtil#createJepExpression(DbControl, String, DynamicQuery)
*/
public class ExtraValueFunction
	implements JepExpressionFunction
{
	private final DbControl dc;
	private final Map<Integer, Integer> extraToIndex;
	private final BioAssaySet bas;
	private int numParameters;
	private SqlResult result;

	/**
		Create a new instance of this function. The new instance cannot be used
		to dynamically evaluate expressions. It should only be used for converting
		JEP formulas to {@link Expression}:s.
	 	@param dc DbControl to use when accessing the database 
	 	@param bas Bioassay set for this extra value function. 
		@see Jep#formulaToExpression(String, JepFunction[])
	*/
	public ExtraValueFunction(DbControl dc, BioAssaySet bas)
	{
		this.dc = dc;
		this.extraToIndex = null;
		this.bas = bas;
	}
	/**
		Create a new instance of this function which can be used
		to dynamically evaluate expressions.
	 	@param dc DbControl to use when accessing the database 
		@param extraToIndex An map from extra value ID to column indexes in
			the SqlResult
		@see #setSqlResult(SqlResult)
	*/
	public ExtraValueFunction(DbControl dc, Map<Integer, Integer> extraToIndex)
	{
		this.dc = dc;
		this.extraToIndex = extraToIndex;
		this.bas = null;
	}	
	/*
		From the JepFunction interface
		-------------------------------------------
	*/
	/**
		@return The string "xtra"
	*/
	public String getFunctionName()
	{
		return "xtra";
	}
	// -------------------------------------------
	/*
		From the JepExpressionFunction interface
		-------------------------------------------
	*/
	/**
		Use the {@link Dynamic#extraValue(ExtraValue)} method to create an
		expression referencing an extra value.
	*/
	public Expression toExpression(Node node)
	{
		int numChildren = node.jjtGetNumChildren();
		if (numChildren != 1)
		{
			throw new BaseException("Invalid number of expressions for 'xtra' function: " + numChildren);
		}
		// First try to load by external id
		String arg = Jep.nodeToString(node.jjtGetChild(0));
		int extraValueId = Values.getInt(arg, -1);
		ExtraValueType xType = extraValueId == -1 ? ExtraValueType.getByExternalId(dc, arg) : null;
		ExtraValue xv = xType != null ? bas.getExtraValue(xType) : ExtraValue.getById(dc, extraValueId);
		return Dynamic.extraValue(xv);
	}
	// -------------------------------------------
	/*
		From the PostfixMathCommandI interface
		-------------------------------------------
	*/
	/**
		@return Always 1
	*/
	public int getNumberOfParameters()
	{
		return 1;
	}
	public void setCurNumberOfParameters(int n)
	{
		this.numParameters = n;
	}
	public boolean checkNumberOfParameters(int n)
	{
		return n == 1;
	}
	@SuppressWarnings({"unchecked"})
	public void run(Stack stack)
		throws ParseException
	{
		if (stack == null || stack.empty()) 
		{
			throw new ParseException("Stack is empty");
		}
		Object extraId = stack.pop();
		if (extraId instanceof Number)
		{
			stack.push(xtra(((Number)extraId).intValue()));
		}
		else
		{
			throw new ParseException("Invalid parameter type: " + extraId + "; expected number");
		}
	}
	// -------------------------------------------

	/**
		Set a new {@link SqlResult} object that will be used the next time the
		JEP expression is evaluated.
		@param result The result object
	*/
	public void setSqlResult(SqlResult result)
	{
		this.result = result;
	}

	/**
		Get the value of the specified extra value of the current sql result.
		@param extraId The extra value ID
	 	@return an extra value as an Object.
	 	@throws ParseException If no result has been specified or
	 		if no extra value was found.
	*/
	public Object xtra(Integer extraId)
		throws ParseException
	{
		if (result == null)
		{
			throw new ParseException("No result object has been specified for function xtra("+extraId+")");
		}
		if (extraToIndex == null || extraToIndex.get(extraId) == null)
		{
			throw new ParseException("Extra value not found for function xtra("+extraId+")");
		}
		try
		{
			return result.getObject(extraToIndex.get(extraId));
		}
		catch (SQLException ex)
		{
			throw new ParseException(ex.getMessage());
		}
	}
}
