/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep;

import java.util.Stack;

import org.nfunk.jep.FunctionTable;
import org.nfunk.jep.JEP;
import org.nfunk.jep.ParseException;
import org.nfunk.jep.Token;
import org.nfunk.jep.function.PostfixMathCommandI;

/**
	Extension to the JEP parser that provides better error handling for
	unknown functions. The normal error message when JEP encounters
	an unknown function is: Syntax Error (implicit multiplication not enabled).
	This is not very informative so this class improves the error handling
	by giving a message: Unknown function foo() instead.
	
	@author Nicklas
	@since 2.15
	@base.modified $Date $
*/
public class FunctionSafeJep 
	extends JEP 
{
	
	/**
		Create a new function-safe JEP parser.
	*/
	public FunctionSafeJep() 
	{
		super();
	}
	
	/**
		Overrides the regular function lookup table by replacing it with a
		function table that pretends that all functions has been registered,
		but then generates an error message if it doesn't exists.
	*/
	@Override
	public void initFunTab()
	{
		this.funTab = new SafeFunctionTable(this);
	}
	
	/**
		Adds an error message to the "real" JEP parser.
	*/
	@SuppressWarnings("unchecked")
	void addError(String error)
	{
		this.errorList.add(error);
	}
	
	/**
		Function table implementation that pretends that all possible functions
		exists, but sends an error message if someone tries to use a function that
		doesn't really exists.
	*/
	static class SafeFunctionTable
		extends FunctionTable
	{
		
		private static final long serialVersionUID = 4430148600636575401L;
		private final FunctionSafeJep jep;
		SafeFunctionTable(FunctionSafeJep jep) 
		{
			super();
			this.jep = jep;
		}
		

		/**
			If the key is already registered, return TRUE. If the key is not
			registered and is a function, register it as an unknown function,
			set an error message and return TRUE. Otherwise return FALSE.
		*/
		@Override
		public boolean containsKey(Object key)
		{
			boolean exists = super.containsKey(key);
			//System.out.println("containsKey: " + key + "=" + exists +";nextToken="+nextToken((String)key));
			if (!exists && key instanceof String)
			{
				String name = (String)key;
				if ("(".equals(nextToken(name)))
				{
					super.put(key, new UnknownFunction(name));
					jep.addError("Unknown function: " + name + "()");
					exists = true;
				}
			}
			return exists;
		}
		
		/**
			Get the next token in the parser. If this is a "(", the 
			parse is currently looking up a function name.
			@param token The current token
		*/
		private String nextToken(String token)
		{
			String next = null;
			Token t = jep.parser.token;
			while (t != null)
			{
				if (token.equals(t.image))
				{
					if (t.next != null) next = t.next.image;
					break;
				}
				t = t.next;
			}
			return next;
		}
	}
	
	/**
		Represents an unknown function that appears in an expression.
		The error is usually catched at parse time, otherwise it will
		be catched at evaluation time by this class.
	*/
	static class UnknownFunction
		implements PostfixMathCommandI
	{

		private final String name;
		UnknownFunction(String name)
		{
			this.name = name;
		}
		
		/*
			From the PostfixMathCommandI interface
			-------------------------------------------
		*/
		@Override
		public boolean checkNumberOfParameters(int numParameters) 
		{
			return true;
		}
		@Override
		public int getNumberOfParameters() 
		{
			return -1;
		}
		@Override
		public void run(Stack paramStack) 
			throws ParseException 
		{
			throw new ParseException("Unknown function: " + name + "()");
		}
		@Override
		public void setCurNumberOfParameters(int numParameters) 
		{}
		// ---------------------------------------
		
	}
	
}
