/**
	$Id: AutoChildNodeLoader.java 4740 2009-02-05 13:39:25Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Node;

/**
	Node loader implementation that automatically loads child nodes.
	When any of the <code>create...</code> methods are called the actual
	node loading is delegated to a parent node loader. If this loader returns
	a non-null node, {@link #loadChildNodes(DbControl, OverviewContext, Node)}
	is automatically invoked on it.
	<p>
	The automatic loading can of course only be done one step down, since
	the new children probably needs a different parent node loader. It is 
	possible to make the loading fully recursive by if the {@link
	NodeLoaderFactory} returned from {@link OverviewContext#getNodeLoaderFactory()}
	wrap all node loaders it create by an instance of this class.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-05 14:39:25 +0100 (Thu, 05 Feb 2009) $
*/
public class AutoChildNodeLoader<I>
	implements NodeLoader<I>
{

	private final NodeLoader<I> parent;
	public AutoChildNodeLoader(NodeLoader<I> parent)
	{
		this.parent = parent;
	}
	/*
		From the NodeLoader interface
		-----------------------------
	*/
	/**
		Call the same method on the 'parent' node loader. If a node is created
		{@link #loadChildNodes(DbControl, OverviewContext, Node)} is called
		on the new node.
		@return The node that the parent node loader creates (may be null)
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node result = parent.createForwardNode(dc, context, parentNode);
		if (result != null) loadChildNodes(dc, context, result);
		return result;
	}

	/**
		Call the same method on the 'parent' node loader. If a node is created
		{@link #loadChildNodes(DbControl, OverviewContext, Node)} is called
		on the new node.
		@return The node that the parent node loader creates (may be null)
	*/
	@Override
	public Node createPropertyNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node result = parent.createPropertyNode(dc, context, parentNode);
		if (result != null) loadChildNodes(dc, context, result);
		return result;
	}

	/**
		Call the same method on the 'parent' node loader. If a node is created
		{@link #loadChildNodes(DbControl, OverviewContext, Node)} is called
		on the new node.
		@return The node that the parent node loader creates (may be null)
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node childNode)
	{
		Node result = parent.createReverseNode(dc, context, childNode);
		if (result != null) loadChildNodes(dc, context, result);
		return result;
	}

	/**
		Call the same method on the 'parent' node loader. If a node is created
		{@link #loadChildNodes(DbControl, OverviewContext, Node)} is called
		on the new node.
		@return The node that the parent node loader creates (may be null)
	*/
	@Override
	public Node createRootNode(DbControl dc, OverviewContext context, I item)
	{
		Node result = parent.createRootNode(dc, context, item);
		if (result != null) loadChildNodes(dc, context, result);
		return result;
	}

	/**
		Call the same method on the 'parent' node loader.
		@return The result of the parent call
	*/
	@Override
	public boolean loadChildNodes(DbControl dc, OverviewContext context, Node node)
	{
		return parent.loadChildNodes(dc, context, node);
	}
	// --------------------------------

	@Override
	public String toString()
	{
		return super.toString() + "-->" + parent.toString();
	}
	
}
