/**
	$Id: ScanLoader.java 4740 2009-02-05 13:39:25Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.Scan;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.cache.IndexedCacheKey;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for scans. The forward-loading
	direction goes from {@link Hybridization}:s -&gt; {@link Scan}:s -&gt; 
	{@link RawBioAssay}:s. The reverse-loading direction is the opposite 
	direction.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-05 14:39:25 +0100 (Thu, 05 Feb 2009) $
*/
public class ScanLoader
	extends BasicItemNodeLoader<Scan>
{

	public ScanLoader()
	{
		super(Item.SCAN, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<Scan>("scan", "Scan"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading scan nodes from a given hybridization node.
		The returned node is a folder-type node containing item nodes for 
		each scan that has been created from the hybridization.
		@return The folder node for the scans
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node hybridizationNode)
	{
		NodeFactory<Scan> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		Hybridization hyb = (Hybridization)hybridizationNode.getItem(dc);
		ItemQuery<Scan> scanQuery = context.initQuery(hyb.getScans(), "name");

		// If we have a labeled extract as parent we need to know the array number
		// to be able to load the correct raw bioassays later -- The same hyb may have
		// other raw bioassays that are NOT related to the same labeled extract
		Node labeledExtractNode = hybridizationNode.getFirstParent(new ItemTypeFilter(Item.LABELEDEXTRACT));
		int arrayIndex = 0;
		if (labeledExtractNode != null)
		{
			LabeledExtract extract = (LabeledExtract)labeledExtractNode.getItem();
			arrayIndex = hyb.getCreationEvent().getSourceGroup(extract);
		}

		ItemResultIterator<Scan> it = scanQuery.iterate(dc);
		while (it.hasNext())
		{
			Scan scan = it.next();
			if (folderNode == null)
			{
				folderNode = new Node("scans", "Scans", hybridizationNode, ChildNodeDirection.FORWARD);
			}
			Object cacheKey = arrayIndex == 0 ? scan : new IndexedCacheKey(scan, arrayIndex);
			createItemNode(nf, scan, cacheKey, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, hybridizationNode, false);
		return folderNode;
	}
	
	/**
		Create a reverse-loading scan node from the raw bioassay node.
		@return A scan node, or null if the raw bioassay doesn't have a scan
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		NodeFactory<Scan> nf = getNodeFactory(dc, context);
		Scan scan = null;
		boolean denied = false;
		int arrayNum = 0;
		try
		{
			RawBioAssay rba = (RawBioAssay)rawBioAssayNode.getItem(dc);
			arrayNum = rba.getArrayNum();
			scan = rba.getScan();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Object cacheKey = new IndexedCacheKey(scan, arrayNum);
		Node scanNode = createItemNode(nf, scan, cacheKey, denied, 
				rawBioAssayNode, ChildNodeDirection.REVERSE);
		return scanNode;
	}
	// --------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads property nodes of a scan. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Scanner: {@link HardwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Images: {@link ImageLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node scanNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, scanNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, scanNode);
		getNodeLoader(context, Item.HARDWARE).createPropertyNode(dc, context, scanNode);
		getNodeLoader(context, Item.IMAGE).createPropertyNode(dc, context, scanNode);
	}

	/**
		Loads all raw bioassay nodes that are children of the given scan node.
		@see RawBioAssayLoader#createForwardNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node scanNode)
	{
		getNodeLoader(context, Item.RAWBIOASSAY).createForwardNode(dc, context, scanNode);
	}
	
	/**
		Loads the hybridization node that is the parent of this scan. 
		@see HybridizationLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node scanNode)
	{
		getNodeLoader(context, Item.HYBRIDIZATION).createReverseNode(dc, context, scanNode);
	}
	// ---------------------------------------
	
}
