/**
	$Id: ProtocolValidator.java 4764 2009-02-16 19:29:47Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.ProtocolType;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.OverviewUtil;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for protocols. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_PROTOCOL}
	<li>Access denied: {@link Validator#DENIED_PROTOCOL}
	<li>Incorrect protocol type: {@link Validator#INCORRECT_PROTOCOLTYPE} 
	<li>Non-default protocol: {@link Validator#NONDEFAULT_PROTOCOL}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-16 20:29:47 +0100 (Mon, 16 Feb 2009) $
*/
public class ProtocolValidator
	extends NameableNodeValidator<Protocol>
{
	
	public ProtocolValidator()
	{
		super(Validator.MISSING_PROTOCOL, Validator.DENIED_PROTOCOL);
	}

	
	/* 
		From BasicNodeValidator class
		-----------------------------
	*/
	/**
		Checks if the protocol is of the correct protocol type
		and, if a project is active, the protocol is the same as
		the default for the project.
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		Protocol protocol = (Protocol)node.getItem();
		BasicItem parentItem = parentNode.getItem();
		Project.Default defaultType = Project.getDefaultType(Item.PROTOCOL, parentNode.getItemType());
		String protocolTypeName = "protocol";
		String parentTypeName = "item";
		if (defaultType != null) 
		{
			protocolTypeName = defaultType.getShortName().toLowerCase();
			parentTypeName = defaultType.getAttachesToItemType().toString().toLowerCase();
		}
		
		// Check if using project default protocol
		Project project = context.getProject();
		BasicItem defaultProtocol = OverviewUtil.getProjectDefaultItem(dc, project, defaultType);
		if (defaultProtocol != null && !defaultProtocol.equals(protocol))
		{
			context.createFailure(Validator.NONDEFAULT_PROTOCOL, parentNode, null,
				new Fix("Change protocol of " + parentTypeName, parentItem),
				new Fix("Change default " + protocolTypeName + " for project", project)
			);
		}

		// Check protocol type
		try
		{
			ProtocolType pt = protocol.getProtocolType();
			String requiredId = defaultType == null ? null : defaultType.getType();
			if (requiredId != null && !requiredId.equals(pt.getSystemId()))
			{
				context.createFailure(Validator.INCORRECT_PROTOCOLTYPE, parentNode, null,
					new Fix("Change protocol of " + parentTypeName, parentItem),
					new Fix("Change protocol type to '" + protocolTypeName + "'", protocol)
				);
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}
	
	/**
		@return Suggested fix is to add a protocol to the parent item
	*/
	@Override
	protected Fix getMissingItemFix(Node parentNode)
	{
		Project.Default defaultType = Project.getDefaultType(Item.PROTOCOL, parentNode.getItemType());
		String parentTypeName = "item";
		if (defaultType != null) 
		{
			parentTypeName = defaultType.getAttachesToItemType().toString().toLowerCase();
		}
		BasicItem parentItem = parentNode.getItem();
		return new Fix("Add protocol to " + parentTypeName, parentItem);
	}
	// --------------------------------
}
