/*
	$Id: MedianRatioNormalization.java 5103 2009-09-28 08:45:48Z nicklas $

	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007, 2008 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007, 2008 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.FloatParameterType;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Values;

/**
	 @author enell, Nicklas
	 @version 2.0
	 @base.modified $Date: 2009-09-28 10:45:48 +0200 (Mon, 28 Sep 2009) $
*/

public class MedianRatioNormalization
	extends AbstractAnalysisPlugin
	implements InteractivePlugin, SignalTarget
{

	/**
	   Plugin description
	*/
	private static final About about = new AboutImpl
	(
		"Normalization: Median ratio",
		("This normalizer behaves differently depending on the array type.\n\n" +

		 "1-channel arrays: Intensities are changed in two different ways depending " +
		 "on user selection.\n" +
		 "a) For absolute intensity data the intensities should be scaled with the " +
		 "median intensity, Inew=Iold/Imedian, such that resulting median is 1.\n" +
		 "b) For log intensity data the intensities should be subtracted with the median " +
		 "intensity, log(Inew)=log(Iold)-log(Imedian), such that resulting median is 0.\n\n" + 

		 "2-channel arrays: Intensities are scaled such " +
		 "that their geometric mean (sqrt(i1 * i2)) is kept constant but " +
		 "the median of their ratios (i1/i2) is shifted to 1.\n\n" +

		 "All array types: The user can optionally exclude spots; all spots " +
		 "with at least one intensity below a user selectable threshold are " +
		 "removed. User selectable (low/high intensity) fractions of the " +
		 "remaining spots are ignored in median calculation."),
		 Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
		"2006-8, Base 2 development team",
		null,
		null,
		"http://base.thep.lu.se"
	);

	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	// The default FloatParameterType is set in
	// getConfigureJobParameters() since it depends on the number of
	// channels.
	private PluginParameter<Float> minIntensityParameter = new PluginParameter<Float>
	(
		"minIntensity",
		"Minimum intensity",
		"Use spots with both int greater than [value]",
		null
	);

	private static final PluginParameter<Float> highExcludeParameter = new PluginParameter<Float>
	(
		"highExclude",
		"Exclude top",
		"A percentage of the highest values that are ignored when calculating the median.",
		new FloatParameterType(0F, 100F, 5F, true)
	);
	
	private static final PluginParameter<Float> lowExcludeParameter = new PluginParameter<Float>
	(
		"lowExclude",
		"Exclude low",
		"A percentage of the lowest values that are ignored when calculating the median.",
		new FloatParameterType(0F, 100F, 5F, true)
	);

	/**
		 List of median normalization methods.

		 The method to use depends whether data is log values or
		 not. Absolute intensities should be scaled with the median
		 whereas log intensities should be subtracted with the median
		 intensity.

		   Inew = Iold/Imedian for absolute intensities.

		   log(Inew) = log(Iold) - log(Imedian) for log intensities.

		 Used with 1-channel data only
	*/
	private static final
	PluginParameter<String> methodParameter = new PluginParameter<String>
		( "method",
			"Normalization method",
			"Scale: Scaling of intensities, Inew=Iold/Imedian. Use with absolute intensities.\n" +
			"Subtract: Subtraction of intensities, log(Inew)=log(Iold)-log(Imedian), Use with " +
			"log intensities.",
			new StringParameterType(255, "Scale", true, 1, 0, 0,
															Arrays.asList(
																new String[] { "Scale" ,
																							 "Subtract" }
																)
				)
			);

	/**
		 Used with 2-channel data only
	 */
	private static final PluginParameter<Integer> blockGroupParameter = new PluginParameter<Integer>
	(
		"blockGroup",
		"Blockgroup size",
		"The number of arraydesign blocks in each group. Use '0' to normalize all spots as a single group.",
		new IntegerParameterType(0, null, 0, true)
	);

	/**
		Section for normalization options
	*/
	protected static final PluginParameter<String> medianRatioSection = new PluginParameter<String>(
		"medianRatioSection",
		"Normalization options",
		"", 
		null
	);

	private RequestInformation configureJob;

	private ThreadSignalHandler signalHandler;
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	public About getAbout()
	{
		return about;
	}
	public boolean supportsConfigurations()
	{
		return false;
	}
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request use access to Experiment:s and read access to Job:s.
		@since 2.0.2
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.JOB, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	public void run(Request request, Response response, ProgressReporter progress)
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_EXECUTE))
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				BioAssaySet source = getSourceBioAssaySet(dc);
				String childName = Values.getString((String)job.getValue(CHILD_NAME), source.getName());
				String childDescription = (String)job.getValue(CHILD_DESCRIPTION);
				float minIntensity = (Float)job.getValue(minIntensityParameter.getName());
				float lowExclude = (Float) job.getValue(lowExcludeParameter.getName());
				float highExclude = (Float) job.getValue(highExcludeParameter.getName());
				Job thisJob = getCurrentJob(dc);
				
				BioAssaySet child = null;
				int rawChannels = source.getRawDataType().getChannels();
				switch (rawChannels)
				{
					case 1:
						String method = (String) job.getValue(methodParameter.getName());
						child = normalize1ch(dc, source, thisJob, minIntensity, lowExclude,
										highExclude, method.equals("Scale"), progress);
						break;
					case 2:
						int blockGroupSize =
							(Integer) job.getValue(blockGroupParameter.getName());
						child = normalize(dc, source, thisJob, minIntensity, lowExclude, 
								highExclude, blockGroupSize, progress);
							break;
					default:
							throw new InvalidDataException(rawChannels + 
								"-channel arrays not supported");
				}
				
				child.setName(childName);
				child.setDescription(childDescription);
				dc.commit();
				int normalizedSpots = child.getNumSpots();
				response.setDone(normalizedSpots + " spots normalized, " + (source.getNumSpots() - normalizedSpots) + " spots removed");
			}
			catch (Throwable ex)
			{
				response.setError(ex.getMessage(), Arrays.asList(ex));
			}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		else
		{
			response.setError("Unknown command: " + command, null);
		}
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	public String isInContext(GuiContext context, Object item)
	{
		String message = super.isInContext(context, item);	
		if (message == null)
		{
			BioAssaySet bas = (BioAssaySet)item;
			int nofChannels = bas.getRawDataType().getChannels();
			if (nofChannels > 2)
			{
				message = ("Only 1- or 2-channel data is supported, not " +
									 nofChannels + "-channel.");
			}
			else if (bas.getNumSpots() <= 0)
			{
				message = "This plug-in requires spot data in the database";
			}
			else if (bas.getMaxRawMappingsForSpot() != 1)
			{
				message = "Normalization can't be done on merged data.";
			}
		}
		return message;
	}

	public RequestInformation getRequestInformation(GuiContext context, String command) throws BaseException
	{
		RequestInformation requestInformation = null;
		
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters();
		}
		return requestInformation;
	}

	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				// Validate job parameters
				RequestInformation ri = getConfigureJobParameters();
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				if ((Float)request.getParameterValue(lowExcludeParameter.getName()) + (Float)request.getParameterValue(lowExcludeParameter.getName()) > 100)
				{
					response.setError("The sum of low exclude and high exclude cannot be greater than 100", null);
					return;
				}
				
				// Source bioassay set
				storeValue(job, request, ri.getParameter(SOURCE_BIOASSAYSET));
								
				// Child name, description and transformation
				storeValue(job, request, ri.getParameter(CHILD_NAME));
				storeValue(job, request, ri.getParameter(CHILD_DESCRIPTION));

				// MedianRatio options
				// 1- and 2-channel common
				storeValue(job, request, minIntensityParameter);
				storeValue(job, request, highExcludeParameter);
				storeValue(job, request, lowExcludeParameter);
				// 1-channel only
				storeValue(job, request, methodParameter);
				// 2-channel only
				storeValue(job, request, blockGroupParameter);
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------
	
	/**
		Normalise the source bioassay set using MEDIAN-RATIO normalization.
		@param dc The DbControl to use for database access
		@param source The source bioassay set that is going to be normalized
		@param job The job in which the normalization is done.
		@param minIntensity All spots which have a lower intensity value
			in either channel will be filtered out
		@param lowExclude A percentage of the spots with lowest ratio that
			are excuded in the median calculation
		@param highExclude A percentage of the spots with highest ratio
			that are excluded in the median calculation
		@param scale If true, scale intensities with median intensity,
		  otherwise subtract median intensity
	 	@param progress Progress reporter for the caller to keep track of
	 	the plugin's progress. Null is allowed.
		@return The normalized bioassayset
		@since 2.7
	*/
	private BioAssaySet
	normalize1ch(DbControl dc, BioAssaySet source, Job job,
							 float minIntensity, float lowExclude, float highExclude,
							 boolean scale, ProgressReporter progress)
	{
		// Create Transformation
		Transformation t = source.newTransformation(job);
		t.setName(about.getName());
		dc.saveItem(t);

		// Create the normalized bioassay set
		BioAssaySet child = t.newProduct(null, "new", true);
		dc.saveItem(child);

		// Batcher for inserting normalized data
		SpotBatcher batcher = child.getSpotBatcher();
		
		// Expressions used to get data
		Expression ch1 = Dynamic.column(VirtualColumn.channelIntensity(1));

		// Create restriction: ch1 > minIntensity
		Restriction intensityRestriction = Restrictions.and
			( Restrictions.gt(Dynamic.column(VirtualColumn.channelIntensity(1)),
												Expressions.aFloat(minIntensity)) );

		// Create restriction: column = :bioAssayColumn
		Restriction bioAssayRestriction = Restrictions.eq(
				Dynamic.column(VirtualColumn.COLUMN),
				Expressions.parameter("bioAssayColumn")
			);

		// Get the number of spots to be normalized
		DynamicSpotQuery query = source.getSpotData();
		query.restrictPermanent(intensityRestriction);
		long numSpots = query.count(dc);
		query.reset();
		long normalizedSpots = 0;
		if (progress != null)
			progress.display((int)(normalizedSpots / numSpots * 100),
											 normalizedSpots + " spots normalized");

		// Add to query to retrieve spot data: COLUMN, POSITION, and ch1
		// We use a parameter to restrict the query to return data for one
		// bioassay at a time
		query.select(Dynamic.select(VirtualColumn.POSITION));
		query.select(Selects.expression(ch1, "ch1"));
		query.restrict(bioAssayRestriction);
		query.order(Orders.asc(ch1));

		// Normalize one bioassay at a time
		List<BioAssay> assays = source.getBioAssays().list(dc);
		try
		{
			for (BioAssay assay : assays)
			{
				checkInterrupted();

				// query spot data for this bioassay
				short bioassayColumn = assay.getDataCubeColumnNo();
				query.setParameter("bioAssayColumn", (int)bioassayColumn, Type.INT);
				DynamicResultIterator it = query.iterate(dc);
				int positionIndex = it.getIndex(VirtualColumn.POSITION.getName());
				int ch1Index = it.getIndex("ch1");

				// Prepare arrays for holding data
				int nofAssaySpots = assay.getNumSpots(); // maximum number of spots
				float[] intensity = new float[nofAssaySpots];
				int[] position = new int[nofAssaySpots];

				// Copy bioassay data to SpotData objects
				int nofChildSpots=0; // the number of spots surviving intensity cut
				while (it.hasNext())
				{
					SqlResult r = it.next();
					intensity[nofChildSpots]=r.getFloat(ch1Index);
					position[nofChildSpots]=r.getInt(positionIndex);
					++nofChildSpots;
				}
				it.close();

				// Exlude low and high values in median calculation
				int lowIndex = (int)((nofChildSpots-1) * lowExclude/100);
				int highIndex = (int)((nofChildSpots-1) * (1 - highExclude/100));
				float median = median(intensity,lowIndex,highIndex);

				for (int i = 0; i<nofChildSpots; ++i)
				{
					batcher.insert(bioassayColumn, position[i],
												 (scale ? intensity[i]/median : intensity[i]-median) );
				}
				normalizedSpots += nofChildSpots;
				if (progress != null)
					progress.display((int)((normalizedSpots * 100L) / numSpots),
													 normalizedSpots + " spots normalized");
			}
			batcher.flush();
		}
		catch (SQLException e)
		{
			throw new BaseException(e);
		}
		return child;
	}

	/**
		Normalise the source bioassay set using MEDIAN-RATIO normalization. NOTE!
		This normalization method only works with 2-channel data.
		
		@param dc The DbControl to use for database access
		@param source The source bioassay set that is going to be normalized
		@param job The job that is doing the normalization, or null
		@param minIntensity All spots which have a lower intensity value in either channel
			will be filtered out
		@param lowExclude A percentage of the spots with lowest ratio that are excuded in the
			median calculation
		@param highExclude A percentage of the spots with highest ratio that are excluded in the
			median calculation
		@param blockGroupSize
	 	@param progress Progress reporter for the caller to keep track of the plugin's progress. Null is allowed. 
		@return The normalized bioassayset
		@since 2.5; This method was removed in 2.7.0 by accident, added in 2.7.2 again
	*/
	public BioAssaySet normalize(DbControl dc, BioAssaySet source, Job job, 
			float minIntensity, float lowExclude, float highExclude, 
			int blockGroupSize, ProgressReporter progress)
	{
		if (progress != null) progress.display(0, "Preparing to normalize...");
	
		boolean useBlocks = blockGroupSize > 0;
		
		// Create Transformation
		Transformation t = source.newTransformation(job);
		t.setName(about.getName());
		dc.saveItem(t);
		
		// Create the normalized bioassay set
		BioAssaySet child = t.newProduct(null, "new", true);
		dc.saveItem(child);
		
		// Batcher for inserting normalized data
		SpotBatcher batcher = child.getSpotBatcher();
		
		// Expressions used to get data
		Expression block = useBlocks ? Dynamic.rawData("block") : Expressions.integer(0);
		Expression ch1 = Dynamic.column(VirtualColumn.channelIntensity(1));
		Expression ch2 = Dynamic.column(VirtualColumn.channelIntensity(2));
		
		//	 Create restriction: ch1 > minIntensity and ch2 > minIntensity
		Restriction intensityRestriction = Restrictions.and
		(
			Restrictions.gt(Dynamic.column(VirtualColumn.channelIntensity(1)), Expressions.aFloat(minIntensity)), 
			Restrictions.gt(Dynamic.column(VirtualColumn.channelIntensity(2)), Expressions.aFloat(minIntensity))
		);
		
		// Create restriction: column = :bioAssayColumn
		Restriction bioAssayRestriction = Restrictions.eq(
				Dynamic.column(VirtualColumn.COLUMN), 
				Expressions.parameter("bioAssayColumn")
			);
		
		// Count number of spots that is going to be normalized
		DynamicSpotQuery query = source.getSpotData();
		query.restrictPermanent(intensityRestriction);
		checkInterrupted();
		long numSpots = query.count(dc);
		query.reset();
		long normalizedSpots = 0;
		if (progress != null) progress.display((int)(normalizedSpots / numSpots * 100), normalizedSpots + " spots normalized");
		
		// Create query to retrieve spot data: COLUMN, POSITION, ch1, ch2, block
		// We use a parameter to restrict the query to return data for one bioassay at a time
		query.select(Dynamic.select(VirtualColumn.POSITION));
		query.select(Selects.expression(ch1, "ch1"));
		query.select(Selects.expression(ch2, "ch2"));
		query.select(Selects.expression(block, "block"));
		query.restrict(bioAssayRestriction);
		if (useBlocks) query.order(Orders.asc(block));
	
		// Normalize one bioassay at a time
		List<BioAssay> assays = source.getBioAssays().list(dc);
		try
		{
			for (BioAssay assay : assays)
			{
				checkInterrupted();
	
				// Prepare list for holding data
				int assaySpots = assay.getNumSpots();
				List<SpotData> data = new ArrayList<SpotData>(assaySpots);
				
				// Load spot data for this bioassay
				short bioassayColumn = assay.getDataCubeColumnNo();
				query.setParameter("bioAssayColumn", (int)bioassayColumn, Type.INT);
				
				DynamicResultIterator it = query.iterate(dc);
				int positionIndex = it.getIndex(VirtualColumn.POSITION.getName());
				int ch1Index = it.getIndex("ch1");
				int ch2Index = it.getIndex("ch2");
				int blockIndex = it.getIndex("block");
				
				// Copy bioassay data to SpotData objects
				while (it.hasNext())
				{
					checkInterrupted();
					SqlResult r = it.next();
					SpotData spot = new SpotData(r.getInt(positionIndex),
						r.getFloat(ch1Index), r.getFloat(ch2Index), r.getInt(blockIndex));
					data.add(spot);
				}
				it.close();
				
				// Continue with next bioassay if there is no data
				int dataSize = data.size();
				if (dataSize == 0)
				{
					/* #### CONTINUE-STATEMENT #### */
					continue;
				}
				// Get range of block numbers - NOTE! query must return spots sorted in block order
				int minBlock = data.get(0).block;
				int maxBlock = data.get(data.size()-1).block;
				
				int fromIndex = 0;
				int toIndex = 0;
				int fromBlock = minBlock;
				// Normalize each block range independently: fromBlock + blockGroupSize --> toBlock
				while (fromBlock <= maxBlock)
				{
					checkInterrupted();
					// Find start and end index for current block range
					int toBlock = fromBlock + blockGroupSize - 1;
					if (toBlock > maxBlock) toBlock = maxBlock;
					if (toBlock < fromBlock) toBlock = fromBlock;
					fromIndex = toIndex;
					// Data is sorted by block; find index of last spot with: block <= toBlock
					// spot given by toIndex should not be included
					while (toIndex < dataSize && data.get(toIndex).block <= toBlock)
					{
						++toIndex;
					}
					
					if (toIndex > fromIndex)
					{
						// Exlude low and high values
						int count = toIndex - fromIndex;
						int lowIndex = fromIndex + (int)Math.floor(count * lowExclude / 100);
						int highIndex = toIndex - (int)Math.floor(count * highExclude / 100);
						
						if (highIndex > lowIndex)
						{
							List <SpotData> toNormalize = data.subList(fromIndex, toIndex);
							Collections.sort(toNormalize);
							// Find median ratio in the current block range
							double median = median(toNormalize);
							
							// Correct all spot intensities
							double sqrtMedRatio = Math.sqrt(median);
							double invSqrtMedRatio = 1 / sqrtMedRatio;
							for (int j = fromIndex; j < toIndex; ++j)
							{
								checkInterrupted();
								SpotData spot = data.get(j);
								float newCh1 = (float)(spot.ch1 * invSqrtMedRatio);
								float newCh2 = (float)(spot.ch2 * sqrtMedRatio);
								batcher.insert(bioassayColumn, spot.position, newCh1, newCh2);
							}
						}
						normalizedSpots += toIndex - fromIndex;
						if (progress != null) progress.display((int)((normalizedSpots * 100L) / numSpots), normalizedSpots + " spots normalized");
					}
					fromBlock = toBlock + 1;
				}
			}
			batcher.flush();
			batcher.close();
		}
		catch (SQLException e)
		{
			throw new BaseException(e);
		}
		return child;
	}
	
	private RequestInformation getConfigureJobParameters()
	{
		DbControl dc = null;
		try
		{
			if (configureJob == null)
			{
				dc = sc.newDbControl();
				BioAssaySet bas = getCurrentBioAssaySet(dc);
				int nofChannels=bas.getRawDataType().getChannels();

				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>(7);

				// 2-channel default value
				FloatParameterType fpt=new FloatParameterType(0F , null, 0F, true);
				if (nofChannels==1) fpt=new FloatParameterType(null , null, 0F, true);
				minIntensityParameter = new PluginParameter<Float>
				(
					minIntensityParameter.getName(),
					minIntensityParameter.getLabel(),
					minIntensityParameter.getDescription(),
					fpt
					);
				
				// Source and child bioassay set parameters
				parameters.add(getSourceBioAssaySetParameter(null, null));
				parameters.add(getChildNameParameter(null, null, bas.getName()));
				parameters.add(getChildDescriptionParameter(null, null, null));				
				
				// MediaRatio options
				parameters.add(medianRatioSection);
				parameters.add(minIntensityParameter);
				parameters.add(highExcludeParameter);
				parameters.add(lowExcludeParameter);
				String description="Set minimum intensity, low exclude, high exclude, and select method";
				if (nofChannels == 1) {
					parameters.add(methodParameter);
				}
				else if (nofChannels == 2)
				{
					parameters.add(blockGroupParameter);
					description=("Set minimum intensity, low exclude, high exclude, " +
											 "and block group parameter");
				}
				configureJob = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Configure job",
					description,
					parameters
				);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}
	
	private float median(float[] sortedArray, int first, int last)
	{
		int length=(last-first+1);
		int mid=first+length/2;
		return (length%2==0 ? (sortedArray[mid-1]+sortedArray[mid])/2 : sortedArray[mid]);
	}

	private double median(List<SpotData> data)
	{
		int size = data.size();
		int half = size / 2;
		if (size % 2 == 0)
		{
			return Math.sqrt(data.get(half - 1).ratio * data.get(half).ratio);
		}
		else
		{
			return data.get(half).ratio;
		}
	}
	
	
	private static class SpotData
		implements Comparable<SpotData>
	{
		final int position;
		final float ch1;
		final float ch2;
		final double ratio;
		final int block;

		SpotData(int position, float ch1, float ch2, int block)
		{
			this.position = position;
			this.ch1 = ch1;
			this.ch2 = ch2;
			this.block = block;
			this.ratio = ch1 / ch2;
		}
		/**
			Sort by ratio in ascending order.
		*/
		public int compareTo(SpotData o)
		{
			return this.ratio < o.ratio ? -1 : 1;
		}
	}
}
