/*
	$Id: ReporterMapFlatFileImporter.java 4985 2009-06-23 09:27:27Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArrayDesignBlock;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BlockInfo;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.FeatureBatcher;
import net.sf.basedb.core.FeatureIdentificationMethod;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.FileType;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.ReporterBatcher;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.data.FeatureData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.parser.ConfigureByExample;
import net.sf.basedb.util.parser.ConstantMapper;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Collections;
import java.util.Set;

/**
	A plugin that imports features to an ArrayDesign from a flat file. The ArrayDesign
	will not have any connections to plates.
	
	@base.modified $Date: 2009-06-23 11:27:27 +0200 (Tue, 23 Jun 2009) $
	@author Enell
	@version 2.0
*/
public class ReporterMapFlatFileImporter
	extends AbstractFlatFileImporter
	implements InteractivePlugin, ConfigureByExample
{

	private static final About about = 
		new AboutImpl
		(
			"Reporter map importer",
			"This plugin is used to import features to an ArrayDesign from a simple flat file.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2006, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);

	private static final Set<GuiContext> guiContexts = 
		Collections.singleton(new GuiContext(Item.ARRAYDESIGN, GuiContext.Type.ITEM));

	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	private static final StringParameterType requiredColumnMapping = new StringParameterType(255, null, true);
	private static final StringParameterType optionalColumnMapping = new StringParameterType(255, null, false);
	
	private static final PluginParameter<String> featureIdentificationParameter = 
		new PluginParameter<String>(
			"featureIdentification",
			"Identify features by",
			"Choose which method to use for identifying features: \n" +
			"COORDINATES: Use block, meta-grid, row and column coordinates\n" +
			"POSITION: Use a position number only\n" +
			"FEATURE_ID: Use the feature ID value (string)\n" +
			"In all cases, the identifier value must be unique.",
			new StringParameterType(255, "COORDINATES", true, 1, 0, 0, 
				Arrays.asList(new String[] {"COORDINATES", "POSITION", "FEATURE_ID"}))
		);
	
	private static final PluginParameter<String> reporterIdColumnMapping = new PluginParameter<String>(
		"reporterIdColumnMapping",
		"Reporter ID",
		"Mapping that picks the reporter's external ID from the data columns. " +
		"For example: \\ID\\",
		requiredColumnMapping
		);
	
	private static final PluginParameter<String> featureIdColumnMapping = new PluginParameter<String>(
		"featureIdColumnMapping",
		"Feature ID",
		"Mapping that picks the feature's ID from the data columns. This column is " +
		"only used when the array design uses the FEATURE_ID method for " +
		"identifying features. In the other cases, the value is just stored as it is." +
		"For example: \\Feature ID\\",
		optionalColumnMapping
		);

	private static final PluginParameter<String> positionColumnMapping = new PluginParameter<String>(
		"positionColumnMapping",
		"Position",
		"Mapping that picks the feature's position from the data columns. This column is only " +
		"used when the array design uses the POSITION method " +
		"for identifying features. The other identification methods will always " +
		"overwrite the position with an autogenerated value." +
		"Mapping example: \\Position\\",
		optionalColumnMapping
		);
	
	
	private static final PluginParameter<String> blockColumnMapping = new PluginParameter<String>(
		"blockColumnMapping",
		"Block",
		"Mapping that picks the feature's block number from the data columns. This column is only " +
		"used when the array design uses the COORDINATES method " +
		"for identifying features. In the other cases, the value is just stored as it is. " +
		"You must specify either this mapping or mappings for the meta coordinates when using " +
		"the COORDINATES method. " +
		"Example: \\Block\\",
		optionalColumnMapping
		);

	private static final PluginParameter<String> metaGridXColumnMapping = new PluginParameter<String>(
		"metaGridXColumnMapping",
		"Meta grid X",
		"Mapping that picks the feature's meta grid X coordinate from the data columns. This column is only " +
		"used when the array design uses the COORDINATES method " +
		"for identifying features. In the other cases, the value is just stored as it is. " +
		"You must specify either this mapping or mappings for the block number when using " +
		"the COORDINATES method. Example: \\Meta grid X\\",
		optionalColumnMapping
		);

	private static final PluginParameter<String> metaGridYColumnMapping = new PluginParameter<String>(
		"metaGridYColumnMapping",
		"Meta grid Y",
		"Mapping that picks the feature's meta grid Y coordinate from the data columns. This column is only " +
		"used when the array design uses the COORDINATES method " +
		"for identifying features. In the other cases, the value is just stored as it is. " +
		"You must specify either this mapping or mappings for the block number when using " +
		"the COORDINATES method. Example: \\Meta grid Y\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> columnColumnMapping = new PluginParameter<String>(
		"columnColumnMapping",
		"Column",
		"Mapping that picks the feature's column position in a block from the data columns. " +
		"This column is only used when the array design uses the COORDINATES method " +
		"for identifying features. In the other cases, the value is just stored as it is. " +
		"This value is required when using the COORDINATES method. Example: \\Column\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> rowColumnMapping = new PluginParameter<String>(
		"rowColumnMapping",
		"Row",
		"Mapping that picks the feature's row position in a block from the data columns. " +
		"This column is only used when the array design uses the COORDINATES method " +
		"for identifying features. In the other cases, the value is just stored as it is. " +
		"This value is required when using the COORDINATES method. Example: \\Row\\",
		optionalColumnMapping
		);
	
	protected static final PluginParameter<String> missingReporterErrorParameter = new PluginParameter<String>(
		"missingReporterError",
		"Reporter not found",
		"How to handle errors that are caused by a reporter not beeing present in the datbase. If not specified the " +
			"default error handling is used.\n\n"+
			"null = Import the data but set the reporter to null\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
					Arrays.asList( new String[] { "null", "skip", "fail"} ))
		);
	
	protected static final PluginParameter<String> duplicateFeatureErrorParameter = new PluginParameter<String>(
		"duplicateFeatureError",
		"Duplicate feature",
		"How to handle errors that are caused by more than one feature having the same " +
			"identifier. If not specified the default error handling is used.\n\n" +
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
					Arrays.asList( new String[] {"skip", "fail"} ))
		);

	
	private RequestInformation configurePlugin;
	private RequestInformation configureJob;
	
	private DbControl dc;
	private FeatureBatcher batcher;
	private ReporterBatcher reporterBatcher;
	private FileSetMember reporterMapMember;
	private ArrayDesign arrayDesign;
	private Map<BlockInfo, ArrayDesignBlock> blocks;
	private FlatFileParser ffp;

	private ItemParameterType<ArrayDesign> arrayDesignType;
	private PluginParameter<ArrayDesign> arrayDesignParameter;
	
	private List<PluginParameter<String>> allColumnMappings;
	
	private boolean errorIfNotFound;
	private Mapper reporterMapper;
	private Mapper featureIdMapper;
	private Mapper positionMapper;
	private Mapper blockMapper;
	private Mapper columnMapper;
	private Mapper rowMapper;
	private Mapper metaGridXMapper;
	private Mapper metaGridYMapper;

	private int numFeatures;
	private int numBlocks;
	
	/**
		Create a new importer.
	*/
	public ReporterMapFlatFileImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	public About getAbout()
	{
		return about;
	}
	public boolean supportsConfigurations()
	{
		return true;
	}
	public boolean requiresConfiguration()
	{
		return true;
	}
	/**
	 	Request write access to ArrayDesign:s, read access to Reporter:s,
	 	DataFileType:s, Platform:s and File:s.
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.ARRAYDESIGN, null, EnumSet.of(Permission.WRITE)));
			permissions.add(new Permissions(Item.REPORTER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.PLATFORM, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DATAFILETYPE, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	// -------------------------------------------
	
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	/**
		Return a set containing the element [ARRAYDESIGN, ITEM].
	*/
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}

	/**
		Returns null if the item is a {@link ArrayDesign} that doesn't already have 
		any features and isn't an affy design.
		@throws PermissionDeniedException If the array design already has features
			or if the logged in user doesn't have write permission
	*/
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;
		if (item == null)
		{
			message = "The object is null";
		}
		else if (!(item instanceof ArrayDesign))
		{
			message = "The object is not an ArrayDesign: " + item;
		}
		else
		{
			ArrayDesign ad = (ArrayDesign)item;
			if (ad.isFileOnlyPlatform())
			{
				message = "Array design is using a file-only platform";
			}
			else if (ad.getNumDbFeatures() > 0)
			{
				throw new PermissionDeniedException("The array design already has features");
			}
			else
			{
				ad.checkPermission(Permission.WRITE);
			}
		}
		return message;
	}
	/**
		The {@link Request#COMMAND_CONFIGURE_PLUGIN} command will ask for
		parser regular expressions and column mappings.
		The {@link Request#COMMAND_CONFIGURE_JOB} command will ask for
		a file and the array design that features should be added to.
	*/
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigurePluginParameters(context);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
			ItemContext fileContext = sc.getCurrentContext(Item.FILE);
			fileContext.setPropertyFilter(FileType.getPropertyFilter(FileType.REPORTER_MAP));
			fileContext.setPropertyFilter(getPrimaryLocationFilter());
		}
		return requestInformation;
	}
	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_PLUGIN} and
		{@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
			{
				RequestInformation ri = getConfigurePluginParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				
				// Check that enough mappings have been set for the 
				// specific feature identification method
				String fiParameter = (String)request.getParameterValue("featureIdentification");
				FeatureIdentificationMethod fiMethod = fiParameter == null ?
					FeatureIdentificationMethod.COORDINATES :
					FeatureIdentificationMethod.valueOf(fiParameter);
				
				List<Throwable> mapErrors = new ArrayList<Throwable>();
				if (fiMethod == FeatureIdentificationMethod.COORDINATES)
				{
					String blockMapping = (String)request.getParameterValue("blockColumnMapping");
					String metaGridXMapping = (String)request.getParameterValue("metaGridXColumnMapping");
					String metaGridYMapping = (String)request.getParameterValue("metaGridYColumnMapping");
					String columnMapping = (String)request.getParameterValue("columnColumnMapping");
					String rowMapping = (String)request.getParameterValue("rowColumnMapping");
					
					if (blockMapping == null && (metaGridXMapping == null || metaGridYMapping == null))
					{
						mapErrors.add(new InvalidDataException("You must at least map either block or meta coordinates to columns."));
					}
					if (columnMapping == null)	mapErrors.add(new InvalidUseOfNullException("Column"));
					if (rowMapping == null)	mapErrors.add(new InvalidUseOfNullException("Row"));
				}
				else if (fiMethod == FeatureIdentificationMethod.POSITION)
				{
					String positionMapping = (String)request.getParameterValue("positionColumnMapping");
					if (positionMapping == null) mapErrors.add(new InvalidUseOfNullException("Position"));
				}
				else if (fiMethod == FeatureIdentificationMethod.FEATURE_ID)
				{
					String positionMapping = (String)request.getParameterValue("featureIdColumnMapping");
					if (positionMapping == null) mapErrors.add(new InvalidUseOfNullException("Feature ID"));
				}
				else
				{
					response.setError("Unknown feature identification method: " + fiMethod, null);
					return;
				}
				if (mapErrors.size() > 0)
				{
					response.setError("Missing column mapping for feature identification method: " 
						+ fiMethod, mapErrors);
					return;
				}
				
				// Parser settings
				storeValue(configuration, request, headerRegexpParameter);
				storeValue(configuration, request, dataHeaderRegexpParameter);
				storeValue(configuration, request, dataSplitterRegexpParameter);
				storeValue(configuration, request, trimQuotesParameter);
				storeValue(configuration, request, ignoreRegexpParameter);
				storeValue(configuration, request, dataFooterRegexpParameter);
				storeValue(configuration, request, minDataColumnsParameter);
				storeValue(configuration, request, maxDataColumnsParameter);
				storeValue(configuration, request, ri.getParameter(Parameters.CHARSET_PARAMETER));

				// Column mappings
				storeValue(configuration, request, featureIdentificationParameter);
				storeValue(configuration, request, complexMappings);
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter<String> pp : getAllColumnMappings())
				{
					String mapExpression = (String)request.getParameterValue(pp.getName());
					checkColumnMapping(mapExpression, allowComplex, pp.getLabel());
					storeValue(configuration, pp, mapExpression);
				}

				response.setDone("Plugin configuration complete");
			}
			else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				storeValue(job, request, arrayDesignParameter);
				storeValue(job, request, fileParameter);
				storeValue(job, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
			
				// Error handling parameters
				storeValue(job, request, defaultErrorParameter);
				storeValue(job, request, missingReporterErrorParameter);
				storeValue(job, request, duplicateFeatureErrorParameter);
				storeValue(job, request, invalidUseOfNullErrorParameter);
				storeValue(job, request, numberFormatErrorParameter);
				storeValue(job, request, stringTooLongErrorParameter);
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------

	/*
		From the AbstractFlatFileReporter class
		-------------------------------------------
	*/
	/**
		Create a {@link DbControl} and a {@link FeatureBatcher}. 
		Check that the {@link ArrayDesign} doesn't already have features.
	*/
	@Override
	protected void begin(FlatFileParser ffp)
		throws BaseException
	{
		super.begin(ffp);
		dc = sc.newDbControl();
		arrayDesign = (ArrayDesign)job.getValue("arrayDesign");
		File reporterMapFile = (File)job.getValue("file");
		arrayDesign = ArrayDesign.getById(dc, arrayDesign.getId());
		if (arrayDesign.getNumDbFeatures() > 0)
		{
			throw new PermissionDeniedException("The array design already has features. "+arrayDesign.getName()+"["+arrayDesign.getId()+"]");
		}
		String fiParameter = (String)configuration.getValue("featureIdentification");
		FeatureIdentificationMethod fiMethod = fiParameter == null ?
			FeatureIdentificationMethod.COORDINATES :
			FeatureIdentificationMethod.valueOf(fiParameter);
		reporterMapMember = FileStoreUtil.setGenericDataFile(dc, arrayDesign, 
				FileType.REPORTER_MAP, DataFileType.GENERIC_REPORTER_MAP, reporterMapFile);
		batcher = arrayDesign.getFeatureBatcher(fiMethod, job.getJob());
		reporterBatcher = ReporterBatcher.getNew(dc);
		this.ffp = ffp;
		// Error handling for "missing reporter"
		String method = (String)job.getValue("missingReporterError");
		errorIfNotFound = !"null".equals(method);
		if (method != null)
		{
			addErrorHandler(ItemNotFoundException.class, new SimpleErrorHandler("skip".equals(method)));
		}
		// Error handling for "duplicate feature"
		method = (String)job.getValue("duplicateFeatureError");
		if (method != null)
		{
			addErrorHandler(ItemAlreadyExistsException.class, new SimpleErrorHandler("skip".equals(method)));
		}
		if (job.getValue("file") != null)
		{
			reporterBatcher.setUpdateSource("File: " + ((File)job.getValue("file")).getName());
		}
		
		numFeatures = 0;
		numBlocks = 0;
		blocks = new HashMap<BlockInfo, ArrayDesignBlock>();
	}

	/**
		Initialise column <code>Mapper</code>:s.
	*/
	@Override
	protected void beginData()
	{
		boolean cropStrings = ("crop".equals(job.getValue("stringTooLongError")));
		Mapper nullMapper = new ConstantMapper((String)null);
		reporterMapper = getMapper(ffp, (String)configuration.getValue("reporterIdColumnMapping"), 
			cropStrings ? ReporterData.MAX_EXTERNAL_ID_LENGTH : null, nullMapper);
		featureIdMapper = getMapper(ffp, (String)configuration.getValue("featureIdColumnMapping"), 
			cropStrings ? FeatureData.MAX_EXTERNAL_ID_LENGTH : null, nullMapper);
		positionMapper = getMapper(ffp, (String)configuration.getValue("positionColumnMapping"), null, nullMapper);
		blockMapper = getMapper(ffp, (String)configuration.getValue("blockColumnMapping"), null, nullMapper);
		columnMapper = getMapper(ffp, (String)configuration.getValue("columnColumnMapping"), null, nullMapper);
		rowMapper = getMapper(ffp, (String)configuration.getValue("rowColumnMapping"), null, nullMapper);
		metaGridXMapper = getMapper(ffp, (String)configuration.getValue("metaGridXColumnMapping"), null, nullMapper);
		metaGridYMapper = getMapper(ffp, (String)configuration.getValue("metaGridYColumnMapping"), null, nullMapper);
	}	
	
	@Override
	protected void handleData(FlatFileParser.Data data) 
		throws BaseException
	{
		String externalId = reporterMapper.getValue(data);
		String featureId = featureIdMapper.getValue(data);
		ReporterData reporter = externalId == null ? 
			null : reporterBatcher.getByExternalId(externalId, errorIfNotFound);
		BlockInfo bi;

		Integer block = blockMapper.getInt(data);
		Integer metaGridX = metaGridXMapper.getInt(data);
		Integer metaGridY = metaGridYMapper.getInt(data);
		
		bi = new BlockInfo(block == null ? 0 : block, 
				metaGridX == null ? 0 : metaGridX, 
				metaGridY == null ? 0 : metaGridY);

		ArrayDesignBlock adb = blocks.get(bi);
		if (adb == null)
		{
			adb = arrayDesign.addArrayDesignBlock(bi);
			blocks.put(bi, adb);
			numBlocks++;
		}
		
		FeatureData feature = batcher.newFeature(adb, reporter);
		feature.setExternalId(featureId);
		Integer position = positionMapper.getInt(data);
		Integer row = rowMapper.getInt(data);
		Integer column = columnMapper.getInt(data);
		feature.setPosition(position == null ? 0 : position);
		if (row != null)
		{
			feature.setRow(row);
			if (row > adb.getBlockSizeY()) adb.setBlockSizeY(row);
		}
		if (column != null)
		{
			feature.setColumn(column);
			if (column > adb.getBlockSizeX()) adb.setBlockSizeX(feature.getColumn());
		}
		
		batcher.insert(feature);
		numFeatures++;
	}

	/**
		Close and commit/rollback the FeatureBatcher and DbControl.
	*/
	@Override
	protected void end(boolean success)
		throws BaseException
	{
		if (blocks != null) blocks.clear();
		if (reporterMapMember != null) reporterMapMember.setValid(success, null);
		try
		{
			if (batcher != null) batcher.close();
			if (reporterBatcher != null) reporterBatcher.close();
			if (dc != null && success)
			{
				dc.commit();
			}
		}
		finally
		{
			if (dc != null) dc.close();
			super.end(success);
		}
	}
	/**
		Return <code>x features inserted; y blocks inserted</code>.
	*/
	protected String getSuccessMessage(int skippedLines)
	{
		String skipped = skippedLines > 0 ? "; " + skippedLines + " spot(s) skipped due to errors" : "";
		return numFeatures + " feature(s) inserted; "
			+ numBlocks + " block(s) inserted"
			+ skipped;
	}
	// -------------------------------------------

	private List<PluginParameter<String>> getAllColumnMappings()
	{
		if (allColumnMappings == null)
		{
			allColumnMappings = new ArrayList<PluginParameter<String>>();
			allColumnMappings.add(reporterIdColumnMapping);
			allColumnMappings.add(featureIdColumnMapping);
			allColumnMappings.add(positionColumnMapping);
			allColumnMappings.add(blockColumnMapping);
			allColumnMappings.add(metaGridXColumnMapping);
			allColumnMappings.add(metaGridYColumnMapping);
			allColumnMappings.add(columnColumnMapping);
			allColumnMappings.add(rowColumnMapping);
		}
		return allColumnMappings;
	}
	
	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob == null)
		{
			arrayDesignType = new ItemParameterType<ArrayDesign>(ArrayDesign.class, null, true, 1, null);
			arrayDesignParameter = new PluginParameter<ArrayDesign>(
				"arrayDesign",
				"Array Design",
				"The array design assigned to the imported features",
				arrayDesignType);
			
			// Parameters for CONFIGURE_JOB
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			parameters.add(arrayDesignParameter);

			ArrayDesign design = null;
			int currentArrayDesignId = sc.getCurrentContext(Item.ARRAYDESIGN).getId();
			List<File> reporterMapFiles = null;
			dc = sc.newDbControl();
			try
			{
				design = ArrayDesign.getById(dc, currentArrayDesignId);
				reporterMapFiles = FileStoreUtil.getGenericDataFiles(dc, design, FileType.REPORTER_MAP);
			}
			catch (Throwable t)
			{}
			finally
			{
				if (dc != null) dc.close();
			}
			
			// The reporter map file to import from - if a file already hase
			// been attached to the array design use it as a default choice
			PluginParameter<File> fileParameter = new PluginParameter<File>(
				"file",
				"Reporter map file",
				"The file with reporter information to import features from",
				new FileParameterType(reporterMapFiles == null || reporterMapFiles.isEmpty() ? 
					null : reporterMapFiles.get(0), true, 1)
			);
			parameters.add(fileParameter);
			parameters.add(Parameters.charsetParameter(null, null, 
					(String)configuration.getValue(Parameters.CHARSET_PARAMETER)));
			
			// Error handling parameters
			parameters.add(errorSection);
			parameters.add(defaultErrorParameter);
			parameters.add(missingReporterErrorParameter);
			parameters.add(duplicateFeatureErrorParameter);
			parameters.add(invalidUseOfNullErrorParameter);
			parameters.add(numberFormatErrorParameter);
			parameters.add(stringTooLongErrorParameter);
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Select file to import features from",
				"Here you select which file to import the features from.",
				parameters
			);
		}
		return configureJob;
	}

	private RequestInformation getConfigurePluginParameters(GuiContext context)
	{
		if (configurePlugin == null)
		{
			// Parameters object for CONFIGURE_PLUGIN
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			// Parser regular expressions
			parameters.add(parserSection);
			parameters.add(headerRegexpParameter);
			parameters.add(dataHeaderRegexpParameter);
			parameters.add(dataSplitterRegexpParameter);
			parameters.add(trimQuotesParameter);
			parameters.add(ignoreRegexpParameter);
			parameters.add(dataFooterRegexpParameter);
			parameters.add(minDataColumnsParameter);
			parameters.add(maxDataColumnsParameter);
			parameters.add(Parameters.charsetParameter(null, null, null));
			
			// Column mappings
			parameters.add(mappingSection);
			parameters.add(featureIdentificationParameter);
			parameters.add(complexMappings);
			parameters.addAll(getAllColumnMappings());

			configurePlugin = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_PLUGIN,
				"Parser settings",
				"Enter the regular expressions used to parse the text file " +
				"and the column mappings used to find matching properties for the features.",
				parameters
			);
		}
		return configurePlugin;
	}
	
}
