/**
	$Id: ArrayBatchImporter.java 5249 2010-02-17 12:38:58Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.ProtocolType;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing array batch items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2010-02-17 13:38:58 +0100 (Wed, 17 Feb 2010) $
*/
public class ArrayBatchImporter
	extends AbstractItemImporter<ArrayBatch>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.ARRAYBATCH, GuiContext.Type.LIST));

	private static final About about =
		new AboutImpl
	(
			"Array batch importer",
			"Imports and updates array batches in a batch.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2008, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
	);

	
	protected static final PluginParameter<String> hardwareColumnMapping = new PluginParameter<String>(
			"hardwareColumnMapping",
			"Print robot",
			"Mapping that picks the name or ID of the print robot from the data columns. " +
			"The plug-in will first try to find a robot with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Robot\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> arrayDesignColumnMapping = new PluginParameter<String>(
			"arrayDesignColumnMapping",
			"Array design",
			"Mapping that picks the name or ID of the array design from the data columns. " +
			"The plug-in will first try to find a design with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Array design\\",
			optionalColumnMapping
			);
	
	public ArrayBatchImporter()
	{}

	/*
		From the Plugin interface
		------------------------------------
	*/
	@Override
	public About getAbout()
	{
		return about;
	}
	// ------------------------------------

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper protocolMapper;
	private Mapper hardwareMapper;
	private Mapper arrayDesignMapper;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME,  
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	/**
		Adds column mappings for name, description,
		protocol, robot, and array design
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(arrayDesignColumnMapping);
		parameters.add(protocolColumnMapping);
		parameters.add(hardwareColumnMapping);
	}
	
	/**
		Calls {@link ArrayBatch#getQuery()}.
	*/
	@Override
	protected ItemQuery<ArrayBatch> createItemQuery()
	{
		return ArrayBatch.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? ArrayBatch.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? ArrayBatch.MAX_DESCRIPTION_LENGTH : null, null);
		protocolMapper = getMapper(ffp, (String)job.getValue("protocolColumnMapping"), null, null);
		hardwareMapper = getMapper(ffp, (String)job.getValue("hardwareColumnMapping"), null, null);
		arrayDesignMapper = getMapper(ffp, (String)job.getValue("arrayDesignColumnMapping"), null, null);
	}
	
	@Override
	protected ArrayBatch createItem(DbControl dc, FlatFileParser.Data data)
	{
		ArrayDesign design = null;
		if (arrayDesignMapper != null) 
		{
			design = findArrayDesign(dc, FallbackIdMethod.NAME_OR_ID, arrayDesignMapper.getValue(data));
		}

		ArrayBatch batch = ArrayBatch.getNew(dc, design);
		updateItem(dc, batch, data);
		return batch;
	}
	
	@Override
	protected void updateItem(DbControl dc, ArrayBatch batch, FlatFileParser.Data data)
	{
		if (nameMapper != null) batch.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) batch.setDescription(descriptionMapper.getValue(data));
		if (protocolMapper != null) 
		{
			String nameOrId = protocolMapper.getValue(data);
			ProtocolType type = ProtocolType.getById(dc, SystemItems.getId(ProtocolType.PRINTING));
			Protocol protocol = findProtocol(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			if (nameOrId == null || protocol != null) batch.setProtocol(protocol);
		}
		if (hardwareMapper != null)
		{
			String nameOrId = hardwareMapper.getValue(data);
			Hardware robot = findPrintRobot(dc, FallbackIdMethod.NAME_OR_ID, nameOrId);
			if (nameOrId == null || robot != null) batch.setPrintRobot(robot);
		}
	}
	// ------------------------------------
}
