/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.executor;

import java.io.IOException;
import java.util.Arrays;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.AbstractPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;

/**
	Abstract base class that is useful for plug-ins that need to implement
	the {@link IOSupport} interface and don't want to be visible as 
	standalone plug-ins. 

	@author nicklas
	@since 2.15
	@see ExternalProgramExecutor
	@base.modified $Date $
*/
public abstract class AbstractIOSupport
	extends AbstractPlugin
	implements IOSupport
{
	
	private ExternalProgramExecutor executor;
	
	public AbstractIOSupport()
	{}
	
	/*
		From Plugin & InteractivePlugin interfaces
		------------------------------------------
	*/
	/**
		@return Always FALSE
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	/**
		@return Always FALSE
	*/
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
		@return Always null
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return null;
	}
	/**
		@return Always null
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		return null;
	}
	
	/**
		Accepts the {@link IOSupport#COMMAND_EXECUTE_IO_JOB} command. The default
		implementation starts a new transaction and then calls 
		{@link #performIO(DbControl, ProgressReporter)} which should be implemented by the 
		subclass.
	*/
	@Override
	public void run(Request request, Response response, ProgressReporter progress)
	{
		String command = request.getCommand();
		if (command.equals(IOSupport.COMMAND_EXECUTE_IO_JOB))
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				String message = performIO(dc, progress);
				dc.commit();
				response.setDone(message);
			}
			catch (Throwable t)
			{
				response.setError(t.getMessage(), Arrays.asList(t));
			}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		else
		{
			response.setError("Unknown command: "+command, Arrays.asList(new Throwable()));
		}		
	}
	// --------------------------------------------
	/*
		From the IOSupport interface
		----------------------------
	*/
	@Override
	public void init(SessionControl sc, ParameterValues configuration, 
			ParameterValues job, ExternalProgramExecutor executor)
	{
		super.init(sc, configuration, job);
		this.executor = executor;
	}
	// ----------------------------

	/**
		Perform the IO operation.
		@param dc A DbControl that can be used by the subclass
		@param progress An optional progress reporter
		@return A message that is returned as to the external program executor
	*/
	protected abstract String performIO(DbControl dc, ProgressReporter progress)
		throws IOException;
	
	
	/**
		Get the external program executor that invoked the plug-in.
	*/
	protected ExternalProgramExecutor getExecutor()
	{
		return executor;
	}

}
