/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.executor;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.plugins.BfsExporterPlugin;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.bfs.AnnotationWriter;
import net.sf.basedb.util.bfs.FilenameGenerator;
import net.sf.basedb.util.bfs.GenericDiskDataWriterFactory;
import net.sf.basedb.util.bfs.MetadataWriter;
import net.sf.basedb.util.bfs.SequenceFilenameGenerator;
import net.sf.basedb.util.bfs.SuggestedFilenameGenerator;
import net.sf.basedb.util.export.spotdata.AnnotationAssayField;
import net.sf.basedb.util.export.spotdata.AssayField;
import net.sf.basedb.util.export.spotdata.BfsExporter;
import net.sf.basedb.util.export.spotdata.DynamicField;
import net.sf.basedb.util.export.spotdata.ExportableFieldConverter;
import net.sf.basedb.util.export.spotdata.ExportableFieldFactory;
import net.sf.basedb.util.export.spotdata.MatrixBfsExporter;
import net.sf.basedb.util.export.spotdata.SerialBfsExporter;
import net.sf.basedb.util.export.spotdata.StandardFieldConverter;
import net.sf.basedb.util.formatter.IdentifiableFormatter;
import net.sf.basedb.util.formatter.MultiFormatter;
import net.sf.basedb.util.formatter.NameableFormatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

/**
	Exporter implementation for the external program executor that
	exports data to BFS format.
	
	@author nicklas
	@since 2.15
	@base.modified $Date $
*/
public class BfsExporterSupport
	extends AbstractIOSupport
{

	private static final About about =
		new AboutImpl
		(
			"BFS exporter for the ExternalProgramExecutor",
			"Support plug-in for the external program executor that exports data to the " +
			"external program in BFS format. This plug-in can't be used as a standalone " +
			"plug-in.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2010, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);

	
	public BfsExporterSupport()
	{}

	/*
		From the Plugin interface
		-------------------------
	*/
	@Override
	public About getAbout()
	{
		return about;
	}
	@Override
	public MainType getMainType()
	{
		return Plugin.MainType.EXPORT;
	}	
	// -------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		try
		{
			String command = request.getCommand();
			if (command.equals(IOSupport.COMMAND_CONFIGURE_IO_PLUGIN))
			{
				RequestInformation ri = getConfigureExecutorParameters(command);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found " +
														"in the request", errors);
					return;
				}
				
				// Check that fields are unique
				try
				{
					checkUniqueFields((String)request.getParameterValue("spotFields"), "spot fields");
					checkUniqueFields((String)request.getParameterValue("reporterFields"), "reporter fields");
					checkUniqueFields((String)request.getParameterValue("assayFields"), "assay fields");
				}
				catch (InvalidDataException ex)
				{
					response.setError(ex.getMessage(), null);
					return;
				}
				
				// Export options
				storeValue(configuration, request, ri.getParameter(BfsExporterPlugin.PARAMETER_FILEFORMAT));
				storeValue(configuration, request, ri.getParameter(BfsExporterPlugin.PARAMETER_FILENAMEGENERATOR));
				storeValue(configuration, request, ri.getParameter("metadataFile"));
				
				storeValue(configuration, request, ri.getParameter("spotFields"));
				storeValue(configuration, request, ri.getParameter("reporterFields"));
				storeValue(configuration, request, ri.getParameter("assayFields"));
				response.setDone("Executor configuration complete");
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}

	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(IOSupport.COMMAND_CONFIGURE_IO_PLUGIN))
		{
			requestInformation = getConfigureExecutorParameters(command);
		}
		return requestInformation;
	}
	// -----------------------------------------

	private RequestInformation configureExecutor;
	
	private RequestInformation getConfigureExecutorParameters(String command)
	{
		if (configureExecutor != null) return configureExecutor;
	
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			// File format parameter
			parameters.add(BfsExporterPlugin.getFileFormatParameter());
			// Name of the metadata file
			parameters.add(new PluginParameter<String>(
				"metadataFile", 
				"Metadata filename", 
				"The name of the metadata file name. The defult value is 'stdin.txt' which is " +
				"is automatically sent to the program on standard input. It is possible to use " +
				"other filenames but the program is then responsible to open the file itself. " +
				"Other files are created automatically based on the selected file name generation " +
				"strategy",
				new StringParameterType(255, "stdin.txt", true))
			);
			// File name generation strategy parameter
			parameters.add(BfsExporterPlugin.getFileNameGeneratorParameter());
			
			// Exported fields
			parameters.add(new PluginParameter<String>(
					"reporterFields",
					"Reporter fields",
					"Name all reporter fields that you wish to include in the export. " +
					"Use one value per line. The exporter will try to find matching " +
					"fields as follows:\n" +
					" 1. Extra value\n" +
					" 2. Formula\n" +
					" 3. Static properties: Name, ExternalId, etc.",
					new StringParameterType(65535, null, false)
				));

			parameters.add(new PluginParameter<String>(
					"spotFields",
					"Spot fields",
					"Name all spot fields that you wish to include in the export. " +
					"Use one value per line. The exporter will try to find matching " +
					"fields as follows:\n" +
					" 1. Extra value\n" +
					" 2. Formula\n" +
					" 3. Raw data property\n" +
					" 4. Channel intensity: Ch 1, Ch 2, etc.",
					new StringParameterType(65535, null, false)
				));

			parameters.add(new PluginParameter<String>(
					"assayFields",
					"Assay fields",
					"Name all assay fields that you wish to include in the export. " +
					"Use one value per line. The exporter will try to find matching " +
					"fields as follows:\n" +
					" 1. Experimental factor\n" +
					" 2. Assay annotation\n" +
					" 3. Static properties: Name, Description or Id\n",
					new StringParameterType(65535, null, false)
				));
		
			configureExecutor = new RequestInformation(
					command,
					"Select export options.",
					"Select file format options and which fields to export. Additional " +
					"fields may be defined as job parameters and selected at runtime.",
					parameters
				);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureExecutor;
	}

	private void checkUniqueFields(String fields, String parameter)
	{
		if (fields == null) return;
		Set<String> all = new HashSet<String>();
		for (String fld : fields.split("\\n"))
		{
			if (!all.add(fld.trim())) 
			{
				throw new InvalidDataException("Duplicate entry in " + parameter + ": " + fld);
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	protected String performIO(DbControl dc, ProgressReporter progress)
		throws IOException
	{
		// Get job configuration parameters
		ExternalProgramExecutor executor = getExecutor();
		BioAssaySet source = executor.getCurrentBioAssaySet(dc);
		java.io.File workDir = executor.getWorkingDirectory();
		String format = (String)configuration.getValue(BfsExporterPlugin.PARAMETER_FILEFORMAT);
		String filenameGen = (String)configuration.getValue(BfsExporterPlugin.PARAMETER_FILENAMEGENERATOR);
		String metadataFile = Values.getString((String)configuration.getValue("metadataFile"), "stdin.txt");
		
		// Reporter fields
		ExportableFieldConverter fieldConverter = new StandardFieldConverter(dc, source);
		List<DynamicField> reporterFields = new ArrayList<DynamicField>();
		String tmp = (String)configuration.getValue("reporterFields");
		if (tmp != null)
		{
			for (String reporterField : tmp.split("\\n"))
			{
				reporterFields.add(fieldConverter.getReporterField(reporterField.trim(), true));
			}
		}
		// Spot fields
		List<DynamicField> spotFields = new ArrayList<DynamicField>();
		tmp = (String)configuration.getValue("spotFields");
		if (tmp != null)
		{
			for (String spotField : tmp.split("\\n"))
			{
				spotFields.add(fieldConverter.getSpotField(spotField.trim(), true));
			}
		}
		// Assay fields
		List<AssayField> assayFields = new ArrayList<AssayField>();
		tmp = (String)configuration.getValue("assayFields");
		if (tmp != null)
		{
			for (String assayField : tmp.split("\\n"))
			{
				assayFields.add(fieldConverter.getAssayField(assayField.trim(), true));
			}
		}

		List<ProgramParameter> programParameters = executor.getProgramParameters();
		
		// Create the BFS exporter
		if (progress != null) progress.display(0, "Exporting ...");
		BfsExporter exporter = null;
		if (SerialBfsExporter.SUBTYPE.equals(format))
		{
			exporter = new SerialBfsExporter();
		}
		else if (MatrixBfsExporter.SUBTYPE.equals(format))
		{
			exporter = new MatrixBfsExporter();
		}
		else
		{
			throw new BaseException("Unknown format. "+format);
		}
		exporter.setDbControl(dc);
		exporter.setProgressReporter(progress);
		exporter.setSource(source);
		
		// Create files & file factory
		java.io.File metadata = executor.createFile(metadataFile);
		exporter.setMetadataWriter(MetadataWriter.create(metadata));

		// Export parameter values
		MultiFormatter parameterFormatter = new MultiFormatter(new ToStringFormatter<Object>(), true);
		parameterFormatter.registerFormatter(Nameable.class, new NameableFormatter());
		parameterFormatter.registerFormatter(BioAssay.class, new IdentifiableFormatter());
		for (ProgramParameter pp : programParameters)
		{
			ParameterDefinition pdef = pp.getParameterDefinition();
			String name = pdef.getName();
			List values = pp.getValues();
			String parameterType = pdef.getType();
			if (values != null && values.size() > 0)
			{
				exporter.setParameter(name, Values.toStrings(values, parameterFormatter));
				if ("factor".equals(parameterType))
				{
					// Add annotation fields to the exporter
					for (Object annotationType : values)
					{
						AnnotationAssayField fld = ExportableFieldFactory.annotation((AnnotationType)annotationType, null, null);
						if (!assayFields.contains(fld)) assayFields.add(fld);
					}
				}
				else if ("spotformula".equals(parameterType))
				{
					// Add formula fields to the exporter
					for (Object formula : values)
					{
						DynamicField fld = ExportableFieldFactory.formula((Formula)formula, null, null);
						if (!spotFields.contains(fld)) spotFields.add(fld);
					}
				}
			}
		}
		
		// Reporter fields
		if (reporterFields.size() > 0)
		{
			AnnotationWriter reporterWriter = AnnotationWriter.create(executor.createFile("rdata.txt"));
			exporter.setReporterAnnotationsWriter(reporterWriter);
			exporter.addReporterFields(reporterFields);
		}
		
		// Spot fields
		if (spotFields.size() > 0)
		{
			FilenameGenerator filenameGenerator = "sequence".equals(filenameGen) ?
					new SequenceFilenameGenerator("sdata-", ".txt", 1) : 
					new SuggestedFilenameGenerator();
			exporter.setDataWriterFactory(
					new GenericDiskDataWriterFactory(workDir, filenameGenerator));
			exporter.addSpotFields(spotFields);
		}

		// Assay fields
		if (assayFields.size() > 0)
		{
			AnnotationWriter assayWriter = AnnotationWriter.create(executor.createFile("pdata.txt"));
			exporter.setAssayAnnotationsWriter(assayWriter);
			exporter.addAssayFields(assayFields);
		}

		exporter.doExport();
		return "Bioassay set " + source.getName() + " successfully exported to " + workDir;
	}
	
}
