/**
	$Id: Parameters.java 5271 2010-03-15 13:49:40Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.util;

import java.nio.charset.Charset;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;

import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.StringParameterType;

/**
	Utility class for creating some useful plug-in parameters.
	
	@author nicklas
	@version 2.6
	@base.modified $Date: 2010-03-15 14:49:40 +0100 (Mon, 15 Mar 2010) $
*/
public class Parameters
{

	/**
		Section definition for grouping error handling options.
	*/
	protected static final PluginParameter<String> errorSection = new PluginParameter<String>(
		"errorSection",
		"Error handling",
		"Select how errors should be handled by the plugin.",
		null
	);

	/**
		The name of the parameter created by {@link #errorSection(String, String)}.
	*/
	public static final String ERROR_SECTION = "errorSection";
	
	/**
		Create a parameter starting an error section.
		@param title The title of the section, or null for the default (Error handling)
		@param description A description, or null for the default
		@return A new plug-in parameter
	*/
	public static final PluginParameter<String> errorSection(String title, String description)
	{
		return new PluginParameter<String>(
			ERROR_SECTION,
			title != null ? title : "Error handling",
			description != null ? description : "Select how errors should be handled by the plugin.",
			null
		);
	}
	
	/**
		The name of the parameter created by {@link #defaultError(String, String, String, String...)}.
	*/
	public static final String DEFAULT_ERROR = "defaultError";
	
	/**
		Create the default error handling parameter.
		@param title The title for the parameter or null for the default (Default error handling)
		@param description The description or null for the default
		@param defaultValue The default value or null to use "fail"
		@param options Options to select from, if null or empty, the "fail" and "skip" are used
		@return A new plug-in parameter
	*/
	public static final PluginParameter<String> defaultError(String title, String description, 
		String defaultValue, String...options)
	{
		if (options == null || options.length == 0)
		{
			options = new String[] { "skip", "fail"};
		}
		if (defaultValue == null) defaultValue = "fail";
		if (title == null) title = "Default error handling";
		if (description == null) description = 
			"The default way to handle an error when no other error handling options have been " +
			"specified to take care of specific exceptions.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message (default)";
		
		return new PluginParameter<String>(
				DEFAULT_ERROR,	title,	description,
				new StringParameterType(255, defaultValue, true, 1, 0, 0, 
						Arrays.asList( options ))
			);
	}
	
	/**
		The name of the parameter created by {@link #stringTooLongError(String, String, String, String...)}.
	*/
	public static final String STRING_TOO_LONG_ERROR = "stringTooLongError";
	
	/**
		Create the parameter for handling too long strings.
		
		@param title The title for the parameter or null for the default (String too long)
		@param description The description or null for the default
		@param defaultValue The default value or null to use the default error handling
		@param options Options to select from, if null or empty, the "crop", "fail" and "skip" are used
		@return A new plug-in parameter
	*/
	public static final PluginParameter<String> stringTooLongError(String title, String description, 
		String defaultValue, String...options)
	{
		if (options == null || options.length == 0)
		{
			options = new String[] { "crop", "skip", "fail" };
		}
		if (title == null) title = "String too long";
		if (description == null) description = 
			"How to handle errors that are caused by too long string values. " +
			"If no value is specified the default error handling is used.\n\n"+
			"crop = Crop the string to the maximum allowed length\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message";
		
		return new PluginParameter<String>(
				STRING_TOO_LONG_ERROR,	title,	description,
				new StringParameterType(255, defaultValue, false, 1, 0, 0, 
						Arrays.asList( options ))
			);
	}	
	

	/**
		The name of the parameter created by {@link #numberFormatError(String, String, String, String...)}.
	*/
	public static final String NUMBER_FORMAT_ERROR = "numberFormatError";
	
	/**
		Create the parameter for handling invalid numbers.
		
		@param title The title for the parameter or null for the default (Invalid numeric value)
		@param description The description or null for the default
		@param defaultValue The default value or null to use the default error handling
		@param options Options to select from, if null or empty, the "null", "fail" and "skip" are used
		@return A new plug-in parameter
	*/
	public static final PluginParameter<String> numberFormatError(String title, String description, 
		String defaultValue, String...options)
	{
		if (options == null || options.length == 0)
		{
			options = new String[] { "null", "skip", "fail" };
		}
		if (title == null) title = "Invalid numeric value";
		if (description == null) description = 
			"How to handle errors that are caused by strings that can't be converted " +
			"to a numeric value. If no value is specified the default error handling is used.\n\n"+
			"null = Insert a null value and continue\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message";
		
		return new PluginParameter<String>(
				NUMBER_FORMAT_ERROR,	title,	description,
				new StringParameterType(255, defaultValue, false, 1, 0, 0, 
						Arrays.asList( options ))
			);
	}
	
	
	/**
		The name of the parameter created by {@link #invalidUseOfNullError(String, String, String, String...)}.
	*/
	public static final String INVALID_USE_OF_NULL_ERROR = "invalidUseOfNullError";
	
	/**
		Create the parameter for handling required but missing values.
		
		@param title The title for the parameter or null for the default (Missing a required value)
		@param description The description or null for the default
		@param defaultValue The default value or null to use the default error handling
		@param options Options to select from, if null or empty, the "fail" and "skip" are used
		@return A new plug-in parameter
	*/
	public static final PluginParameter<String> invalidUseOfNullError(String title, String description, 
		String defaultValue, String...options)
	{
		if (options == null || options.length == 0)
		{
			options = new String[] { "skip", "fail" };
		}
		if (title == null) title = "Missing a required value";
		if (description == null) description = 
			"How to handle errors that are caused by missing a required value. If no value " +
			"is specified the default error handling is used.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message";
		
		return new PluginParameter<String>(
				INVALID_USE_OF_NULL_ERROR,	title, description,
				new StringParameterType(255, defaultValue, false, 1, 0, 0, 
						Arrays.asList( options ))
			);
	}	
	
	/**
		The name of the parameter created by {@link #numberOutOfRangeError(String, String, String, String...)}.
	*/
	public static final String NUMBER_OUT_OF_RANGE_ERROR = "numberOutOfRangeError";
	
	/**
		Create the parameter for handling numeric values that
		are outside the allowed range.
		
		@param title The title for the parameter or null for the default (Numeric value out of range)
		@param description The description or null for the default
		@param defaultValue The default value or null to use the default error handling
		@param options Options to select from, if null or empty, the "fail" and "skip" are used
		@return A new plug-in parameter
	*/
	public static final PluginParameter<String> numberOutOfRangeError(String title, String description, 
		String defaultValue, String...options)
	{
		if (options == null || options.length == 0)
		{
			options = new String[] { "skip", "fail" };
		}
		if (title == null) title = "Numeric value out of range";
		if (description == null) description = 
			"How to handle errors that are caused by numbers that are outside the allowed" +
			"range. If no value is specified the default error handling is used.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message";
		
		return new PluginParameter<String>(
				NUMBER_OUT_OF_RANGE_ERROR,	title, description,
				new StringParameterType(255, defaultValue, false, 1, 0, 0, 
						Arrays.asList( options ))
			);
	}
	
	/**
		The name of the parameter that is created by {@link #charsetParameter(String, String, String)}
	*/
	public static final String CHARSET_PARAMETER = "charset";
	
	/**
		Enumerates all available character sets.
		@see #charsetParameter(String, String, String)
	*/
	private static final StringParameterType charsetType = 
		new StringParameterType(255, Config.getCharset(), false, 1, 0, 0, 
		new ArrayList<String>(Charset.availableCharsets().keySet()));
		
	
	/**
		Parameter definition that asks for the character encoding used in the
		file that is imported. This can be both a job parameter and a configuration
		parameter. The implementation first checks in the job parameter and
		if not found in the configuration parameter. If still not found the
		default charset is used as specified by {@link Config#getCharset()}.
		@param title The title to use for the parameter or null to use the default label
			(Character set)
		@param description The description to use for the parameter or null to use the
			default description
		@param defaultValue The default value for the character set or null to use
			the system default
	*/
	public static final PluginParameter<String> charsetParameter(String title, String description, 
			String defaultValue)
	{
		if (title == null) title = "Character set";
		if (description == null) description = 
			"The character set to use when reading the file. This setting overrides " +
			"the character set specified by the file. If neither this parameter nor the " +
			"file specifies a character set, the system default is used " +
			"(" + Config.getCharset() + ").";
		if (defaultValue == null) defaultValue = Config.getCharset();
		return new PluginParameter<String>(
				CHARSET_PARAMETER, title, description, defaultValue, charsetType
		);
	}
	
	
	/**
		The name of the parameter that is created by 
		{@link #decimalSeparatorParameter(String, String, String)}
	*/
	public static final String DECIMAL_SEPARATOR_PARAMETER = "decimalSeparator";
	
	/**
		Parameter definition that asks for the decimal separator used by numeric values 
		when parsed from strings. If no specific value is used number parsing is done by 
		Float.valueOf or Double.valueOf.
		@param title The title to use for the parameter or null to use the default title
			(Decimal separator)
		@param description The description to use for the parameter or null to use the
			default description
		@param defaultValue The default value for the decimal separator
	*/
	public static final PluginParameter<String> decimalSeparatorParameter(String title, String description, 
		String defaultValue)
	{
		if (title == null) title = "Decimal separator";
		if (description == null) description = 
			"The decimal separator used in numeric values, " +
			"if not specified dot is assumed.";
		
		return new PluginParameter<String>(
				DECIMAL_SEPARATOR_PARAMETER, title, description, defaultValue, 
				new StringParameterType(
						255, "dot", false, 1, 0, 0,
						Arrays.asList(new String[] { "dot", "comma" })
				)
		);
	}

	/**
		The name of the parameter that is created by 
		{@link #decimalSeparatorParameter(String, String, String)}
		@since 2.8
	*/
	public static final String DATE_FORMAT_PARAMETER = "dateFormat";
	
	/**
		Parameter definition that asks for the format used by date values when parsed from
		strings. See {@link SimpleDateFormat} for more info about template values.
		@param title The title to use for the parameter or null to use the default title
			(Date format)
		@param description The description to use for the parameter or null to use the
			default description
		@param defaultValue The default value for the date format
		@since 2.8
	*/
	public static final PluginParameter<String> dateFormatParameter(String title, String description, 
		String defaultValue)
	{
		if (title == null) title = "Date format";
		if (description == null) description = 
			"The format used for date values, if not specified, yyyy-MM-dd is assumed.\n\n" +
			"<b>Common patterns</b>: \n" +
			"yy = two-digit year\n" +
			"yyyy = four-digit year\n" +
			"MM = numerical month\n" +
			"MMM = name of month\n" +
			"dd = day in month\n" +
			"See http://java.sun.com/javase/6/docs/api/java/text/SimpleDateFormat.html for a complete list\n\n" +
			"<b>Examples</b>:\n" +
			"yyyy-MM-dd: international standard\n" +
			"MM/dd/yy: american\n" +
			"dd/MM/yy: british\n" +
			"See http://en.wikipedia.org/wiki/Calendar_date " +
			"for more information.";
		
		return new PluginParameter<String>(
				DATE_FORMAT_PARAMETER, title, description, defaultValue, 
				new StringParameterType(
					255, "yyyy-MM-dd", false)
				);
	}
	
	/**
		The name of the parameter that is created by 
		{@link #logFileParameter(String, String, String)}
		@since 2.8
	*/
	public static final String LOGFILE_PARAMETER = "logFile";
	
	/**
		Parameter definition that asks for the path to a log file that
		a plug-in can use for logging detailed information about it's
		progress.
		
		@param title The title to use for the parameter or null to use the default title
			(Log file)
		@param description The description to use for the parameter or null to use the
			default description
		@param defaultValue The default path
		@since 2.8
	*/
	public static PluginParameter<String> logFileParameter(String title, String description, String defaultValue)
	{
		if (title == null) title = "Log file";
		if (description == null)
		{
			description = "Specify a file path for logging detailed information about the plug-ins " +
				"work. Logging will be disabled if this parameter is empty.";
		}
		return new PluginParameter<String>(
				LOGFILE_PARAMETER, title, description, defaultValue,
				new PathParameterType(Path.Type.FILE, defaultValue, false)
		);
	}

	/**
		The name of the parameter that is created by 
		{@link #logFileParameter(String, String, String)}
		@since 2.8
	*/
	public static final String DRY_RUN_PARAMETER = "dryRun";

	/**
		Parameter definition that enable/disable dry-run of a plug-in. A dry run
		should do everything the plug-in normally does, except that the result should
		not be committed to the database. A dry-run should be combined with
		a log file (see {@link #logFileParameter(String, String, String)}.
		
		@param title The title to use for the parameter or null to use the default title
			(Dry run)
		@param description The description to use for the parameter or null to use the
			default description
		@param defaultValue The default value
		@since 2.8
	*/
	public static PluginParameter<Boolean> dryRunParameter(String title, String description, Boolean defaultValue)
	{
		if (title == null) title = "Dry run";
		if (description == null)
		{
			description = "Enables dry-run of the plug-in. It will do it's work as normal, " +
					"but not commit the result to the database. Enabling this option is mostly " +
					"useful if a 'Log file' is also selected.";
		}
		return new PluginParameter<Boolean>(
				DRY_RUN_PARAMETER, title, description, defaultValue,
				new BooleanParameterType()
		);
	}
	
	/**
		Clone a plug-in parameter but give it a different default value.
		
		@param pp The plug-in parameter to clone
		@param defaultValue The new default value
		@return The cloned plug-in parameter
		@since 2.15
	*/
	public static <T> PluginParameter<T> cloneParameter(PluginParameter<T> pp, T defaultValue)
	{
		PluginParameter<T> clone = new PluginParameter<T>(
				pp.getName(), 
				pp.getLabel(),
				pp.getDescription(), 
				defaultValue, 
				pp.getParameterType());
		return clone;
	}
	
	
	/**
		The name of the parameter that is created by
		{@link #childNameParameter(String, String, String)}
		@since 2.15
	*/
	public static final String CHILD_NAME = "childName";
	
	/**
		Create a plugin parameter that asks for the name of the child bioassay set.
		@param label The label to use for the parameter or null to use the default label
			(Child name)
		@param description The description to use for the parameter or null to use the
			default description (The name of the child bioassay set)
		@param defaultValue The default value for the child name or null to use
			the default value (New bioassay set)
		@return A plugin parameter
		@since 2.15
	*/
	public static PluginParameter<String> childNameParameter(String label, String description, String defaultValue)
	{
		if (label == null) label = "Child name";
		if (description == null) description = "The name of the child bioassay set";
		if (defaultValue == null) defaultValue = "New bioassay set";
		return new PluginParameter<String>
			(
				CHILD_NAME, label, description, defaultValue, 
				new StringParameterType(Nameable.MAX_NAME_LENGTH, null, true)
			);
	}
	
	/**
		The name of the parameter that asks is created by
		{@link #childDescriptionParameter(String, String, String)}
		@since 2.15
	*/
	public static final String CHILD_DESCRIPTION = "childDescription";
	
	/**
		Create a plugin parameter that asks for the description of the child bioassay set.
		@param label The label to use for the parameter or null to use the default label
			(Child description)
		@param description The description to use for the parameter or null to use the
			default description (An optional description of the child bioassay set)
		@param defaultValue The default value for the child name or null to not use any
			default value
		@return A plugin parameter
		@since 2.15
	*/
	public static PluginParameter<String> childDescriptionParameter(String label, String description, String defaultValue)
	{
		if (label == null) label = "Child description";
		if (description == null) description = "An optional description of the child bioassay set";
		return new PluginParameter<String>
			(
				CHILD_DESCRIPTION, label, description, defaultValue, 
				new StringParameterType(BioAssaySet.MAX_DESCRIPTION_LENGTH, null, false)
			);
	}
	
	/**
		The name of the parameter that asks created by
		{@link #transformationNameParameter(String, String, String)}
		@since 2.15
	*/
	public static final String TRANSFORMATION_NAME = "transformationName";
	
	/**
		Create a plugin parameter that asks for the name of the transformation.
		@param label The label to use for the parameter or null to use the default label
			(Transformation name)
		@param description The description to use for the parameter or null to use the
			default description (The name of the transformation)
		@param defaultValue The default value for the transformation name or null to not
			use any default
		@return A plugin parameter
		@since 2.15
	*/
	public static PluginParameter<String> transformationNameParameter(String label, String description, String defaultValue)
	{
		if (label == null) label = "Transformation name";
		if (description == null) description = "The name of the transformation";
		return new PluginParameter<String>
			(
					TRANSFORMATION_NAME, label, description, defaultValue, 
					new StringParameterType(Nameable.MAX_NAME_LENGTH, null, true)
			);
	}

	
}
