/*
	$Id$

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.ConsoleProgressReporter;
import net.sf.basedb.util.IntensityCalculator;
import net.sf.basedb.util.IntensityCalculatorUtil;

public class TestAnalyzeUtil
{
	
	private static Map<String, Integer> itemIds = new HashMap<String, Integer>();
	private static boolean isSetup = false;
	
	static int getItemId(String key)
	{
		Integer id = itemIds.get(key);
		return id == null ? 0 : id;
	}
	
	static void setItemId(String key, int id)
	{
		if (id != 0)
		{
			itemIds.put(key, id);
		}
	}
	
	static boolean hasItemId(String key)
	{
		return itemIds.containsKey(key);
	}
	
	static boolean setupExperiment()
	{
		if (isSetup) return false;
		write("++Setting up experiment");
		// Import reporters
		TestReporter.test_import_from_file("data/test.rawdata.import.txt", 
				"\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*", "\\t", 4, 3);
		
		// Create array design
		int arrayDesignId = TestArrayDesign.test_create(Platform.GENERIC, false);
		TestArrayDesign.test_import_from_file(arrayDesignId, "data/test.rawdata.import.txt", 
			"\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*", null, 0, 2, 1, 4);
		setItemId("arraydesign.generic", arrayDesignId);
		
		// Create raw bioassays and import raw data
		int rba1 = TestRawBioAssay.test_create(Platform.GENERIC, "genepix", "Genepix.1", 0, 0, 0, arrayDesignId, 1, false);
		int rba2 = TestRawBioAssay.test_create(Platform.GENERIC, "genepix", "Genepix.2", 0, 0, 0, arrayDesignId, 1, false);
		TestRawBioAssay.test_import_from_file(rba1, "data/test.rawdata.import.txt");
		TestRawBioAssay.test_import_from_file(rba2, "data/test.rawdata.import.txt");
		setItemId("rawbioassay.generic.1", rba1);
		setItemId("rawbioassay.generic.2", rba2);

		// Experimental factors
		int drugFactor = TestAnnotationType.test_create("Drug resistance", Type.STRING, 0, null, null, Item.RAWBIOASSAY, 1, new String[] { "high", "medium", "low" }, false);
		int tempFactor = TestAnnotationType.test_create("Temperature", Type.FLOAT, 0, null, null, Item.RAWBIOASSAY, 1, null, false);
		TestAnnotation.test_annotatate(Item.RAWBIOASSAY, rba1, drugFactor, 0, "medium");
		TestAnnotation.test_annotatate(Item.RAWBIOASSAY, rba2, drugFactor, 0, "high");
		TestAnnotation.test_annotatate(Item.RAWBIOASSAY, rba1, tempFactor, 0, 25.0);
		TestAnnotation.test_annotatate(Item.RAWBIOASSAY, rba2, tempFactor, 0, 30.0);
		setItemId("factor.drug-resistance", drugFactor);
		setItemId("factor.temperature", tempFactor);
		
		// The experiment
		int experiment1 = TestExperiment.test_create("genepix", true);
		TestExperiment.test_add_experimental_factors(experiment1, drugFactor, tempFactor);
		TestExperiment.test_add_rawbioassay(experiment1, rba1, rba2);
		setItemId("experiment.genepix", experiment1);
		
		// Some other useful stuff - extra value types
		int intExtraId = TestExtraValueType.test_create("test.int", Type.INT, false);
		int floatExtraId = TestExtraValueType.test_create("test.float", Type.FLOAT, false);
		int stringExtraId = TestExtraValueType.test_create("test.string", Type.STRING, false);
		setItemId("extravaluetype.int", intExtraId);
		setItemId("extravaluetype.float", floatExtraId);
		setItemId("extravaluetype.string", stringExtraId);
		
		// formulas
		int formulaId = TestFormula.test_create("Mean FG", Formula.Type.INTENSITY_EXPRESSION,
				"genepix", new String[] { "raw('ch1FgMean')", "raw('ch2FgMean')" }, false);
		setItemId("formula.intensity.genepix.mean", formulaId);

		// create a root bioassay set
		int rootId = test_create_root_bioassayset(experiment1, "mean", "Default root");
		setItemId("bioassayset.root", rootId);

		isSetup = true;
		return true;
	}
	
	
	static int test_create_root_bioassayset(int experimentId, String intensityFormula, String name)
	{
		if (experimentId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		long time;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, experimentId);
			
			List<RawBioAssay> rbaCollection = e.getRawBioAssays().list(dc);
			
			RawDataType rdt = e.getRawDataType();
			IntensityCalculator iCalc = IntensityCalculatorUtil.createJepIntensityCalculator(
				dc, rdt, rdt.getIntensityFormula(intensityFormula).getExpressions());
			
			time = System.currentTimeMillis();
			BioAssaySet root = IntensityCalculatorUtil.createRootBioAssaySet(
				dc, e, rbaCollection, null, iCalc, TestUtil.getSilent() ? null : new ConsoleProgressReporter());
			root.setName(name);
			root.getTransformation().setName(name);
			dc.commit();

			time = System.currentTimeMillis()-time;
			int numSpots = root.getNumSpots();
			id = root.getId();
			write("--Create root bioassayset using calculator OK ("+numSpots+" spots inserted; "+time+" ms)");
		}
		catch (Throwable ex)
		{
			write("--Create root bioassayset using calculator FAILED");
			ex.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static int clone_bioassay_set(int bioassayset_id, String cloneName)
	{
		if (bioassayset_id == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet bas = BioAssaySet.getById(dc, bioassayset_id);
			
			Transformation t = bas.newTransformation(null);
			t.setName(cloneName);
			BioAssaySet clone = t.newProduct(null, null, true);
			clone.setName(cloneName);
			dc.saveItem(t);
			dc.saveItem(clone);
			dc.commit();
			id = clone.getId();

			write("--Clone bioassayset OK ("+cloneName+")");
		}
		catch (Throwable ex)
		{
			write("--Clone bioassayset OK ("+cloneName+")");
			ex.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
		
	}
	
	static boolean test_delete_bioassayset(int bioAssaySetId)
	{
		if (bioAssaySetId == 0) return true;
		boolean ok = true;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet bas = BioAssaySet.getById(dc, bioAssaySetId);
			dc.deleteItem(bas);
			dc.commit();
			write("--Delete bioassayset OK");
		}
		catch (Throwable ex)
		{
			write("--Delete bioassayset FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return ok;
	}
	
	static void cleanupExperiment()
	{
		if (!isSetup) return;
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		write("--Cleaning up experiment");
		
		TestExperiment.test_delete(getItemId("experiment.genepix"));
		TestRawBioAssay.test_delete(getItemId("rawbioassay.generic.1"));
		TestRawBioAssay.test_delete(getItemId("rawbioassay.generic.2"));
		
		TestAnnotationType.test_delete(getItemId("factor.drug-resistance"));
		TestAnnotationType.test_delete(getItemId("factor.temperature"));
		TestExtraValueType.test_delete(getItemId("extravaluetype.int"));
		TestExtraValueType.test_delete(getItemId("extravaluetype.float"));
		TestExtraValueType.test_delete(getItemId("extravaluetype.string"));
		TestFormula.test_delete(getItemId("formula.intensity"));

		TestArrayDesign.test_delete(getItemId("arraydesign.generic"));
		
		TestReporter.test_delete();
	}
	
	static int getBioAssaySetIdByName(int experimentId, String name)
	{
		if (experimentId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, experimentId);
			ItemQuery<BioAssaySet> query = e.getBioAssaySets();
			query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
			List<BioAssaySet> list = query.list(dc);
			if (list.size() > 0) id = list.get(0).getId();
			dc.commit();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static void write(String message)
	{
		System.out.println(message);
	}
}
