/*
  $Id: TestAnnotationTypeCategory.java 5060 2009-08-19 07:02:11Z nicklas $

  	Copyright (C) 2006 Nicklas Nordborg, Dominic Oyeniran
  	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

import net.sf.basedb.core.*;

import java.util.Date;

/**
	Tests the implementation of annotation type category.
	@author Dominic Oyeniran
	@version 2.0

	@see AnnotationTypeCategory
*/
public class TestAnnotationTypeCategory 
{

	static boolean ok = true;
	
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing annotation type categories");
		write_header();
		// Standard tests: create, load, list annotation type categories
		int id = test_create(true);
		int id2 = test_create(false);
		test_load(id);

		test_list(-1);

		// extra test: add categories to annotation type
		int id_int = TestAnnotationType.test_create(null, Type.INT, 0, 10, null, Item.SAMPLE, 1, null, true);
		test_add_categories(id_int, id, id2);
		test_list_categories(id_int, 2);

		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_remove_categories(id_int, id, id2);
		TestAnnotationType.test_delete(id_int);

		// Standard test: Delete
		test_delete(id);
		test_delete(id2);
		write("++Testing annotation type categories "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.ANNOTATIONTYPECATEGORY)) return 0;
		int id = 0;
		DbControl dc = null;
		try 
		{
			dc = TestUtil.getDbControl();
			AnnotationTypeCategory atgc = AnnotationTypeCategory.getNew(dc);
			if (setAll)
			{
				atgc.setName("Test Category");
				atgc.setDescription("Added at " + new Date());
			}
			dc.saveItem(atgc);
			dc.commit();
			dc=TestUtil.getDbControl();
			dc.reattachItem(atgc, false);
			id=atgc.getId();
			write_item(id, atgc);
			write("--Create annotation type category OK");
		}
		catch (Throwable ex)
		{
			write("--Create annotation type category FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally 
		{
			if (dc != null) dc.close();
		}

		return id; 
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			AnnotationTypeCategory atgc = AnnotationTypeCategory.getById(dc, id);
			write_item(0, atgc);
			write("--Load annotation type category OK");
		}
		catch (Throwable ex)
		{
			write("--Load annotation type category FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}

	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<AnnotationTypeCategory> list = AnnotationTypeCategory.getQuery().list(dc);
			for (int i = 0; i<list.size(); i++)
			{
				write_item(i, list.get(i));
			}
			if (expectedResults >= 0 && expectedResults != list.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+list.size());
			}
			write("--List annotation types categories OK ("+list.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List annotation types categories FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			AnnotationTypeCategory at = AnnotationTypeCategory.getById(dc, id);
			dc.deleteItem(at);
			dc.commit();
			write("--Delete annotation type category OK");
		}
		catch (Throwable ex)
		{
			write("--Delete annotation type category FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}


	static void write(String message)
	{
		System.out.println(message);
	}

	static void write_header ()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription");
			write("-- \t-- \t--------- \t-----------");
		}
	}

	static void write_item(int i, AnnotationTypeCategory atgc)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
			write(i+":\t"+atgc.getId()+"\t"+atgc.getName()+"\t"+atgc.getDescription());
	}


	static void test_add_categories(int annotationTypeId, int... categoryIds)
	{
		if (annotationTypeId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			AnnotationType at = AnnotationType.getById(dc, annotationTypeId);
			for (int categoryId : categoryIds)
			{
				if (categoryId != 0)
				{
					at.addCategory(AnnotationTypeCategory.getById(dc, categoryId));
				}
			}
			dc.commit();
			write("--Add categories to annotation type OK");

		}
		catch (Throwable ex)
		{
			write("--Add categories to annotation type FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_remove_categories(int annotationTypeId, int... categoryIds)
	{
		if (annotationTypeId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			AnnotationType at = AnnotationType.getById(dc, annotationTypeId);
			for (int categoryId : categoryIds)
			{
				if (categoryId != 0)
				{
					at.removeCategory(AnnotationTypeCategory.getById(dc, categoryId));
				}
			}
			dc.commit();
			write("--Remove categories to annotation type OK");

		}
		catch (Throwable ex)
		{
			write("--Remove categories to annotation type FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_categories(int annotationTypeId, int expectedResults)
	{
		if (annotationTypeId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			AnnotationType at = AnnotationType.getById(dc, annotationTypeId);
			ItemResultList<AnnotationTypeCategory> list = at.getCategories().list(dc);
			for (int i = 0; i<list.size(); i++)
			{
				write_item(i, list.get(i));
			}
			if (expectedResults >= 0 && expectedResults != list.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+list.size());
			}
			write("--List categories for annotation type OK ("+list.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List categories for annotation type FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

}

