/*
	$Id$
	
	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.Date;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.Trashcan;

public class TestArrayBatch
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	public static boolean test_all()
	{
		write("++Testing array batch");
		write_header();
		// Standard tests: create, load, list
		int arraydesign_id = TestArrayDesign.test_create(Platform.GENERIC, false);
		int id = test_create(arraydesign_id, true);
		int id2 = test_create(arraydesign_id, false);
		test_load(id);
		test_list(-1);

		// Extra tests:

		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();

		// Standard test: Delete
		test_delete(id);
		test_delete(id2);
		TestArrayDesign.test_delete(arraydesign_id);
		write("++Testing array batch "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemQuery<ArrayBatch> q = ArrayBatch.getQuery();
			ItemResultList<ArrayBatch> l = q.list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}

			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List array batches OK ("+l.size()+")");
		}
		catch (Exception ex)
		{
			write("--List array batches FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ArrayBatch ab = ArrayBatch.getById(dc, id);
			dc.deleteItem(ab);
			dc.commit();
			write("--Delete array batch OK");
		}
		catch (Exception ex)
		{
			write("--Delete array batch FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ArrayBatch ab = ArrayBatch.getById(dc, id);
			write_item(0, ab);
			write("--Load array batch OK");
		}
		catch (Exception ex)
		{
			write("--Load array batch FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	public static int test_create(int arraydesign_id, boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.ARRAYBATCH)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ArrayDesign ad = ArrayDesign.getById(dc, arraydesign_id);
			ArrayBatch ab = ad.newArrayBatch();
			if (setAll)
			{
				ab.setName("Test array batch");
				ab.setDescription("Added at "+new Date());
			}
			dc.saveItem(ab);
			dc.commit();
			id = ab.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(ab, false);
			write_item(0, ab);
			write("--Create array batch OK");
		}
		catch (Throwable  ex)
		{
			write("--Create array batch FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<ArrayBatch> l = ArrayBatch.getQuery().list(dc);
			Trashcan.delete(dc.getSessionControl(), l, true, null);
			dc.commit();
			write("--Delete all array batches OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all array batches FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tArray design");
			write("-- \t-- \t--------- \t-----------\t------------");
		}
	}
	
	static void write_item(int i, ArrayBatch ab)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+ab.getId()+"\t"+ab.getName()+"\t"+
			ab.getDescription()+"\t"+ab.getArrayDesign());
	}
	
	static void write(String message)
	{
		System.out.println(message);
	}
}
