/*
	$Id: TestBaseFileExporterImporter.java 5277 2010-03-22 12:45:37Z nicklas $
	
	Copyright (C) 2010 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	at your option) any later version.
	
	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.ConsoleProgressReporter;
import net.sf.basedb.util.basefile.BaseFileWriter;
import net.sf.basedb.util.export.spotdata.BaseFileExporter;
import net.sf.basedb.util.export.spotdata.DynamicField;
import net.sf.basedb.util.export.spotdata.ExportableFieldFactory;
import net.sf.basedb.util.export.spotdata.MatrixBaseFileExporter;
import net.sf.basedb.util.export.spotdata.SerialBaseFileExporter;
import net.sf.basedb.util.importer.spotdata.BaseFileImporter;

import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;


public class TestBaseFileExporterImporter
{
	static boolean ok = true;
	
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		boolean wasSetup = TestAnalyzeUtil.setupExperiment();
		write("++Testing BaseFileExporter and BaseFileImporter");
		try
		{
			int experimentId = TestAnalyzeUtil.getItemId("experiment.genepix");
			int rootBasId = TestAnalyzeUtil.getItemId("bioassayset.root");
			if (experimentId == 0 || rootBasId == 0)
			{
				ok = false;
				write("Test experiment was not setup correctly.");
				return ok;
			}
			
			int fileSerialNoMerge = TestFile.test_create("serial.nomerge.txt", false, false);
			int fileSerialMerge = TestFile.test_create("serial.merge.txt", false, false);
			int fileMatrixNoMerge = TestFile.test_create("matrix.nomerge.txt", false, false);
			int fileMatrixMerge = TestFile.test_create("matrix.merge.txt", false, false);
			
			int exportRoot = TestAnalyzeUtil.clone_bioassay_set(rootBasId, "BaseFileExporter");
			int exportNoMerge = TestAnalyzeUtil.clone_bioassay_set(exportRoot, "No merge");
			int exportMerge = TestAnalyzeUtil.clone_bioassay_set(exportRoot, "Merge");
			test_export(exportNoMerge, fileSerialNoMerge, false, false);
			test_export(exportMerge, fileSerialMerge, false, true);
			test_export(exportNoMerge, fileMatrixNoMerge, true, false);
			test_export(exportMerge, fileMatrixMerge, true, true);
	
			test_import(exportNoMerge, fileSerialNoMerge, "Import serial BASEfile");
			test_import(exportNoMerge, fileMatrixNoMerge, "Import matrix BASEfile");
			test_import(exportMerge, fileSerialMerge, "Import merged serial BASEfile");
			test_import(exportMerge, fileMatrixMerge, "Import merged matrix BASEfile");
			
			if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
			TestAnalyzeUtil.test_delete_bioassayset(exportRoot);
	
			TestFile.test_delete(fileSerialNoMerge);
			TestFile.test_delete(fileSerialMerge);
			TestFile.test_delete(fileMatrixNoMerge);
			TestFile.test_delete(fileMatrixMerge);
		}
		finally
		{
			if (wasSetup) TestAnalyzeUtil.cleanupExperiment();
			write("++Testing BaseFileExporter and BaseFileImporter "+(ok ? "OK" : "Failed")+"\n");
		}
		return ok;
	}
	
	/**
		The BioAssaySet thats beeing tested must be the one created by this class.
	*/
	static void test_export(int bioAssaySetId, int fileId, 
			boolean matrix, boolean mergeReporters)
	{
		if (bioAssaySetId == 0 || fileId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet bas = BioAssaySet.getById(dc, bioAssaySetId);
			File file = File.getById(dc, fileId);
			ItemQuery<BioAssay> query = bas.getBioAssays();
			query.order(Orders.asc(Hql.property("name")));
			query.order(Orders.asc(Hql.property("id")));
			List<BioAssay> assays = new ArrayList<BioAssay>(query.list(dc));
			
			// Initialise exporter
			BaseFileExporter bfe = matrix ? new MatrixBaseFileExporter() : new SerialBaseFileExporter();
			bfe.setDbControl(dc);
			bfe.setSource(bas);
			bfe.setAverageOnReporter(mergeReporters);
			if (!TestUtil.getSilent()) bfe.setProgressReporter(new ConsoleProgressReporter());
			bfe.setWriter(new BaseFileWriter(new OutputStreamWriter(file.getUploadStream(false))));
			
			// Parameters that go in the 'settings' section
			bfe.setParameter("b", "some parameter");
			bfe.setParameter("a", "some other parameter");

			// Configure the reporter fields to export
			List<DynamicField> reporterFields = new ArrayList<DynamicField>();
			reporterFields.add(ExportableFieldFactory.simple(Dynamic.column(VirtualColumn.POSITION), "position", Type.INT, Formula.AverageMethod.NONE, null));
			reporterFields.add(ExportableFieldFactory.reporter("id", "reporter", Type.INT, Formula.AverageMethod.NONE, null));
			reporterFields.add(ExportableFieldFactory.reporter("name", null, Type.STRING, Formula.AverageMethod.NONE, null));
			bfe.addReporterFields(reporterFields);

			// Configure the spot fields to export
			ItemQuery<Formula> ratio = Formula.getQuery(Formula.Type.COLUMN_EXPRESSION, 
					bas.getRawDataType(), bas.getIntensityTransform());
				ratio.restrict(Restrictions.eq(Hql.property("name"), Expressions.string("Ratio, ch1 / ch2")));
				if (ratio.count(dc) < 1)
				{
					throw new Exception("Cant find formula 'Ratio, ch1 / ch2'");
				}
			
			List<DynamicField> spotFields = new ArrayList<DynamicField>();
			spotFields.add(ExportableFieldFactory.channel(1, "intensity1", null));
			spotFields.add(ExportableFieldFactory.channel(2, "intensity2", null));
			spotFields.add(ExportableFieldFactory.rawData("row", null, Type.INT, Formula.AverageMethod.NONE, null));
			spotFields.add(ExportableFieldFactory.rawData("flags", null, Type.INT, Formula.AverageMethod.NONE, null));
			spotFields.add(ExportableFieldFactory.formula(ratio.list(dc).get(0), null, null));
			bfe.addSpotFields(spotFields);
			
			bfe.doExport();
			FileStoreUtil.setDataFile(dc, bas, 
					matrix ? DataFileType.BASEFILE_MATRIX : DataFileType.BASEFILE_SERIAL, 
					file);
			if (!TestUtil.getSilent()) write("");
			dc.commit();
			
			write("--Exporting basefile [matrix:"+matrix+", mergeReporters:"+mergeReporters+"] OK");
		}
		catch (Throwable ex)
		{
			write("--Exporting basefile [matrix:"+matrix+", mergeReporters:"+mergeReporters+"] FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_import(int bioAssaySetId, int fileId, 
			String childName)
	{
		if (bioAssaySetId == 0 || fileId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet parent = BioAssaySet.getById(dc, bioAssaySetId);
			File file = File.getById(dc, fileId);
			
			Transformation t = parent.newTransformation(null);
			t.setName("Importing BASEfile: " + file.getName());
			dc.saveItem(t);
			
			BaseFileImporter importer = new BaseFileImporter();
			importer.setDbControl(dc);
			importer.setSourceFile(file);
			importer.setTransformation(t);
			if (!TestUtil.getSilent()) importer.setProgressReporter(new ConsoleProgressReporter());
			BioAssaySet child = importer.doImport();
			child.setName(childName);
			dc.commit();
			if (!TestUtil.getSilent()) write("");
			write("--Importing BASEfile for " + childName + " OK");
		}
		catch (Throwable ex)
		{
			write("--Importing BASEfile for " + childName + " FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
			

	static void write(String message)
	{
		System.out.println(message);
	}
	

	
}
