/*
	$Id: TestJob.java 5270 2010-03-15 10:42:35Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ParameterInfo;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginConfigurationRequest;
import net.sf.basedb.core.PluginExecutionRequest;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.PluginResponse;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.plugin.Response;

import net.sf.basedb.util.ConsoleProgressReporter;

import java.util.List;
import java.util.Arrays;

public class TestJob
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing job");
		write_header();
		int definitionId = TestPluginDefinition.test_create("NullPlugin", null);
		int configurationId = TestPluginConfiguration.test_create(definitionId);
		TestPluginConfiguration.test_configure(configurationId);
		
		// Standard tests: create, load, list
		int id = test_create(configurationId);
		test_list();
		test_load(id);
		
		// extra test: configure, execute
		test_configure(id);
		test_execute(id, false);
		
		// Standard test: Delete
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_delete(id);
		TestPluginConfiguration.test_delete(configurationId);
		TestPluginDefinition.test_delete(definitionId);
		write("++Testing job "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}
	
	static boolean test_execute(int id, boolean expectError)
	{
		if (id == 0) return ok;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Job job = Job.getById(dc, id);
			
			PluginExecutionRequest request = 
				job.execute(TestUtil.getSilent() ? null : new ConsoleProgressReporter(), TestUtil.getLocalIp());
			dc.close();
			PluginResponse response = request.invoke();
			Response.Status status = response.getStatus();
			if (expectError)
			{
				if (status != Response.Status.ERROR)
				{
					throw new BaseException("Expected job to fail. Got " + 
						status + "; " + response.getMessage());
				}
			}
			else
			{
				if (status == Response.Status.ERROR)
				{
					Throwable cause = response.getErrorList() != null &&
						response.getErrorList().size() > 0 ? response.getErrorList().get(0) : null;
					throw new BaseException(response.getMessage(), cause);
				}
			}
			
			dc = TestUtil.getDbControl();
			job = Job.getById(dc, id);
			if (!TestUtil.getSilent()) write("");
			write_item(0, job);
			write("--Execute job OK ("+response.getMessage()+")");
		}
		catch (Throwable ex)
		{
			write("--Execute job FAILED");
			ex.printStackTrace();
			ok = false;
			return false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return true;
	}

	static void test_configure(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Job job = Job.getById(dc, id);
			
			PluginConfigurationRequest pr = job.configure(null);
			pr.getRequestInformation();

			pr.setParameterValue("oneString", "aString");
			pr.setParameterValues("manyStrings", Arrays.asList("A", "C"));
			pr.setParameterValue("selectOneString", "B");
			pr.setParameterValues("selectMultipleStrings", Arrays.asList("One", "Three"));
			PluginResponse response = pr.invoke();
			if (response.getStatus() == Response.Status.DONE)
			{
				response.saveParameters(dc);
				dc.commit();
			}
			else
			{
				throw new BaseException(response.getMessage(), response.getErrorList().get(0));
			}
			
			dc = TestUtil.getDbControl();
			job = Job.getById(dc, id);

			List<?> value = job.getParameterValues("oneString");
			if (value == null || value.size() != 1 || !value.get(0).equals("aString"))
			{
				throw new Exception("Cant find the correct value for job parameter 'oneString'");
			}
			value = job.getParameterValues("manyStrings");
			if (value == null || value.size() != 2 || !value.contains("A") || !value.contains("C"))
			{
				throw new Exception("Cant find the correct value for job parameter 'manyStrings'");
			}
			value = job.getParameterValues("selectOneString");
			if (value == null || value.size() != 1 || !value.get(0).equals("B"))
			{
				throw new Exception("Cant find the correct value for job parameter 'oneString'");
			}
			value = job.getParameterValues("selectMultipleStrings");
			if (value == null || value.size() != 2 || !value.contains("One") || !value.contains("Three"))
			{
				throw new Exception("Cant find the correct value for job parameter 'manyStrings'");
			}

			write_item(0, job);
			write_values(job);
			write("--Configure job OK");
		}
		catch (Throwable ex)
		{
			write("--Configure job FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static int test_create(int configurationId)
	{
		if (configurationId == 0 || !TestUtil.hasPermission(Permission.CREATE, Item.JOB)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginConfiguration pc = PluginConfiguration.getById(dc, configurationId);
			Job job = Job.getNew(dc, pc.getPluginDefinition(), pc, null);
			job.setName(pc.getName());
			dc.saveItem(job);
			dc.commit();
			id = job.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(job, false);
			write_item(0, job);
			write("--Create job OK");
		}
		catch (Throwable ex)
		{
			write("--Create job FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
		return id;
	}
	
	static void test_list()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemQuery<Job> q = Job.getQuery();
			ItemResultList<Job> l = q.list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			write("--List job OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List job FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Job job = Job.getById(dc, id);
			write_item(0, job);
			write_values(job);
			write("--Load job OK");
		}
		catch (Throwable ex)
		{
			write("--Load job FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Job job = Job.getById(dc, id);
			dc.deleteItem(job);
			dc.commit();
			write("--Delete job OK");
		}
		catch (Throwable ex)
		{
			write("--Delete job FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tType  \tStatus\tStarted\tEnded\tConfiguration");
			write("-- \t-- \t--------- \t-----------\t------\t------\t-------\t-----\t-------------");
		}
	}
	
	static void write_item(int i, Job job)
		throws BaseException
	{
		if (!TestUtil.getSilent())
		{
			System.out.println(i+":\t"+job.getId()+"\t"+job.getName()+"\t"+job.getDescription()+"\t"+job.getJobType()+
				"\t"+job.getStatus()+"\t"+job.getStarted()+"\t"+job.getEnded()+"\t"+job.getPluginConfiguration());
		}
	}
	
	static void write_values(Job job)
		throws BaseException
	{
		if (!TestUtil.getSilent())
		{
			System.out.println("Job values:");
			for(String name : job.getParameterNames())
			{
				ParameterInfo pi = job.getParameterInfo(name);
				System.out.println((pi.getLabel() == null ? name : pi.getLabel()) + "=" + pi.getValues());
			}
		}
	}
	
	static void write_request_information(RequestInformation ri)
	{
		if (TestUtil.getSilent()) return;

		write("Request information for command: "+ri.getCommand());
		write("\tTitle:\t"+ri.getTitle());
		write("\tDescription:\t"+ri.getDescription());
		for (PluginParameter<?> pp : ri.getParameters())
		{
			ParameterType pt = pp.getParameterType();
			System.out.println("\tParameter:\t"+pp.getName() + "\t" + pp.getLabel() + "\t" + pt );
		}
	}

	
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void setRequestParameterValues(PluginConfigurationRequest request, Parameter... parameters)
	{
		// Set default values first
		RequestInformation ri = request.getRequestInformation();
		for (PluginParameter<?> pp : ri.getParameters())
		{
			request.setParameterValue(pp.getName(), pp.getDefaultValue());
		}
		// Then, set specified parameters
		if (parameters != null)
		{
			for (Parameter p : parameters)
			{
				if (ri.getParameter(p.name) != null)
				{
					request.setParameterValue(p.name, p.value);
				}
			}
		}
	}
	
	static class Parameter
	{
		final String name;
		final Object value;
		Parameter(String name, Object value)
		{
			this.name = name;
			this.value = value;
		}
	}

}
