/*
	$Id: TestPlateMapping.java 5060 2009-08-19 07:02:11Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.core.data.MappingCoordinate;
import java.util.Date;
import java.util.List;
import java.util.ArrayList;

public class TestPlateMapping
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing plate mapping");
		write_header();
		// Standard tests: create, load, list
		int sourceGeometryId = TestPlateGeometry.test_create(5, 5, false);
		int destinationGeometryId = TestPlateGeometry.test_create(10, 10, false);

		int id = test_create(sourceGeometryId, destinationGeometryId);
		test_load(id);
		test_list(-1);

		// Extra tests: apply mapping to plates
		int sourcePlateTypeId = TestPlateType.test_create(sourceGeometryId, false);
		int destinationPlateTypeId = TestPlateType.test_create(destinationGeometryId, false);
		int[] sourcePlateIds = new int[4];
		for (int i = 0; i < 4; i++)
		{
			sourcePlateIds[i] = TestPlate.test_create(sourcePlateTypeId, false);
		}
		int mappedId = test_apply_mapping(id, sourcePlateIds, destinationPlateTypeId);
		TestPlate.test_list(-1);

		// Extra test: list parents and children to plates
		test_list_parents(mappedId, 4);
		test_list_children(sourcePlateIds[0], 1);

		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		TestPlate.test_delete(mappedId);
		for (int i = 0; i < 4; i++)
		{
			TestPlate.test_delete(sourcePlateIds[i]);
		}
		TestPlateType.test_delete(destinationPlateTypeId);
		TestPlateType.test_delete(sourcePlateTypeId);

		// Standard test: Delete
		test_delete(id);
		TestPlateGeometry.test_delete(sourceGeometryId);
		TestPlateGeometry.test_delete(destinationGeometryId);

		write("++Testing plate mapping "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int sourceGeometryId, int destinationGeometryId)
	{
		if (sourceGeometryId == 0 || destinationGeometryId == 0 || 
			!TestUtil.hasPermission(Permission.CREATE, Item.PLATEMAPPING)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PlateGeometry source = PlateGeometry.getById(dc, sourceGeometryId);
			PlateGeometry destination = PlateGeometry.getById(dc, destinationGeometryId);
			PlateMapping pm = PlateMapping.getNew(dc, source, 4, destination, 1);
			pm.setName("Test plate mapping");
			pm.setDescription("Maps 4 5x5 plates to 1 10x10 plate");

			for (int plate = 0; plate < 4; plate++)
			{
				int rowOffset = (plate == 0 || plate == 1) ? 0 : 5;
				int columnOffset = (plate == 0 || plate == 2) ? 0 : 5;
				for (int row = 0; row < 5; row++)
				{
					for (int column = 0; column < 5; column++)
					{
						MappingCoordinate from = new MappingCoordinate(plate, row, column);
						MappingCoordinate to = new MappingCoordinate(
							0, row + rowOffset, column + columnOffset);
						pm.setSourceCoordinate(to, from);
					}
				}
			}
			dc.saveItem(pm);
			dc.commit();
			id = pm.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(pm, false);
			write_item(0, pm);
			write("--Create plate mapping OK");
		}
		catch (Exception ex)
		{
			write("--Create plate mapping FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PlateMapping pm = PlateMapping.getById(dc, id);
			write_item(0, pm);
			write("--Load plate mapping OK");
		}
		catch (Throwable ex)
		{
			write("--Load plate mapping FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<PlateMapping> l = PlateMapping.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List plate mappings OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List plate mappings FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PlateMapping pm = PlateMapping.getById(dc, id);
			dc.deleteItem(pm);
			dc.commit();
			write("--Delete plate mapping OK");
		}
		catch (Throwable ex)
		{
			write("--Delete plate mapping FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tSource\tDestination");
			write("-- \t-- \t--------- \t-----------\t------\t-----------");
		}
	}
	static void write_item(int i, PlateMapping pm)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+pm.getId()+"\t"+pm.getName()+"\t"+
			pm.getDescription()+"\t"+pm.getSourceGeometry()+"["+pm.getSourceCount()+"]\t"+
			pm.getDestinationGeometry()+"["+pm.getDestinationCount()+"]\t");
	}
	static void write_item(int i, Plate p, int sourceIndex)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+p.getId()+"\t"+p.getName()+"\t"+sourceIndex);
	}
	static void write(String message)
	{
		System.out.println(message);
	}

	static int test_apply_mapping(int plateMappingId, int[] sourcePlateIds, int destinationPlateTypeId)
	{
		if (plateMappingId == 0 || destinationPlateTypeId == 0) return 0;
		DbControl dc = null;
		int id = 0;
		try
		{
			dc = TestUtil.getDbControl();
			PlateMapping pm = PlateMapping.getById(dc, plateMappingId);
			PlateType destinationType = PlateType.getById(dc, destinationPlateTypeId);
			List<Plate> sourcePlates = new ArrayList<Plate>(sourcePlateIds.length);
			for (int i = 0; i < sourcePlateIds.length; i++)
			{
				if (sourcePlateIds[i] != 0)
				{
					sourcePlates.add(Plate.getById(dc, sourcePlateIds[i]));
				}
			}
			List<Plate> destinationPlates = pm.apply(sourcePlates, destinationType);
			Plate p = destinationPlates.get(0);
			p.setName("Test apply mapping");
			p.setDescription("Added at "+new Date());
			dc.saveItem(p);
			dc.commit();
			id = p.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(p, false);
			TestPlate.write_item(0, p);
			write("--Apply plate mapping OK");
		}
		catch (Exception ex)
		{
			write("--Apply plate mapping FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_list_parents(int plateId, int expectedResults)
	{
		if (plateId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Plate p = Plate.getById(dc, plateId);
			ItemResultList<Plate> l = p.getParentPlates().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				Plate parent = l.get(i);
				write_item(i, parent, p.getSourceIndex(parent));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List parent plates OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List parents plates FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_children(int plateId, int expectedResults)
	{
		if (plateId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Plate p = Plate.getById(dc, plateId);
			ItemResultList<Plate> l = p.getChildPlates().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				Plate child = l.get(i);
				write_item(i, child, child.getDestinationIndex());
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List child plates OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List child plates FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}



}

