/*
	$Id: TestRawBioAssay.java 5060 2009-08-19 07:02:11Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.util.ConsoleProgressReporter;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.parser.FlatFileParser;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

public class TestRawBioAssay
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing raw bioassays");
		write_header();
		
		int hybridizationId = TestHybridization.test_create(0, false);
		int scanId = TestScan.test_create(hybridizationId, 0, 0, false);
		int arrayDesignId = TestArrayDesign.test_create(Platform.GENERIC, false);
		int arrayDesignId2 = TestArrayDesign.test_create(Platform.GENERIC, false);
		
		TestReporter.test_import_from_file("data/test.rawdata.import.txt", 
			"\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*", "\\t", 4, 3);
		TestArrayDesign.test_import_from_file(arrayDesignId, "data/test.rawdata.import.txt", 
			"\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*", null, 0, 2, 1, 4);
		TestArrayDesign.test_import_from_file(arrayDesignId2, "data/test.rawdata.import.txt", 
				"\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*", null, 0, 2, 1, 4);
		
		int softwareId = TestSoftware.test_create(SystemItems.getId(SoftwareType.FEATURE_EXTRACTION), null, false);
		int protocolId = TestProtocol.test_create(SystemItems.getId(ProtocolType.FEATURE_EXTRACTION), null, false);
		
		// Standard tests: create, load, list
		int id = test_create(Platform.GENERIC, "genepix", null, scanId, softwareId, protocolId, arrayDesignId, 1, true);
		int id2 = test_create(Platform.GENERIC, "genepix", null, 0, 0, 0, 0, 1, false);		
		int id3 = test_create(Platform.AFFYMETRIX, null, null, 0, 0, 0, 0, 1, false);
		
		test_load(id);
		
		// Test adding raw data -- stored in database
		test_add_rawdata(id2, 10);
		test_import_from_file(id, "data/test.rawdata.import.txt");
		test_load(id);
		test_list(-1);
		write_raw_data_header();
		test_list_raw_data(id, 10, 5, 0, true);
		
		// Test switch array design
		test_change_arraydesign(id, arrayDesignId2);
		
		// Test adding raw data -- stored in files
		int celId = TestFile.test_create("data/test.affymetrix.cel", false, false);
		test_set_file(id3, "affymetrix.cel", celId);
		test_list_files(id3, 1);
		test_list_file_types(id3, 1);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_delete_raw_data(id);

		// Standard test: Delete
		test_delete(id);
		test_delete(id2);
		test_delete(id3);

		TestFile.test_delete(celId);
		TestScan.test_delete(scanId);
		TestHybridization.test_delete(hybridizationId);
		TestArrayDesign.test_delete(arrayDesignId);
		TestArrayDesign.test_delete(arrayDesignId2);
		TestSoftware.test_delete(softwareId);
		TestProtocol.test_delete(protocolId);
		TestReporter.test_delete();
		write("++Testing raw bioassays "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(String platformId, String rawDataType, String name, int scanId, int softwareId, int protocolId, int arrayDesignId, int arrayNum, boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.RAWBIOASSAY)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Platform platform = Platform.getByExternalId(dc, platformId);
			RawBioAssay rba = RawBioAssay.getNew(dc, platform, RawDataTypes.getRawDataType(rawDataType));
			if (setAll)
			{
				rba.setName("Test raw bioassay");
				rba.setDescription("Added at "+new Date());
			}
			if (name != null) rba.setName(name);
			rba.setArrayNum(arrayNum);
			if (softwareId != 0)
			{
				rba.setSoftware(Software.getById(dc, softwareId));
			}
			if (protocolId != 0)
			{
				rba.setProtocol(Protocol.getById(dc, protocolId));
			}
			if (scanId != 0)
			{
				rba.setScan(Scan.getById(dc, scanId));
			}
			if (arrayDesignId != 0)
			{
				rba.setArrayDesign(ArrayDesign.getById(dc, arrayDesignId));
			}
			dc.saveItem(rba);
			dc.commit();
			id = rba.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(rba, false);
			write_item(0, rba);
			write("--Create raw bioassay OK");
		}
		catch (Throwable ex)
		{
			write("--Create raw bioassay FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, id);
			write_item(0, rba);
			if (!TestUtil.getSilent())
			{
				write("Headers:");
				for (String header : rba.getHeaderNames())
				{
					write("\t"+header+"="+rba.getHeader(header));
				}
			}
			write("--Load raw bioassay OK");
		}
		catch (Throwable ex)
		{
			write("--Load raw bioassay FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<RawBioAssay> l = RawBioAssay.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List raw bioassays OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List raw bioassays FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, id);
			dc.deleteItem(rba);
			dc.commit();
			write("--Delete raw bioassay OK");
		}
		catch (Throwable ex)
		{
			write("--Delete raw bioassay FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			List<RawBioAssay> l = new ArrayList<RawBioAssay>(RawBioAssay.getQuery().list(dc));
			dc.commit();
			Trashcan.delete(TestUtil.getSessionControl(), l, true, null);
			write("--Delete all raw bioassays OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all raw bioassays FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tData type\tDb spots\tFile spots\tBytes\tArray Design\tSoftware\tProtocol");
			write("-- \t-- \t--------- \t-----------\t---------\t--------\t-----------\t-----\t------------\t--------\t--------");
		}
	}
	static void write_raw_data_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tBlock\tRow\tColumn\tGrid X\tGrid Y\tX    \tY    \tPosition\tReporter\tFeature\tExtra");
			write("-- \t-- \t-----\t---\t------\t------\t------\t-----\t-----\t--------\t--------\t-------\t-----");
		}
	}
	static void write_item(int i, RawBioAssay rba)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+rba.getId()+"\t"+rba.getName()+"\t"+rba.getDescription()+
				"\t"+rba.getRawDataType().getName()+"\t"+rba.getNumDbSpots()+"\t"+rba.getNumFileSpots() + 
				"\t"+rba.getBytes()+"\t"+rba.getArrayDesign()+
				"\t"+rba.getSoftware()+"\t"+rba.getProtocol());
		}
	}
	
	static void write_item(int i, RawData rd, boolean writeExtended)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+rd.getId()+"\t"+rd.getBlock()+"\t"+rd.getRow()+
				"\t"+rd.getColumn()+"\t"+rd.getMetaGridX()+"\t"+rd.getMetaGridY()+
				"\t"+rd.getX()+"\t"+rd.getY()+"\t"+rd.getPosition()+
				"\t"+rd.getReporter()+"\t"+rd.getFeature()+(writeExtended ? "\t"+rd.getAllExtended() : ""));
		}
	}
	
	static void write_item(int i, FileSetMember member)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+member.getId()+"\t"+member.getFile().getName()+"\t"+
				member.getDataFileType().getName());
		}
	}
	
	static void write_item(int i, DataFileType type)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+type.getId()+"\t"+type.getName()+"\t"+
				type.getDescription());
		}
	}

	
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void test_add_rawdata(int rawBioAssayId, int numSpots)
	{
		if (rawBioAssayId == 0) return;
		DbControl dc = null;
		long time;
		try
		{
			write("--Start adding raw data ("+numSpots+" spots)");
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, rawBioAssayId);
			RawDataType rdt = rba.getRawDataType();
			List<RawDataProperty> extraProperties = rdt.getProperties();
			RawDataBatcher rdb = rba.getRawDataBatcher(null, null);
			int block = 1;
			int row = 1;
			int column = 1;
			time = System.currentTimeMillis();
			for (int i = 0; i < numSpots; i++)
			{
				RawData rd = rdb.newRawData();
				rd.setBlock(block);
				rd.setRow(row);
				rd.setColumn(column);
				rd.setX((float)Math.random() * 50000);
				rd.setY((float)Math.random() * 50000);
				
				Map<String, Object> extra = rd.getAllExtended();
				for (RawDataProperty p : extraProperties)
				{
					Type t = p.getType();
					Object value = null;
					if (t == Type.INT)
					{
						value = (int)(Math.random() * 50000);
					}
					else if (t == Type.FLOAT)
					{
						value =  (float)Math.random() * 50000;
					}
					else if (t == Type.STRING)
					{
						value =  "ABCDEF";
					}
					//System.out.println("Add value: "+p.getName()+": "+t+": "+value);
//					rd.setExtended(p.getName(), value);
					extra.put(p.getName(), value);
				}
				rdb.insert(rd);
				row++;
				if (row == 11)
				{
					row = 1;
					column++;
					if (column == 11)
					{
						column = 1;
						block++;
					}
				}
			}
			dc.commit();
			time = System.currentTimeMillis()-time;
			write("--Add raw data OK ("+time+" ms)");
		}
		catch (Throwable ex)
		{
			write("--Add raw data FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static boolean test_list_raw_data(int rawBioAssayId, int limit, int offset, long expectedTotal, boolean writeExtended)
	{
		if (rawBioAssayId == 0) return ok;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, rawBioAssayId);
			DataQuery<RawData> query = rba.getRawData();
			query.join(Hql.leftJoin(null, "feature", "ft", null, true));
			query.join(Hql.leftJoin(null, "reporter", "rep", null, true));
			query.setMaxResults(limit);
			query.setFirstResult(offset);
			query.setReturnTotalCount(true);

			DataResultIterator<RawData> it = query.iterate(dc);
			int counter = 0;
			while (it.hasNext())
			{
				RawData rd = it.next();
				if (counter == 0)
				{
					write_item(0, RawDataUtil.getRawBioAssay(dc, rd));
				}
				write_item(counter, rd, writeExtended);
				counter++;
			}
			long totalCount = it.getTotalCount();
			if (expectedTotal > 0 && expectedTotal != totalCount)
			{
				throw new Exception("Unexpected number of raw data: " + totalCount +"; expected " +expectedTotal);
			}
			write("--List raw data OK (listed = "+counter+"; total = "+totalCount+")");
		}
		catch (Throwable ex)
		{
			write("--List raw data FAILED");
			ex.printStackTrace();
			ok = false;
			return false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return true;
	}
	
	static void test_delete_raw_data(int rawBioAssayId)
	{
		if (rawBioAssayId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, rawBioAssayId);
			rba.deleteRawData();
			write("--Delete raw data OK");
		}
		catch (Throwable ex)
		{
			write("--Delete raw data FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_import_from_file(int rawBioAssayId, String filename)
	{
		if (rawBioAssayId == 0) return;
		write("--Start import raw data from file (" + filename + ")");
		DbControl dc = null;
		long time;
		int i = 0;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, rawBioAssayId);
			RawDataBatcher rawDataBatcher = rba.getRawDataBatcher(null, null);
			ReporterBatcher reporterBatcher = ReporterBatcher.getNew(dc);
		
			FlatFileParser parser = new FlatFileParser();
			parser.setHeaderRegexp(Pattern.compile("\"(.*)=(.*)\""));
			parser.setDataHeaderRegexp(Pattern.compile("\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*"));
			parser.setDataSplitterRegexp(Pattern.compile("\\t"));
			parser.setInputStream(FileUtil.getInputStream(new java.io.File(filename)), "ISO-8859-1");
			parser.parseHeaders();
			
			for (String header : parser.getHeaderNames())
			{
				String value = parser.getHeader(header);
				if (header != null && value != null) rba.setHeader(header, value);
			}
			
			time = System.currentTimeMillis();
			while (parser.hasMoreData())
			{
				FlatFileParser.Data parsedData = parser.nextData();
				RawData rawData = rawDataBatcher.newRawData();
				rawData.setBlock(Values.getInteger(parsedData.get(0), null));
				rawData.setRow(Values.getInteger(parsedData.get(2), null));
				rawData.setColumn(Values.getInteger(parsedData.get(1), null));
				rawData.setX(Values.getFloat(parsedData.get(5), null));
				rawData.setY(Values.getFloat(parsedData.get(6), null));
				String reporterId = parsedData.get(4);
				rawData.setReporter(reporterId == null ? null : reporterBatcher.getByExternalId(reporterId));
				
				Map<String, Object> extraData = rawData.getAllExtended();
				extraData.put("diameter", Values.getFloat(parsedData.get(7), null));
				extraData.put("ch1FgMedian", Values.getFloat(parsedData.get(8), null));
				extraData.put("ch1FgMean", Values.getFloat(parsedData.get(9), null));
				extraData.put("ch1FgSd", Values.getFloat(parsedData.get(10), null));
				extraData.put("ch1BgMedian", Values.getFloat(parsedData.get(11), null));
				extraData.put("ch1BgMean", Values.getFloat(parsedData.get(12), null));
				extraData.put("ch1BgSd", Values.getFloat(parsedData.get(13), null));
				extraData.put("ch1PercSd1", Values.getInteger(parsedData.get(14), null));
				extraData.put("ch1PercSd2", Values.getInteger(parsedData.get(15), null));
				extraData.put("ch1PercSat", Values.getInteger(parsedData.get(16), null));
				extraData.put("ch2FgMedian", Values.getFloat(parsedData.get(17), null));
				extraData.put("ch2FgMean", Values.getFloat(parsedData.get(18), null));
				extraData.put("ch2FgSd", Values.getFloat(parsedData.get(19), null));
				extraData.put("ch2BgMedian", Values.getFloat(parsedData.get(20), null));
				extraData.put("ch2BgMean", Values.getFloat(parsedData.get(21), null));
				extraData.put("ch2BgSd", Values.getFloat(parsedData.get(22), null));
				extraData.put("ch2PercSd1", Values.getInteger(parsedData.get(23), null));
				extraData.put("ch2PercSd2", Values.getInteger(parsedData.get(24), null));
				extraData.put("ch2PercSat", Values.getInteger(parsedData.get(25), null));

				try
				{
					rawDataBatcher.insert(rawData);
					i++;
				}
				catch (ItemAlreadyExistsException ex)
				{
					// Ignore duplicate features
				}
			}
			dc.commit();
			time = System.currentTimeMillis()-time;

			write("--Import raw data from file OK ("+i+", time="+time+" ms)");
		}
		catch (Throwable ex)
		{
			write("--Import raw data from file FAILED ("+i+")");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_set_file(int rawBioAssayId, String fileType, int fileId)
	{
		if (rawBioAssayId == 0 || fileId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, rawBioAssayId);
			DataFileType type = DataFileType.getByExternalId(dc, fileType);
			File file = File.getById(dc, fileId);
			rba.getFileSet().setMember(file, type);
			dc.commit();
			write("--Set file OK ("+fileType+")");
		}
		catch (Throwable ex)
		{
			write("--Set file FAILED ("+fileType+")");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_list_files(int rawBioAssayId, int expectedResults)
	{
		if (rawBioAssayId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, rawBioAssayId);
			ItemResultList<FileSetMember> l = rba.getFileSet().getMembers().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List files OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List files FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_list_file_types(int rawBioAssayId, int expectedResults)
	{
		if (rawBioAssayId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, rawBioAssayId);
			ItemResultList<DataFileType> l = rba.getFileSet().getMemberTypes().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List file types OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List file types FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_change_arraydesign(int id, int arrayDesignId)
	{
		if (id == 0 || arrayDesignId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			RawBioAssay rba = RawBioAssay.getById(dc, id);
			ArrayDesign ad = ArrayDesign.getById(dc, arrayDesignId);
			ProgressReporter progress = TestUtil.getSilent() ?
					null : new ConsoleProgressReporter();
			rba.updateArrayDesign(ad, null, progress);
			dc.commit();
			write("--Change array design OK");
		}
		catch (Throwable ex)
		{
			write("--Change array design FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

}
