/**
	$Id: SessionClient.java 4513 2008-09-11 20:06:36Z jari $

	Copyright (C) 2007 Martin Svensson, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.ws.client;

import net.sf.basedb.info.VersionInfo;

import org.apache.axis2.AxisFault;

/**
 	Client for handling sessions.
 	{@link net.sf.basedb.core.Session}
  	@author Nicklas, Martin
  	@version 2.5
 */
public class SessionClient 
	extends AbstractRPCClient
{
	private final String url;
	private final String ID;
	
	/**
	 	Creates a new SessionClient using the default service factory.
	 	
	  	@param url The url where web services are located.
	  	@param ID Session id of a logged in session. A new session will be created if this is NULL.
	  	@param externalClientId External id of the client that is registered in BASE.
	  	@throws AxisFault If communication with the web service fails.
	*/
	public SessionClient(String url, String ID, String externalClientId)
		throws AxisFault
	{
		this(url, ID, externalClientId, null);
	}

	/**
	 	Creates a new SessionClient using a specific service factory.
	  	@param url The url where web services are located.
	  	@param ID Session id of a logged in session. A new session will be created if this is NULL.
	  	@param externalClientId External id of the client that is registered in BASE.
	  	@param serviceFactory The service factory to use, or null to
	  		use the default: {@link Factories#getDefaultServiceFactory()}
	  	@throws AxisFault If communication with the web service fails
	  	@since 2.8
	*/
	public SessionClient(String url, String ID, String externalClientId, ServiceFactory serviceFactory)
		throws AxisFault
	{
		super(url, "Session", serviceFactory);
		this.url = url;
		this.ID = ID != null ? ID : invokeBlocking("newSession", String.class, externalClientId);
	}

	
	/**
	 	Gets the URL of the session
	  	@return URL as a String
	 */
	public String getUrl()
	{
		return url;
	}
	
	/**
	 	Gets the id of the session.
	  	@return Session ID as a String
	*/
	public String getId()
	{
		return ID;
	}
	
	/**
	 	Calling the getChallenge method.
	  	@return Gets a random string to use with password encryption.
	  	@throws AxisFault If communication with web service fails.
	 */
	public String getChallenge()
		throws AxisFault
	{
		return invokeBlocking("getChallenge", String.class, ID);
	}
	
	/**
	 	Login to BASE 
	  	@param login Login name on BASE server
	  	@param password Password on BASE server
	  	@param comment A comment to put on the session
	  	@param encrypted If the password should be encrypted when logging in.
	  	@throws AxisFault If something goes wrong when calling the login service
	 */
	public void login(String login, String password, String comment, boolean encrypted)
		throws AxisFault
	{
		invokeBlocking("login", ID, login, password, comment, encrypted);
	}
	
	/**
	 	Logs out from BASE
	  	@throws AxisFault If something goes wrong 
	  		calling the logout service.
	 */
	public void logout()
		throws AxisFault
	{
		invokeBlocking("logout", ID);
	}
	
	/**
	 	Gets how long time it takes before
	 	the session timeout.
	  	@return Number of minutes before the session timeout
	  	@throws AxisFault If something goes wrong calling the services.
	 */
	public int getSessionTimeOut()
		throws AxisFault
	{
		return invokeBlocking("getSessionTimeOut", Integer.class, ID);
	}
	
	/**
	 	Refresh a session to prevent a timeout.
	  	@param ID The id of the session to refresh.
	  	@throws AxisFault If anything goes wrong calling the service.
	 */
	public void refreshSession(String ID)
		throws AxisFault
	{
		invokeBlocking("refreshSession", ID);
	}
	
	
	/**
	  	@return A {@link VersionInfo} object
	  	@throws AxisFault
	*/
	public VersionInfo getVersion()
		throws AxisFault
	{
		return invokeBlocking("getVersion", VersionInfo.class);
	}
}
