/**
	$Id: BioAssaySetService.java 4931 2009-05-14 11:16:12Z nicklas $

	Copyright (C) 2007 Martin Svensson, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.ws.server;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.info.AnnotationInfo;
import net.sf.basedb.info.DataFileTypeInfo;
import net.sf.basedb.info.QueryOptions;

import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import javax.activation.DataSource;

import org.apache.axiom.om.OMElement;

/**
 	Web service for BioAssaySet {@link net.sf.basedb.core.BioAssaySet}.
 	This class provides services to get information and data from BioAssaySet.
  	@author Martin
  	@version 2.5
 */
public class BioAssaySetService 
	extends AbstractRPCService
{	
	private static final String SERIAL_BASEFILE = "Serial BASEfile";
	private static final String MATRIX_BASEFILE = "Matrix BASEfile";
	private static final String MEV = "Mev file";
	
	/**
	 	
	 */
	public BioAssaySetService()
	{}
	
	/**
	 	Gets the available file formats a bioassay set can be exported in 
	 	through webservices. 
	  	@param ID A string that identify the active session
	  	@return A string-array with the available formats or 
	  		null if there is no active session.
	  	@deprecated In 2.12. Use {@link #getDataFileTypes(String, int, QueryOptions)}
	  		instead
	*/
	public String[] getExportFormats(String ID)
	{
		String[] formats = {SERIAL_BASEFILE, MATRIX_BASEFILE, MEV};
		return formats;	
	}
	
	/**
	 	Returns an OMElement with an attached file that contains the exported bioassay set.
	 	Different file formats to export the bioassay set is available
	 	There may be some parameters that needs to be configured 
	 	depending on the selected file format.
	  	@param ID An id to identify the active session with
	  	@param bioAssaySetId The id of the bioassay-set, which should be exported. 
	  	@param format File format of which the bioassay set should be exported in.
	  		Available formats are:
	  		<ul>
	  			<li>{@value #MEV} (deprecated in BASE 2.12). MeV support has been 
	  				completely moved to the MeV extension (http://baseplugins.thep.lu.se/wiki/net.sf.basedb.mev)
	  				This web service will continue to 
	  			<li>{@value #SERIAL_BASEFILE}
	  			<li>{@value #MATRIX_BASEFILE}
	  		</ul>	  			
	  	@param reporterFieldArray The reporter fields to include in the export
	  	@param spotFieldArray The spot fields to include in the export
	  	@param mergeReporters TRUE to merge spot data having the same reporter.
	  		see {@link net.sf.basedb.plugins.BioAssaySetExporter}
	  	@return An OMElement containing the exported file or 
	  		NULL if there an exception is thrown.
		@throws Exception If the data could not be exported to the file.
		@deprecated In 2.12. Spot data should not be exported dynamically. Instead,
			plug-ins should be used and attach the result to the bioassay set and
			the files can then be downloaded when needed. Use
			{@link #downloadDataFile(String, int, String)} instead.
	*/
	@SuppressWarnings("unchecked")
	public OMElement downloadBioAssaySet(String ID, int bioAssaySetId, String format, String[] reporterFieldArray, 
			String[] spotFieldArray, Boolean mergeReporters)
		throws Exception
	{
		if (format == null) return null;
		
		String elementName = "downloadFile";
		OMElement elementWithAttachment = null;
		
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		File exportFile = null;		
		FileOutputStream fos = null;
		try
		{
			BioAssaySet bioassayset = BioAssaySet.getById(dc, bioAssaySetId);
			List<String> reporterFields = new ArrayList<String>();
			List<String> spotFields = new ArrayList<String>();
			if (reporterFieldArray != null) reporterFields = Arrays.asList(reporterFieldArray);
			if (spotFieldArray != null) spotFields = Arrays.asList(spotFieldArray);
			mergeReporters = mergeReporters == null ? false : mergeReporters;
			
			exportFile =  File.createTempFile("BioAssaySetFile", null);
			fos = new FileOutputStream(exportFile);
			
			net.sf.basedb.plugins.BioAssaySetExporter exporter = 
				new net.sf.basedb.plugins.BioAssaySetExporter();
			//Export to MeV-file format			
			if (format.equals(MEV))
			{							
				exporter.exportMeV(bioassayset, fos, null);				
			}						
			//Export the data from the bioassayset as one big matrix in a BASEfile
			else if(format.equals(MATRIX_BASEFILE))
			{
				exporter.exportBaseFileMatrix(bioassayset, fos, Collections.EMPTY_MAP, reporterFields, spotFields, mergeReporters, null);
			}
		 	//Export the data from a bioassayset as a serial BASEfile.			 
			else if (format.equals(SERIAL_BASEFILE))
			{
				exporter.exportBaseFileSerial(bioassayset, fos, Collections.EMPTY_MAP, reporterFields, spotFields, mergeReporters, null);
			}
			else
			{
				throw new BaseException("BASE web services does not support this format: " + format);
			}
			fos.close();
			DataSource source = new TempFileDataSource(exportFile, "text/plain");
			elementWithAttachment = attachFile(source, elementName);
		}
		finally
		{		
			if (dc != null) dc.close();
		}
		
		return elementWithAttachment;
	}
	
	/**
	 	Returns a data file of a certain type attached to an OMElement.
	  	@param ID Used to identify an active session with.
	  	@param bioAssaySetId Id of the bioassay set the file belongs to.
	  	@param dataFileType The external ID of the file type.
	  	@return OMElement with the file attached, or NULL if no file was found.
	  	@see DataFileType
	  	@since 2.12
	*/
	public OMElement downloadDataFile(String ID, int bioAssaySetId, String dataFileType)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		OMElement toDownload = null;
		try
		{
			BioAssaySet bas = BioAssaySet.getById(dc, bioAssaySetId);
			if (bas.hasFileSet())
			{
				FileSet fileSet = bas.getFileSet();
				DataFileType type = DataFileType.getByExternalId(dc, dataFileType);
				if (fileSet.hasMember(type))
				{
					FileSetMember member = fileSet.getMember(type);
					BaseFileDataSource source = new BaseFileDataSource(member.getFile());
					toDownload = attachFile(source, "downloadFile");
				}
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}		
		return toDownload;
		
	}
	
	/**
	 	Checks if the files has been attached to bioassay set.
	  	@param ID Used to identify an active session with
	  	@param bioAssaySetId Id of the bioassay set it concerns
	  	@return TRUE if there are files to download, FALSE otherwise.
	  	@since 2.12
	*/
	public Boolean hasDownloadableData(String ID, int bioAssaySetId)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		BioAssaySet bas = null;
		try
		{
			bas = BioAssaySet.getById(dc, bioAssaySetId);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return bas != null ? bas.hasFileSet() : false;
	}

	/**
	 	Returns information objects for each DataFileType represented in the FileSet
	 	of a BioAssaySet.
	  	@param ID Id of the active session.
	  	@param bioAssaySetId Id of the bioassay set it concerns
	 	@param qOpt Restrictions and includes to put on the query
	  	@return An array with DataFileTypeInfo:s.
	  	@since 2.12
	*/
	public DataFileTypeInfo[] getDataFileTypes(String ID, int bioAssaySetId, QueryOptions qOpt)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		try
		{
			BioAssaySet bas = BioAssaySet.getById(dc, bioAssaySetId);
			if (bas.hasFileSet())
			{			
				ItemQuery<DataFileType> query = bas.getFileSet().getMemberTypes();
				query = util.getConfiguredItemQuery(query, qOpt);
				return util.listToInfo(query.list(dc), DataFileTypeInfo.class);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return new DataFileTypeInfo[0];
	}

	
	/**
		Get the annotation values for one or more annotation types.
		@param ID Id of the active session.
		@param bioAssaySetId Id of the bioassay set it concerns
		@param annotationTypes Array with the ID of the annotation types
		@param findInherited TRUE to look among inherited annotations, FALSE to only look
			at primary annotations
		@return An array with AnnotationInfo object, each element in the array 
			corresponds to the annotation type with the same index in the annotationTypes array
	*/
	public AnnotationInfo[] getAnnotations(String ID, int bioAssaySetId, Integer[] annotationTypes, boolean findInherited)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		try
		{
			BioAssaySet bas = BioAssaySet.getById(dc, bioAssaySetId);
			return util.getAnnotations(dc, bas, annotationTypes, findInherited);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}	
	
}