/**
	$Id: ServicesUtil.java 5256 2010-02-22 10:42:36Z nicklas $

	Copyright (C) 2007 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.ws.server;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.info.AnnotationInfo;
import net.sf.basedb.info.QueryOptions;
import net.sf.basedb.info.QueryParameter;
import net.sf.basedb.info.ToTransferable;
import net.sf.basedb.info.Transferable;

import java.util.EnumSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 	A utility class for common methods in the web services.
  	@author Martin, Nicklas
  	@version 2.5
 */
public class ServicesUtil
{

	/**
	 	Put options and restriction on a item query and returns it 
	  	@param <T> The kind of BasicItem the item query work with.
	  	@param query ItemQuery that should get includes and restriction.
	  	@param qOpt QueryOptions object with restrictions and includes
	  		to put on the ItemQuery
	  	@return A ItemQuery with restriction and includes set.
	 */
	public <T extends BasicItem> ItemQuery<T> getConfiguredItemQuery(ItemQuery<T> query, QueryOptions qOpt)
	{		
		Set<Include> includes = EnumSet.noneOf(Include.class);
		if (qOpt == null) qOpt = new QueryOptions();
		if (qOpt.getIncludeMine()) includes.add(Include.MINE);
		if (qOpt.getIncludeShared()) includes.add(Include.SHARED);
		if (qOpt.getIncludeInProject()) includes.add(Include.IN_PROJECT);
		if (qOpt.getIncludeOthers()) includes.add(Include.OTHERS);
		if (qOpt.getIncludeNotRemoved()) includes.add(Include.NOT_REMOVED);
		
		query.setIncludes(includes);
		if (qOpt.getRestriction() != null)
		{
			query.restrict(Hql.restriction(qOpt.getRestriction(), qOpt.getRestrictionPrefix()));
		}
		QueryParameter[] parameters = qOpt.getParameters();
		if (parameters != null && parameters.length > 0)
		{
			for (QueryParameter qp : parameters)
			{
				Type valueType = Type.valueOf(qp.getType());
				Object value = valueType.parseString(qp.getValue());
				query.setParameter(qp.getName(), value, valueType);
			}
		}
		return query;
	}
	
	/**
		Convert a list of objects to an array of {@link Transferable} objects
		objects. The objects in must be of the same type and implement the
		{@link ToTransferable} interface. The specified transferable class
		must hava a public no-argument constructor.
		
		See for example: {@link net.sf.basedb.core.Experiment#toTransferable(net.sf.basedb.info.ExperimentInfo)}
	  	@param <T> The kind of info object the items can be transfered into.
		@param list A list of {@link ToTransferable} objects
		@param clazz The {@link Transferable} class that the objects in the list
			are beeing converted to
		@return An array info objects of the specified transferable class
	*/
	@SuppressWarnings("unchecked")
	public <T extends Transferable> T[] listToInfo(List<? extends ToTransferable<T>> list, 
		Class<T> clazz)
	{
		T[] result = (T[])java.lang.reflect.Array.newInstance(clazz, list.size());
		try
		{
			int i = 0;
			for (ToTransferable<T> o : list)
			{
				result[i] = o.toTransferable(clazz.newInstance());
				++i;
			}
		}
		catch (InstantiationException ex)
		{
			throw new RuntimeException(ex);
		}
		catch (IllegalAccessException ex)
		{
			throw new RuntimeException(ex);
		}
		return result;
	}
	
	/**
		Load annotation values from an annotatable item.
		@param dc The DbControl to use for database access
		@param annotatable The annotatable item to load annotations from
		@param annotationTypes Array of annotation type ID
		@param findInherited TRUE to find inherited annotations, FALSE to only look
			for primary annotations
		@return An array of {@link AnnotationInfo} objects, each position corresponds
			to the position with the same index in the annotationTypes array. The returned
			array is always the same length as the annotationTypes array. If the annotatable
			item doesn't have any annotations of a particular type, that element is null
	*/
	public AnnotationInfo[] getAnnotations(DbControl dc, Annotatable annotatable, 
		Integer[] annotationTypes, boolean findInherited)
	{
		AnnotationInfo[] result = new AnnotationInfo[annotationTypes.length];
		if (annotatable.isAnnotated())
		{
			AnnotationSet as = annotatable.getAnnotationSet();
			int i = 0;
			for (Integer atId : annotationTypes)
			{
				if (atId != null)
				{
					AnnotationInfo info = new AnnotationInfo();
					AnnotationType at = AnnotationType.getById(dc, atId);
					List<Annotation> annotations = as.findAnnotations(dc, at, findInherited);
					List<Object> values = new LinkedList<Object>();
					for (Annotation a : annotations)
					{
						values.addAll(a.getValues());
					}
					info.setAnnotationTypeId(atId);
					info.setValueType(at.getValueType().name());
					info.setValues(net.sf.basedb.util.Values.toStrings(values, null));
					result[i] = info;
				}
				++i;
			}
		}
		return result;
	}
}
